Package: ggseg3d
Title: Interactive 3D Brain Atlas Visualization
Version: 2.1.1
Authors@R: c(
    person("Athanasia Mo", "Mowinckel", , "a.m.mowinckel@psykologi.uio.no", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-5756-0223")),
    person("Didac", "Vidal-Piñeiro", , "d.v.pineiro@psykologi.uio.no", role = "aut",
           comment = c(ORCID = "0000-0001-9997-9156")),
    person("Center for Lifespan Changes in Brain and Cognition (LCBC), University of Oslo",
           role = "cph"),
    person("three.js authors", role = c("ctb", "cph"),
           comment = "Three.js and OrbitControls libraries, see LICENSE.note")
  )
Description: Plot brain atlases as interactive 3D meshes using 'Three.js'
    via 'htmlwidgets', or render publication-quality static images through
    'rgl' and 'rayshader'. A pipe-friendly API lets you map data onto
    brain regions, control camera angles, toggle region edges, overlay
    glass brains, and snapshot or ray-trace the result. Additional
    atlases are available through the 'ggsegverse' r-universe. Mowinckel
    & Vidal-Piñeiro (2020) <doi:10.1177/2515245920928009>.
License: MIT + file LICENSE
URL: https://github.com/ggsegverse/ggseg3d,
        https://ggsegverse.github.io/ggseg3d/
BugReports: https://github.com/ggsegverse/ggseg3d/issues
Depends: R (>= 4.1)
Imports: cli, dplyr, ggseg.formats, htmlwidgets, knitr, lifecycle,
        rlang, scales, tidyr, webshot2
Suggests: ggseg.meshes, png, rayshader, rgl, rmarkdown, shiny,
        spelling, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/Needs/website: ggsegverse/ggseg.docs
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-22 10:20:29 UTC; athanasm
Author: Athanasia Mo Mowinckel [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5756-0223>),
  Didac Vidal-Piñeiro [aut] (ORCID:
    <https://orcid.org/0000-0001-9997-9156>),
  Center for Lifespan Changes in Brain and Cognition (LCBC), University
    of Oslo [cph],
  three.js authors [ctb, cph] (Three.js and OrbitControls libraries, see
    LICENSE.note)
Maintainer: Athanasia Mo Mowinckel <a.m.mowinckel@psykologi.uio.no>
Repository: CRAN
Date/Publication: 2026-04-22 11:00:02 UTC
Built: R 4.6.0; ; 2026-04-30 11:53:47 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(knit_print,ggsegray)
S3method(prepare_brain_meshes,cerebellar_atlas)
S3method(prepare_brain_meshes,cortical_atlas)
S3method(prepare_brain_meshes,default)
S3method(prepare_brain_meshes,subcortical_atlas)
S3method(prepare_brain_meshes,tract_atlas)
S3method(print,ggsegray)
export(add_glassbrain)
export(aseg)
export(dk)
export(ggseg3d)
export(ggseg3dOutput)
export(ggsegray)
export(pan_camera)
export(renderGgseg3d)
export(resolve_brain_mesh)
export(set_background)
export(set_dimensions)
export(set_edges)
export(set_flat_shading)
export(set_legend)
export(set_orthographic)
export(set_positioning)
export(snapshot_brain)
export(tracula)
export(updateGgseg3dBackground)
export(updateGgseg3dCamera)
import(ggseg.formats)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(ggseg.formats,aseg)
importFrom(ggseg.formats,dk)
importFrom(ggseg.formats,tracula)
importFrom(graphics,par)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,rect)
importFrom(graphics,text)
importFrom(htmlwidgets,createWidget)
importFrom(htmlwidgets,shinyRenderWidget)
importFrom(htmlwidgets,shinyWidgetOutput)
importFrom(htmlwidgets,sizingPolicy)
importFrom(knitr,knit_print)
importFrom(lifecycle,deprecate_warn)
importFrom(lifecycle,deprecated)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(scales,brewer_pal)
importFrom(scales,colour_ramp)
importFrom(scales,gradient_n_pal)
importFrom(scales,rescale)
importFrom(tidyr,unite)
importFrom(webshot2,webshot)

# ggseg3d <img src='man/figures/logo.png' align="right" height="138.5" />

<!-- badges: start -->

[![Coverage
Status](https://raw.githubusercontent.com/ggsegverse/ggseg3d/coverage/badges/coverage.svg)](https://github.com/ggsegverse/ggseg3d/actions/workflows/test-coverage.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/ggseg3d)](https://CRAN.R-project.org/package=ggseg3d)
[![downloads](https://cranlogs.r-pkg.org/badges/last-month/ggseg3d?color=blue)](https://r-pkg.org/pkg/ggseg3d)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![code-quality](https://github.com/ggsegverse/ggseg3d/actions/workflows/code-quality.yaml/badge.svg)](https://github.com/ggsegverse/ggseg3d/actions/workflows/code-quality.yaml)
[![R-CMD-check](https://github.com/ggsegverse/ggseg3d/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ggsegverse/ggseg3d/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

Interactive 3D brain atlas visualization in R. Plot brain parcellations
as WebGL meshes powered by Three.js, or render publication-quality
static images through rgl and rayshader. A pipe-friendly API lets you
map data onto brain regions, control camera angles, toggle region edges,
overlay glass brains, and snapshot the result.

## Installation

Install from CRAN:

``` r
install.packages("ggseg3d")
```

Or get the development version from the [ggsegverse
r-universe](https://ggsegverse.r-universe.dev):

``` r
options(
  repos = c(
    ggsegverse = "https://ggsegverse.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
  )
)
install.packages("ggseg3d")
```

## Atlases

Three atlases ship with the package (via
[ggseg.formats](https://ggsegverse.github.io/ggseg.formats/)):

1.  `dk` – Desikan-Killiany cortical atlas
2.  `aseg` – Automatic subcortical segmentation
3.  `tracula` – White-matter tract atlas

Additional atlases are available through the [ggsegverse
r-universe](https://ggsegverse.r-universe.dev).

## Usage

``` r
ggseg3d(atlas = dk(), hemisphere = "left") |>
  pan_camera("left lateral")
```

<img src="man/figures/README-dk_3d-plot.png" alt="" width="1200" />

Subcortical structures with a translucent glass brain overlay:

``` r
ggseg3d(atlas = aseg()) |>
  add_glassbrain() |>
  pan_camera("right lateral")
```

<img src="man/figures/README-aseg_3d-plot.png" alt="" width="1200" />

See the [package website](https://ggsegverse.github.io/ggseg3d/) for the
full walkthrough, rayshader rendering, and Shiny integration.

## Citation

Mowinckel & Vidal-Piñeiro (2020). *Visualization of Brain Statistics
With R Packages ggseg and ggseg3d.* Advances in Methods and Practices in
Psychological Science.
[doi:10.1177/2515245920928009](https://doi.org/10.1177/2515245920928009)

## Funding

This tool is partly funded by:

**EU Horizon 2020 Grant:** Healthy minds 0-100 years: Optimising the use
of European brain imaging cohorts (Lifebrain).

**Grant agreement number:** 732592.

**Call:** Societal challenges: Health, demographic change and well-being

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## -----------------------------------------------------------------------------
# library(shiny)
# library(ggseg3d)

## -----------------------------------------------------------------------------
# ui <- fluidPage(
#   titlePanel("Brain Atlas Viewer"),
#   sidebarLayout(
#     sidebarPanel(
#       selectInput(
#         "hemi",
#         "Hemisphere",
#         choices = c("left", "right", "both"),
#         selected = "left"
#       )
#     ),
#     mainPanel(
#       ggseg3dOutput("brain", height = "500px")
#     )
#   )
# )
# 
# server <- function(input, output, session) {
#   output$brain <- renderGgseg3d({
#     hemi <- if (input$hemi == "both") NULL else input$hemi
# 
#     ggseg3d(hemisphere = hemi) |>
#       pan_camera("left lateral")
#   })
# }
# 
# shinyApp(ui, server)

## -----------------------------------------------------------------------------
# ggseg3dOutput("brain", width = "100%", height = "600px")

## -----------------------------------------------------------------------------
# renderGgseg3d({
#   ggseg3d() |>
#     set_dimensions(width = 800, height = 600)
# })

## -----------------------------------------------------------------------------
# server <- function(input, output, session) {
#   brain_data <- reactive({
#     tibble(
#       region = c("precentral", "postcentral", "insula"),
#       value = runif(3)
#     )
#   })
# 
#   output$brain <- renderGgseg3d({
#     ggseg3d(
#       .data = brain_data(),
#       atlas = dk(), # nolint [object_usage_linter]
#       colour_by = "value"
#     ) |>
#       pan_camera("left lateral")
#   })
# }

## -----------------------------------------------------------------------------
# ui <- fluidPage(
#   sidebarLayout(
#     sidebarPanel(
#       selectInput(
#         "view",
#         "View",
#         choices = c(
#           "left lateral",
#           "left medial",
#           "right lateral",
#           "right medial"
#         )
#       ),
#       selectInput("bg", "Background", choices = c("white", "black", "grey"))
#     ),
#     mainPanel(
#       ggseg3dOutput("brain")
#     )
#   )
# )
# 
# server <- function(input, output, session) {
#   output$brain <- renderGgseg3d({
#     ggseg3d() |>
#       pan_camera("left lateral")
#   })
# 
#   observeEvent(input$view, {
#     updateGgseg3dCamera("brain", input$view)
#   })
# 
#   observeEvent(input$bg, {
#     updateGgseg3dBackground("brain", input$bg)
#   })
# }

## -----------------------------------------------------------------------------
# library(shiny)
# library(ggseg3d)
# library(dplyr)
# 
# example_data <- tibble(
#   region = c(
#     "precentral",
#     "postcentral",
#     "insula",
#     "superior parietal",
#     "inferior parietal",
#     "supramarginal",
#     "cuneus",
#     "pericalcarine"
#   ),
#   thickness = c(2.5, 2.3, 3.1, 2.2, 2.4, 2.6, 1.8, 1.9),
#   volume = c(8500, 7200, 6800, 9100, 8800, 7500, 4200, 3800)
# )
# 
# ui <- fluidPage(
#   titlePanel("Brain Metrics Explorer"),
#   sidebarLayout(
#     sidebarPanel(
#       selectInput("metric", "Metric", choices = c("thickness", "volume")),
#       selectInput(
#         "view",
#         "Camera View",
#         choices = c(
#           "left lateral",
#           "left medial",
#           "right lateral",
#           "right medial",
#           "left superior",
#           "left inferior"
#         )
#       ),
#       checkboxInput("edges", "Show edges", value = FALSE),
#       selectInput("bg", "Background", choices = c("white", "black", "grey90"))
#     ),
#     mainPanel(
#       ggseg3dOutput("brain", height = "600px")
#     )
#   )
# )
# 
# server <- function(input, output, session) {
#   output$brain <- renderGgseg3d({
#     p <- ggseg3d(
#       .data = example_data,
#       atlas = dk(), # nolint [object_usage_linter]
#       colour_by = input$metric,
#       text_by = input$metric
#     ) |>
#       pan_camera(input$view) |>
#       set_background(input$bg)
# 
#     if (input$edges) {
#       p <- p |> set_edges("black")
#     }
# 
#     p
#   })
# }
# 
# shinyApp(ui, server)

#' Plot 3D brain parcellations
#'
#' `ggseg3d` creates and returns an interactive Three.js brain mesh
#' visualization. Dispatches to atlas-type-specific methods via
#' [prepare_brain_meshes()].
#'
#' @author Athanasia Mowinckel and Didac Piñeiro
#'
#' @param .data A data.frame to use for plot aesthetics. Must include a
#'   column called "region" corresponding to regions.
#' @param atlas A `ggseg_atlas` object containing 3D vertex mappings, or a
#'   string naming an atlas function (deprecated).
#' @param label_by String. Column name used as hover label for each region.
#' @param text_by String. Column name for extra hover text shown below the
#'   region label.
#' @param colour_by String. Column name mapped to mesh colours.
#' @param label,text,colour `r lifecycle::badge("deprecated")` Use
#'   `label_by`, `text_by`, and `colour_by` instead.
#' @param palette String. Vector of colour names or HEX colours. Can also
#'   be a named numeric vector, with colours as names, and breakpoint for
#'   that colour as the value
#' @param na_colour String. Either name, hex of RGB for colour of NA in
#'   colour.
#' @param na_alpha Numeric. A number between 0 and 1 to control
#'   transparency of NA-regions.
#' @param ... Type-specific arguments passed to the atlas method.
#'   See section **Type-specific arguments** below.
#'
#' @template type-specific-args
#'
#' @return an htmlwidget object for interactive 3D brain visualization
#'
#' @seealso [pan_camera()] for camera position, [set_background()] for
#'   background colour, [set_legend()] for legend visibility
#'
#' @importFrom dplyr filter full_join select distinct summarise mutate
#' @importFrom scales colour_ramp brewer_pal rescale gradient_n_pal
#' @importFrom tidyr unite
#' @importFrom htmlwidgets createWidget sizingPolicy
#' @importFrom lifecycle deprecate_warn deprecated
#' @importFrom rlang %||%
#'
#' @examples
#' ggseg3d()
#' ggseg3d(hemisphere = "left") |> pan_camera("left lateral")
#' ggseg3d() |> set_legend(FALSE)
#' ggseg3d() |> set_background("black")
#'
#' @export
ggseg3d <- function(
  .data = NULL,
  atlas = dk(),
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  ...,
  label = deprecated(),
  text = deprecated(),
  colour = deprecated()
) {
  if (lifecycle::is_present(label)) {
    lifecycle::deprecate_warn("2.1.0", "ggseg3d(label=)", "ggseg3d(label_by=)")
    label_by <- label
  }
  if (lifecycle::is_present(text)) {
    lifecycle::deprecate_warn("2.1.0", "ggseg3d(text=)", "ggseg3d(text_by=)")
    text_by <- text
  }
  if (lifecycle::is_present(colour)) {
    lifecycle::deprecate_warn(
      "2.1.0", "ggseg3d(colour=)", "ggseg3d(colour_by=)"
    )
    colour_by <- colour
  }

  atlas <- if (is.character(atlas)) {
    lifecycle::deprecate_warn(
      "2.0.0",
      "ggseg3d(atlas = 'must be a ggseg_atlas object')",
      details = "Pass atlas objects directly, e.g. `atlas = dk()`."
    )
    tryCatch(match.fun(atlas)(), error = function(e) {
      cli::cli_abort("Could not find atlas {.val {atlas}}.")
    })
  } else {
    atlas
  }

  if (!is_unified_atlas(atlas)) {
    cli::cli_abort(c(
      "Atlas must be a {.cls ggseg_atlas} object with 3D data.",
      "i" = "Use atlases from {.pkg ggseg.formats}.",
      "i" = "Create atlases with {.fn ggseg.extra::create_cortical_atlas}."
    ))
  }

  prepared <- prepare_brain_meshes(
    atlas,
    .data = .data,
    label_by = label_by,
    text_by = text_by,
    colour_by = colour_by,
    palette = palette,
    na_colour = na_colour,
    na_alpha = na_alpha,
    ...
  )

  create_ggseg3d_widget(prepared$meshes, prepared$legend_data)
}


# prepare_brain_meshes S3 generic ----

#' Prepare brain meshes and legend data
#'
#' S3 generic that dispatches to atlas-type-specific preparation methods.
#' Builds mesh data structures and legend data from a `ggseg_atlas`.
#'
#' @param atlas A `ggseg_atlas` object
#' @param ... Type-specific arguments passed to methods
#'
#' @return List with `meshes` (list of mesh entries) and `legend_data`
#' @keywords internal
prepare_brain_meshes <- function(atlas, ...) {
  UseMethod("prepare_brain_meshes")
}

#' @export
#' @keywords internal
prepare_brain_meshes.default <- function(atlas, ...) {
  cls <- paste(class(atlas), collapse = "/") # nolint: object_usage_linter
  cli::cli_abort(c(
    "No method for atlas of class {.val {cls}}.",
    "i" = "Expected {.cls cortical_atlas}, {.cls subcortical_atlas},
    {.cls tract_atlas}, or {.cls cerebellar_atlas}."
  ))
}

#' @method prepare_brain_meshes cortical_atlas
#' @param .data Optional user data to merge
#' @param surface Surface type: `"inflated"` (default), `"semi-inflated"`,
#'   `"white"`, `"pial"`. Use `"LCBC"` as alias for `"inflated"`.
#' @param hemisphere Character vector of hemispheres: `"right"`, `"left"`.
#' @param label_by Column name for region hover labels
#' @param text_by Column name for extra hover text
#' @param colour_by Column name for colour values
#' @param palette Colour palette specification
#' @param na_colour Colour for NA values
#' @param na_alpha Transparency for NA regions
#' @param edge_by Column name for region boundary edge grouping
#' @param brain_meshes Optional user-supplied brain meshes
#' @export
#' @rdname prepare_brain_meshes
#' @keywords internal
prepare_brain_meshes.cortical_atlas <- function(
  atlas,
  .data = NULL,
  surface = "LCBC",
  hemisphere = c("right", "left"),
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  edge_by = NULL,
  brain_meshes = NULL,
  ...
) {
  surface <- if (surface == "LCBC") "inflated" else surface

  atlas_data <- prepare_atlas_data(atlas, .data)
  result <- apply_colours_and_legend(
    atlas_data,
    colour_by,
    palette,
    na_colour,
    label_by
  )
  meshes <- build_cortical_meshes(
    result$atlas_data,
    hemisphere,
    surface,
    na_colour,
    edge_by,
    brain_meshes,
    text_by = text_by,
    label_by = label_by
  )

  list(meshes = meshes, legend_data = result$legend_data)
}

#' @method prepare_brain_meshes subcortical_atlas
#' @export
#' @rdname prepare_brain_meshes
#' @keywords internal
prepare_brain_meshes.subcortical_atlas <- function(
  atlas,
  .data = NULL,
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  ...
) {
  atlas_data <- prepare_mesh_atlas_data(atlas, .data)
  result <- apply_colours_and_legend(
    atlas_data,
    colour_by,
    palette,
    na_colour,
    label_by
  )
  meshes <- build_subcortical_meshes(
    result$atlas_data, na_colour,
    text_by = text_by, label_by = label_by
  )

  list(meshes = meshes, legend_data = result$legend_data)
}

#' @method prepare_brain_meshes cerebellar_atlas
#' @param surface_opacity Numeric opacity for the cerebellar surface mesh.
#'   Defaults to `0.3` when deep nuclei are present and `1` otherwise.
#' @export
#' @rdname prepare_brain_meshes
#' @keywords internal
prepare_brain_meshes.cerebellar_atlas <- function(
  atlas,
  .data = NULL,
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  surface_opacity = NULL,
  ...
) {
  has_deep <- !is.null(atlas$data$meshes)
  surface_opacity <- surface_opacity %||% if (has_deep) 0.3 else 1

  atlas_data <- prepare_atlas_data(atlas, .data)
  result <- apply_colours_and_legend(
    atlas_data,
    colour_by,
    palette,
    na_colour,
    label_by
  )
  surface_meshes <- build_cerebellar_meshes(
    result$atlas_data, na_colour,
    text_by = text_by, label_by = label_by,
    opacity = surface_opacity
  )

  if (has_deep) {
    deep_data <- prepare_mesh_atlas_data(atlas, .data)
    deep_result <- apply_colours_and_legend(
      deep_data, colour_by, palette, na_colour, label_by
    )
    deep_meshes <- build_subcortical_meshes(
      deep_result$atlas_data, na_colour,
      text_by = text_by, label_by = label_by
    )
    all_meshes <- c(surface_meshes, deep_meshes)
    legend_data <- merge_legend_data(
      result$legend_data, deep_result$legend_data
    )
  } else {
    all_meshes <- surface_meshes
    legend_data <- result$legend_data
  }

  list(meshes = all_meshes, legend_data = legend_data)
}

#' @method prepare_brain_meshes tract_atlas
#' @param tract_color `"palette"` (default) or `"orientation"`
#'   (direction-based RGB colouring)
#' @param tube_radius Numeric tube radius (default 5 when `NULL`).
#' @param tube_segments Integer tube segment count (default 8 when `NULL`).
#' @export
#' @rdname prepare_brain_meshes
#' @keywords internal
prepare_brain_meshes.tract_atlas <- function(
  atlas,
  .data = NULL,
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  tract_color = c("palette", "orientation"),
  tube_radius = 2,
  tube_segments = 10,
  ...
) {
  tract_color <- match.arg(tract_color)
  color_by <- if (tract_color == "orientation") "orientation" else "colour"

  atlas_data <- prepare_mesh_atlas_data(atlas, .data)
  result <- apply_colours_and_legend(
    atlas_data,
    colour_by,
    palette,
    na_colour,
    label_by
  )
  atlas_data <- result$atlas_data
  atlas_centerlines <- build_centerline_data(atlas, tube_radius, tube_segments)
  meshes <- build_tract_meshes(
    atlas_data,
    na_colour,
    color_by,
    atlas_centerlines,
    text_by = text_by,
    label_by = label_by
  )

  list(meshes = meshes, legend_data = result$legend_data)
}


# Shared helpers ----

#' Apply colour palette and build legend data
#'
#' Shared pipeline step for all atlas types: applies colour palette to
#' atlas data and builds the legend data structure.
#'
#' @param atlas_data Prepared atlas data frame
#' @param colour_by Column name for colour values
#' @param palette Colour palette specification
#' @param na_colour Colour for NA values
#' @param label_by Column name for labels
#'
#' @return List with `atlas_data` and `legend_data`
#' @keywords internal
apply_colours_and_legend <- function(
  atlas_data,
  colour_by,
  palette,
  na_colour,
  label_by
) {
  colour_result <- apply_colour_palette(
    atlas_data,
    colour_by,
    palette,
    na_colour
  )
  atlas_data <- colour_result$data

  # nolint start: object_usage_linter
  legend_data <- build_legend_data(
    is_numeric = colour_result$is_numeric,
    data_min = colour_result$data_min,
    data_max = colour_result$data_max,
    palette = palette,
    pal_colours = colour_result$palette,
    colour_col = colour_by,
    label_col = label_by,
    fill_col = colour_result$fill,
    data = atlas_data
  )
  # nolint end

  list(atlas_data = atlas_data, legend_data = legend_data)
}


#' Build centerline data for tract atlases
#'
#' Extracts centerline data from a tract atlas and assembles the tube
#' generation parameters.
#'
#' @param atlas A `tract_atlas` object
#' @param tube_radius Optional radius override
#' @param tube_segments Optional segment count override
#'
#' @return List with centerlines, tube_radius, tube_segments, or NULL
#' @keywords internal
build_centerline_data <- function(
  atlas,
  tube_radius = NULL,
  tube_segments = NULL
) {
  if (is.null(atlas$data$centerlines)) {
    return(NULL)
  }

  list(
    centerlines = atlas$data$centerlines,
    tube_radius = tube_radius %||% 2,
    tube_segments = tube_segments %||% 10
  )
}

test_that("pan_camera works", {
  p <- ggseg3d() |>
    pan_camera("right lateral")

  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_equal(p$x$options$camera, "right lateral")

  p <- ggseg3d() |>
    pan_camera("right medial")

  expect_equal(p$x$options$camera, "right medial")

  p <- ggseg3d() |>
    pan_camera("left lateral")

  expect_equal(p$x$options$camera, "left lateral")

  p <- ggseg3d() |>
    pan_camera("left medial")

  expect_equal(p$x$options$camera, "left medial")

  p <- ggseg3d() |>
    pan_camera(camera = list(eye = list(x = -3, y = -4, z = -1)))

  expect_equal(
    p$x$options$camera,
    list(eye = list(x = -3, y = -4, z = -1))
  )
})

test_that("default camera is right lateral", {
  p <- ggseg3d()
  expect_equal(p$x$options$camera, "right lateral")
})

test_that("Check glassbrain", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain()

  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  n_meshes_base <- length(ggseg3d(atlas = aseg())$x$meshes)
  expect_true(length(p$x$meshes) > n_meshes_base)

  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = "left")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = "left", colour = "red")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  glassbrain_meshes <- p$x$meshes[vapply(
    p$x$meshes,
    function(m) {
      grepl("glass brain", m$name)
    },
    logical(1)
  )]
  expect_true(length(glassbrain_meshes) > 0)
  expect_equal(glassbrain_meshes[[1]]$colors[[1]], "#FF0000")
})

test_that("add_glassbrain with hex color", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = "left", colour = "#AABBCC")

  glassbrain_meshes <- p$x$meshes[vapply(
    p$x$meshes,
    function(m) {
      grepl("glass brain", m$name)
    },
    logical(1)
  )]

  expect_equal(glassbrain_meshes[[1]]$colors[[1]], "#AABBCC")
})

test_that("add_glassbrain with custom opacity", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = "left", opacity = 0.5)

  glassbrain_meshes <- p$x$meshes[vapply(
    p$x$meshes,
    function(m) {
      grepl("glass brain", m$name)
    },
    logical(1)
  )]

  expect_equal(glassbrain_meshes[[1]]$opacity, 0.5)
})

test_that("add_glassbrain errors on non-ggseg3d object", {
  expect_error(
    add_glassbrain(list()),
    "ggseg3d"
  )
})

test_that("add_glassbrain with both hemispheres", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = c("left", "right"))

  glassbrain_meshes <- p$x$meshes[vapply(
    p$x$meshes,
    function(m) {
      grepl("glass brain", m$name)
    },
    logical(1)
  )]

  expect_length(glassbrain_meshes, 2)
})

test_that("add_glassbrain with inflated surface", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = "left", surface = "inflated")

  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("add_glassbrain works with white surface", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = "left", surface = "white")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

#' Create ggseg3d htmlwidget
#'
#' Creates the final htmlwidget object with default options and sizing policy.
#'
#' @param meshes List of mesh data structures
#' @param legend_data Legend specification (or NULL)
#'
#' @return An htmlwidget object of class "ggseg3d"
#' @keywords internal
create_ggseg3d_widget <- function(meshes, legend_data) {
  options <- list(
    camera = "right lateral",
    showLegend = TRUE,
    backgroundColor = "#ffffff"
  )

  x <- list(
    meshes = meshes,
    options = options,
    colorbar = legend_data
  )

  htmlwidgets::createWidget(
    name = "ggseg3d",
    x = x,
    package = "ggseg3d",
    sizingPolicy = htmlwidgets::sizingPolicy(
      defaultWidth = 600,
      defaultHeight = 500,
      viewer.defaultHeight = 500,
      viewer.defaultWidth = 600,
      browser.defaultHeight = 500,
      browser.defaultWidth = 600,
      knitr.defaultWidth = 600,
      knitr.defaultHeight = 500,
      padding = 0,
      browser.fill = TRUE
    )
  )
}

test_that("create_ggseg3d_widget creates htmlwidget", {
  meshes <- list(
    list(
      name = "test",
      vertices = list(x = c(0, 1), y = c(0, 1), z = c(0, 1)),
      faces = list(i = 0L, j = 1L, k = 2L),
      colors = c("#FF0000", "#00FF00"),
      colorMode = "vertexcolor"
    )
  )

  widget <- create_ggseg3d_widget(meshes, NULL)

  expect_s3_class(widget, c("ggseg3d", "htmlwidget"))
  expect_equal(widget$x$options$camera, "right lateral")
  expect_true(widget$x$options$showLegend)
  expect_equal(widget$x$options$backgroundColor, "#ffffff")
})

test_that("create_ggseg3d_widget includes legend data", {
  meshes <- list()
  legend_data <- list(
    type = "continuous",
    title = "value",
    min = 0,
    max = 10
  )

  widget <- create_ggseg3d_widget(meshes, legend_data)

  expect_equal(widget$x$colorbar, legend_data)
})

test_that("create_ggseg3d_widget has correct sizing policy", {
  meshes <- list()
  widget <- create_ggseg3d_widget(meshes, NULL)

  expect_equal(widget$sizingPolicy$defaultWidth, 600)
  expect_equal(widget$sizingPolicy$defaultHeight, 500)
})

check_ggseg3d <- function(
  p,
  arg = rlang::caller_arg(p),
  call = rlang::caller_env()
) {
  if (!inherits(p, "ggseg3d")) {
    cli::cli_abort(
      "{.arg {arg}} must be a {.cls ggseg3d} widget, not {.obj_type_friendly {p}}.", # nolint: line_length_linter
      call = call
    )
  }
}


#' @importFrom dplyr left_join
merge_atlas_data <- function(.data, atlas_data) {
  cols <- names(atlas_data)[names(atlas_data) %in% names(.data)]

  if (length(cols) == 0) {
    cli::cli_abort(c(
      "No common columns between data and atlas.",
      "i" = "Atlas has: {.field {names(atlas_data)}}",
      "i" = "Data has: {.field {names(.data)}}"
    ))
  }

  merged <- dplyr::left_join(
    atlas_data,
    .data,
    by = cols,
    relationship = "many-to-many"
  )

  unmatched <- .data[!.data[[cols[1]]] %in% atlas_data[[cols[1]]], ]
  if (nrow(unmatched) > 0) {
    unmatched_vals <- unique(unmatched[[cols[1]]]) # nolint: object_usage_linter
    cli::cli_warn(c(
      "Some data rows did not match atlas regions.",
      "i" = "Unmatched: {.val {unmatched_vals}}",
      "i" = "Check for spelling mistakes in: {.field {cols[1]}}"
    ))
  }

  merged
}


col2hex <- function(colour) {
  col <- grDevices::col2rgb(colour)
  grDevices::rgb(
    red = col[1, ] / 255,
    green = col[2, ] / 255,
    blue = col[3, ] / 255
  )
}


make_mesh_entry <- function(
  name,
  vertices,
  faces,
  colors,
  color_mode = "vertexcolor",
  opacity = 1,
  hover_text = NULL,
  boundary_edges = NULL,
  edge_color = NULL,
  edge_width = NULL,
  vertex_labels = NULL,
  vertex_texts = NULL
) {
  entry <- list(
    name = name,
    vertices = list(
      x = unname(as.numeric(vertices$x)),
      y = unname(as.numeric(vertices$y)),
      z = unname(as.numeric(vertices$z))
    ),
    faces = list(
      i = unname(as.integer(faces$i - 1L)),
      j = unname(as.integer(faces$j - 1L)),
      k = unname(as.integer(faces$k - 1L))
    ),
    colors = unname(colors),
    colorMode = color_mode,
    opacity = opacity,
    hoverText = hover_text
  )

  if (!is.null(boundary_edges)) {
    entry$boundaryEdges <- boundary_edges
  }

  if (!is.null(edge_color)) {
    entry$edgeColor <- edge_color
    entry$edgeWidth <- edge_width %||% 1
  }

  if (!is.null(vertex_labels)) {
    entry$vertexLabels <- unname(vertex_labels)
  }

  if (!is.null(vertex_texts)) {
    entry$vertexTexts <- unname(vertex_texts)
  }

  entry
}


find_boundary_edges <- function(faces, vertex_colors) {
  i <- faces$i
  j <- faces$j
  k <- faces$k

  all_v1 <- c(i, j, k)
  all_v2 <- c(j, k, i)

  color1 <- vertex_colors[all_v1]
  color2 <- vertex_colors[all_v2]
  is_boundary <- color1 != color2
  is_boundary[is.na(is_boundary)] <- FALSE

  boundary_v1 <- all_v1[is_boundary]
  boundary_v2 <- all_v2[is_boundary]

  if (length(boundary_v1) == 0) {
    return(list())
  }

  min_v <- pmin(boundary_v1, boundary_v2)
  max_v <- pmax(boundary_v1, boundary_v2)

  n_vertices <- max(c(boundary_v1, boundary_v2))
  edge_keys <- min_v + (max_v - 1L) * n_vertices

  unique_idx <- !duplicated(edge_keys)
  unique_v1 <- boundary_v1[unique_idx]
  unique_v2 <- boundary_v2[unique_idx]

  mapply(
    function(a, b) c(a - 1L, b - 1L),
    unique_v1,
    unique_v2,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
}


get_palette <- function(palette) {
  if (is.null(palette)) {
    palette <- c("#440154", "#21918c", "#fde725")
  }

  if (!is.null(names(palette))) {
    pal_colours <- names(palette)
    pal_values <- unname(palette)
    pal_norm <- range_norm(pal_values)
  } else {
    pal_colours <- palette
    pal_norm <- seq(0, 1, length.out = length(pal_colours))
    pal_values <- seq(0, 1, length.out = length(pal_colours))
  }

  pal_colours <- if (length(palette) == 1) {
    data.frame(
      values = c(pal_values, pal_values + 1),
      norm = c(0, 1),
      orig = c(pal_colours, pal_colours),
      stringsAsFactors = FALSE
    )
  } else {
    data.frame(
      values = pal_values,
      norm = pal_norm,
      orig = pal_colours,
      stringsAsFactors = FALSE
    )
  }

  pal_colours$hex <- gradient_n_pal(
    colours = pal_colours$orig,
    values = pal_colours$values,
    space = "Lab"
  )(pal_colours$values)

  pal_colours
}

range_norm <- function(x) {
  (x - min(x)) / (max(x) - min(x))
}

test_that("get_palette works", {
  expect_equal(
    get_palette("blue"),
    data.frame(
      values = c(0, 1),
      norm = c(0, 1),
      orig = "blue",
      hex = "#0000FF",
      stringsAsFactors = FALSE
    )
  )

  expect_equal(
    get_palette(c("blue" = 1)),
    data.frame(
      values = c(1, 2),
      norm = c(0, 1),
      orig = "blue",
      hex = "#0000FF",
      stringsAsFactors = FALSE
    )
  )

  expect_equal(
    get_palette(NULL),
    structure(
      list(
        values = c(0, 0.5, 1),
        norm = c(0, 0.5, 1),
        orig = c("#440154", "#21918c", "#fde725"),
        hex = c("#440154", "#21918C", "#FDE725")
      ),
      row.names = c(NA, -3L),
      class = "data.frame"
    )
  )

  expect_equal(
    get_palette(c("firebrick", "white", "goldenrod")),
    structure(
      list(
        values = c(0, 0.5, 1),
        norm = c(0, 0.5, 1),
        orig = c("firebrick", "white", "goldenrod"),
        hex = c("#B22222", "#FFFFFF", "#DAA520")
      ),
      row.names = c(NA, -3L),
      class = "data.frame"
    )
  )

  expect_equal(
    get_palette(c("#ffffff", "#d3d3d3", "#32f303")),
    structure(
      list(
        values = c(0, 0.5, 1),
        norm = c(0, 0.5, 1),
        orig = c("#ffffff", "#d3d3d3", "#32f303"),
        hex = c("#FFFFFF", "#D3D3D3", "#32F303")
      ),
      row.names = c(NA, -3L),
      class = "data.frame"
    )
  )

  expect_equal(
    get_palette(c("#ffffff" = 0, "#d3d3d3" = 1, "#32f303" = 2)),
    structure(
      list(
        values = c(0, 1, 2),
        norm = c(0, 0.5, 1),
        orig = c("#ffffff", "#d3d3d3", "#32f303"),
        hex = c("#FFFFFF", "#D3D3D3", "#32F303")
      ),
      row.names = c(NA, -3L),
      class = "data.frame"
    )
  )
})

test_that("col2hex works", {
  expect_equal(col2hex("red"), "#FF0000")
  expect_equal(col2hex("green"), "#00FF00")
  expect_equal(col2hex("blue"), "#0000FF")
  expect_equal(col2hex("white"), "#FFFFFF")
  expect_equal(col2hex("black"), "#000000")
})

test_that("get_palette with named numeric palette", {
  pal <- get_palette(c("blue" = 0, "white" = 50, "red" = 100))

  expect_equal(pal$values, c(0, 50, 100))
  expect_equal(pal$orig, c("blue", "white", "red"))
  expect_equal(pal$norm, c(0, 0.5, 1))
})

test_that("get_palette handles two colors", {
  pal <- get_palette(c("blue", "red"))

  expect_equal(nrow(pal), 2)
  expect_equal(pal$orig, c("blue", "red"))
})

test_that("range_norm normalizes correctly", {
  expect_equal(range_norm(c(0, 50, 100)), c(0, 0.5, 1))
  expect_equal(range_norm(c(10, 20)), c(0, 1))
  expect_equal(range_norm(c(-10, 0, 10)), c(0, 0.5, 1))
})

test_that("merge_atlas_data joins data correctly", {
  atlas_data <- data.frame(
    label = c("a", "b", "c"),
    region = c("region a", "region b", "region c"),
    hemi = c("left", "right", "left"),
    stringsAsFactors = FALSE
  )

  user_data <- data.frame(
    label = c("a", "b"),
    value = c(10, 20),
    stringsAsFactors = FALSE
  )

  result <- merge_atlas_data(user_data, atlas_data)

  expect_true("value" %in% names(result))
  expect_equal(result$value[result$label == "a"], 10)
  expect_true(is.na(result$value[result$label == "c"]))
})

test_that("merge_atlas_data errors on no common columns", {
  atlas_data <- data.frame(
    label = c("a"),
    stringsAsFactors = FALSE
  )

  user_data <- data.frame(
    unrelated = c("x"),
    stringsAsFactors = FALSE
  )

  expect_error(
    merge_atlas_data(user_data, atlas_data),
    "No common columns"
  )
})

test_that("merge_atlas_data warns on unmatched rows", {
  atlas_data <- data.frame(
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )

  user_data <- data.frame(
    label = c("a", "unknown_region"),
    value = c(10, 20),
    stringsAsFactors = FALSE
  )

  expect_warning(
    result <- merge_atlas_data(user_data, atlas_data),
    "did not match"
  )
})

test_that("merge_atlas_data joins on multiple common columns", {
  atlas_data <- data.frame(
    label = c("a", "b", "a"),
    hemi = c("left", "left", "right"),
    region = c("ra", "rb", "ra"),
    stringsAsFactors = FALSE
  )

  user_data <- data.frame(
    label = c("a", "a"),
    hemi = c("left", "right"),
    value = c(100, 200),
    stringsAsFactors = FALSE
  )

  result <- merge_atlas_data(user_data, atlas_data)

  expect_equal(result$value[result$label == "a" & result$hemi == "left"], 100)
  expect_equal(result$value[result$label == "a" & result$hemi == "right"], 200)
})

test_that("check_ggseg3d passes for valid widget", {
  p <- ggseg3d()
  expect_silent(check_ggseg3d(p))
})

test_that("check_ggseg3d errors for non-widget", {
  expect_error(
    check_ggseg3d(list()),
    "ggseg3d"
  )

  expect_error(
    check_ggseg3d(data.frame()),
    "ggseg3d"
  )

  expect_error(
    check_ggseg3d("not a widget"),
    "ggseg3d"
  )
})

test_that("check_ggseg3d reports correct argument name", {
  my_var <- list()
  expect_error(
    check_ggseg3d(my_var),
    "my_var"
  )
})

test_that("range_norm normalizes to 0-1 range", {
  result <- range_norm(c(0, 5, 10))
  expect_equal(result, c(0, 0.5, 1))

  result <- range_norm(c(10, 20, 30, 40))
  expect_equal(result, c(0, 1 / 3, 2 / 3, 1))
})

test_that("range_norm handles single value", {
  result <- range_norm(c(5, 5))
  expect_true(is.nan(result[1]) || result[1] == 0)
})

test_that("find_boundary_edges finds edges between different colors", {
  faces <- data.frame(
    i = c(1, 2),
    j = c(2, 3),
    k = c(3, 4)
  )
  vertex_colors <- c("#FF0000", "#FF0000", "#00FF00", "#00FF00")

  edges <- find_boundary_edges(faces, vertex_colors)

  expect_true(length(edges) > 0)
  edge_vertices <- unlist(edges)
  expect_true(1 %in% edge_vertices || 2 %in% edge_vertices)
})

test_that("find_boundary_edges returns empty for uniform colors", {
  faces <- data.frame(
    i = c(1, 2),
    j = c(2, 3),
    k = c(3, 4)
  )
  vertex_colors <- rep("#FF0000", 4)

  edges <- find_boundary_edges(faces, vertex_colors)

  expect_equal(length(edges), 0)
})

test_that("find_boundary_edges handles single face", {
  faces <- data.frame(i = 1, j = 2, k = 3)
  vertex_colors <- c("#FF0000", "#00FF00", "#0000FF")

  edges <- find_boundary_edges(faces, vertex_colors)

  expect_equal(length(edges), 3)
})

test_that("find_boundary_edges deduplicates shared edges", {
  faces <- data.frame(
    i = c(1, 2),
    j = c(2, 3),
    k = c(3, 3)
  )
  vertex_colors <- c("#FF0000", "#00FF00", "#00FF00", "#00FF00")

  edges <- find_boundary_edges(faces, vertex_colors)

  edge_keys <- vapply(
    edges,
    function(e) paste(sort(e), collapse = "-"),
    character(1)
  )
  expect_equal(length(edge_keys), length(unique(edge_keys)))
})

test_that("find_boundary_edges skips NA groups without NAs in output", {
  faces <- data.frame(
    i = c(1, 2, 3),
    j = c(2, 3, 4),
    k = c(3, 4, 5)
  )
  vertex_colors <- c(NA, NA, "A", "B", NA)

  edges <- find_boundary_edges(faces, vertex_colors)

  expect_true(length(edges) > 0)
  edge_verts <- unlist(edges)
  expect_false(any(is.na(edge_verts)))
})

test_that("find_boundary_edges returns empty for all-NA groups", {
  faces <- data.frame(i = 1, j = 2, k = 3)
  vertex_colors <- c(NA, NA, NA)

  edges <- find_boundary_edges(faces, vertex_colors)

  expect_equal(length(edges), 0)
})

# Stable textual summary of a ggseg3d widget's meshes for use with
# expect_snapshot(). Captures enough shape/color information to catch
# coordinate regressions (e.g. overlapping hemispheres, wrong axis),
# missing meshes, and colour-mapping breakage, without relying on
# WebGL rendering (which varies across machines).

widget_summary <- function(p, coord_digits = 1) {
  meshes <- p$x$meshes
  if (length(meshes) == 0) return(data.frame())

  rows <- lapply(meshes, function(m) {
    v <- m$vertices
    f <- m$faces
    x_range <- round(range(v$x), coord_digits)
    y_range <- round(range(v$y), coord_digits)
    z_range <- round(range(v$z), coord_digits)

    colors <- m$colors %||% character(0)
    colors <- colors[!is.na(colors)]

    data.frame(
      name = m$name %||% NA_character_,
      n_vertices = length(v$x),
      n_faces = if (is.null(f)) NA_integer_ else length(f$i),
      x_min = x_range[1], x_max = x_range[2],
      y_min = y_range[1], y_max = y_range[2],
      z_min = z_range[1], z_max = z_range[2],
      color_mode = m$colorMode %||% NA_character_,
      n_colors = length(unique(colors)),
      color_digest = substr(rlang::hash(sort(unique(colors))), 1, 8),
      opacity = m$opacity %||% 1,
      is_flatmap = isTRUE(m$isFlatmap),
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


# Build a minimal cerebellar_atlas for testing so we do not depend on
# external atlas packages.
make_test_cerebellar_atlas <- function() {
  vertices_data <- data.frame(
    label = c("left_I-IV", "right_I-IV"),
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(0L:99L, 100L:199L)

  structure(
    list(
      atlas = "test_cerebellar",
      type = "cerebellar",
      core = data.frame(
        label = c("left_I-IV", "right_I-IV"),
        region = c("I-IV", "I-IV"),
        hemi = c("left", "right"),
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data),
        class = c("ggseg_data_cerebellar", "ggseg_atlas_data")
      ),
      palette = c("left_I-IV" = "#FF0000", "right_I-IV" = "#00FF00")
    ),
    class = c("cerebellar_atlas", "ggseg_atlas", "list")
  )
}

# Data-level visual regression tests.
#
# These snapshot a textual summary of every mesh produced by ggseg3d so we
# catch regressions in coordinates (hemisphere overlap, axis mix-ups),
# mesh counts, and colour mapping without relying on WebGL rendering.
# See helper-visual.R for the summary helper.

test_that("dk cortical atlas produces stable mesh layout", {
  p <- ggseg3d(hemisphere = c("left", "right"))
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

test_that("dk single hemisphere has medial edge at midline", {
  lh <- ggseg3d(hemisphere = "left")
  rh <- ggseg3d(hemisphere = "right")
  expect_snapshot(print(widget_summary(lh), row.names = FALSE))
  expect_snapshot(print(widget_summary(rh), row.names = FALSE))
})

test_that("dk pial surface produces stable mesh layout", {
  skip_if_not_installed("ggseg.meshes")
  p <- ggseg3d(hemisphere = c("left", "right"), surface = "pial")
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

test_that("aseg subcortical atlas produces stable mesh layout", {
  p <- ggseg3d(atlas = aseg())
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

test_that("cerebellar atlas produces stable mesh layout", {
  p <- ggseg3d(atlas = make_test_cerebellar_atlas())
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

test_that("cortical + glassbrain composes as expected", {
  p <- ggseg3d(hemisphere = c("left", "right")) |>
    add_glassbrain(hemisphere = c("left", "right"), opacity = 0.2)
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

test_that("aseg + glassbrain composes as expected", {
  p <- ggseg3d(atlas = aseg()) |>
    add_glassbrain(hemisphere = c("left", "right"), opacity = 0.15)
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

test_that("cerebellar + glassbrain composes as expected", {
  p <- ggseg3d(atlas = make_test_cerebellar_atlas()) |>
    add_glassbrain(hemisphere = c("left", "right"), opacity = 0.15)
  expect_snapshot(print(widget_summary(p), row.names = FALSE))
})

#' Resolve brain surface mesh
#'
#' Resolves and prepares a brain surface mesh for rendering. Delegates to
#' [ggseg.formats::get_brain_mesh()] for inflated surfaces and to
#' [ggseg.meshes::get_cortical_mesh()] for pial, white, semi-inflated, and
#' other surfaces. Returned meshes use the shared anatomical axis convention
#' (`x` = left-right, `y` = anterior-posterior, `z` = superior-inferior) with
#' `lh` positioned at `x <= 0` and `rh` at `x >= 0`, medial edges meeting at
#' the midline (`x = 0`).
#'
#' @param hemisphere `"lh"` or `"rh"`
#' @param surface Surface type: `"inflated"`, `"semi-inflated"`, `"white"`,
#'   `"pial"`, `"sphere"`, `"smoothwm"`, `"orig"`
#' @param brain_meshes Optional user-supplied mesh data. Passed through to
#'   [ggseg.formats::get_brain_mesh()] for format details.
#'
#' @return list with vertices (data.frame with x, y, z) and faces
#'   (data.frame with i, j, k), or NULL if mesh not found
#'
#' @examples
#' mesh <- resolve_brain_mesh("lh", "inflated")
#' str(mesh, max.level = 1)
#'
#' @export
resolve_brain_mesh <- function(
  hemisphere = c("lh", "rh"),
  surface = c("inflated", "semi-inflated", "white", "pial",
              "sphere", "smoothwm", "orig"),
  brain_meshes = NULL
) {
  hemisphere <- match.arg(hemisphere)
  surface <- match.arg(surface)

  from_ggseg_meshes <- is.null(brain_meshes) && surface != "inflated"

  if (!from_ggseg_meshes) {
    mesh <- ggseg.formats::get_brain_mesh(hemisphere, surface, brain_meshes)
  } else {
    check_ggseg_meshes(surface)
    mesh <- ggseg.meshes::get_cortical_mesh(hemisphere, surface)
  }

  if (is.null(mesh)) return(NULL)

  if (min(mesh$faces$i) == 0) {
    mesh$faces$i <- mesh$faces$i + 1L
    mesh$faces$j <- mesh$faces$j + 1L
    mesh$faces$k <- mesh$faces$k + 1L
  }

  normalize_cortical_mesh(mesh, hemisphere, from_ggseg_meshes)
}


check_ggseg_meshes <- function(surface) {
  if (!requireNamespace("ggseg.meshes", quietly = TRUE)) {
    cli::cli_abort(c(
      "{.pkg ggseg.meshes} is required for {.val {surface}}.",
      "i" = "Install with: {.run install.packages('ggseg.meshes')}"
    ))
  }
}


# ggseg.meshes surfaces are stored rotated 90 deg CCW relative to FreeSurfer
# native axes (see ggseg.meshes/data-raw/make_cortical_meshes.R): stored
# `x` = AP, stored `y` = -LR, stored `z` = SI. ggseg.formats inflated and
# subcortical/cerebellar meshes already use native axes (x = LR, y = AP,
# z = SI). After axis alignment we shift each hemisphere so medial edges sit
# at x = 0 — LH occupies x <= 0, RH occupies x >= 0 — so both hemispheres
# display side-by-side without overlap regardless of surface choice.
normalize_cortical_mesh <- function(mesh, hemisphere, rotate_axes) {
  if (rotate_axes) {
    v <- mesh$vertices
    mesh$vertices$x <- -v$y
    mesh$vertices$y <- v$x
  }

  mesh$vertices$x <- if (hemisphere == "lh") {
    mesh$vertices$x - max(mesh$vertices$x)
  } else {
    mesh$vertices$x - min(mesh$vertices$x)
  }

  mesh
}


#' Map atlas vertex indices to mesh colors
#'
#' Given a ggseg_atlas with vertices column and a brain mesh, creates a color
#' vector for each mesh vertex based on which region it belongs to.
#'
#' @param atlas_data Data frame with region, colour, and vertices columns
#' @param n_vertices Number of vertices in the mesh
#' @param na_colour Color for vertices not in any region
#'
#' @return Character vector of colors, one per mesh vertex
#' @keywords internal
vertices_to_colors <- function(
  atlas_data,
  n_vertices,
  na_colour = "#CCCCCC"
) {
  vertex_colors <- rep(na_colour, n_vertices)

  for (i in seq_len(nrow(atlas_data))) {
    region_vertices <- atlas_data$vertices[[i]]
    region_colour <- atlas_data$colour[i]

    if (length(region_vertices) > 0 && !is.na(region_colour)) {
      idx <- region_vertices + 1L
      idx <- idx[idx >= 1 & idx <= n_vertices]
      vertex_colors[idx] <- region_colour
    }
  }

  vertex_colors
}


#' Map atlas vertex indices to region labels
#'
#' Given a ggseg_atlas with vertices column and a brain mesh, creates a label
#' vector for each mesh vertex based on which region it belongs to.
#'
#' @param atlas_data Data frame with region and vertices columns
#' @param n_vertices Number of vertices in the mesh
#' @param na_label Label for vertices not in any region
#'
#' @return Character vector of labels, one per mesh vertex
#' @keywords internal
vertices_to_labels <- function(
  atlas_data,
  n_vertices,
  na_label = NA_character_
) {
  vertex_labels <- rep(na_label, n_vertices)

  for (i in seq_len(nrow(atlas_data))) {
    region_vertices <- atlas_data$vertices[[i]]
    region_label <- atlas_data$region[i]

    if (length(region_vertices) > 0 && !is.na(region_label)) {
      idx <- region_vertices + 1L
      idx <- idx[idx >= 1 & idx <= n_vertices]
      vertex_labels[idx] <- region_label
    }
  }

  vertex_labels
}


#' Map vertices to text values for hover display
#'
#' Assigns text values to mesh vertices based on a column in atlas data.
#' Used for per-vertex hover text in the Three.js tooltip.
#'
#' @param atlas_data Data frame with vertices list column
#' @param n_vertices Number of vertices in the mesh
#' @param text_col Name of the column containing text values
#'
#' @return Character vector of text values, one per mesh vertex
#' @keywords internal
vertices_to_text <- function(atlas_data, n_vertices, text_col) {
  vertex_text <- rep(NA_character_, n_vertices)

  if (!text_col %in% names(atlas_data)) return(vertex_text)

  for (i in seq_len(nrow(atlas_data))) {
    region_vertices <- atlas_data$vertices[[i]]
    val <- as.character(atlas_data[[text_col]][i])

    if (length(region_vertices) > 0 && !is.na(val)) {
      idx <- region_vertices + 1L
      idx <- idx[idx >= 1 & idx <= n_vertices]
      vertex_text[idx] <- paste0(text_col, ": ", val)
    }
  }

  vertex_text
}


#' Map vertices to group values for edge detection
#'
#' Assigns group values to mesh vertices based on a column in atlas data.
#' Used for computing boundary edges between different groups rather than
#' between different colors.
#'
#' @param atlas_data Data frame with vertices list column and grouping column
#' @param n_vertices Total number of vertices in the mesh
#' @param group_col Name of the column containing group values
#' @param na_group Value for vertices not in any region
#'
#' @return Character vector of group values, one per mesh vertex
#' @noRd
#' @keywords internal
vertices_to_groups <- function(
  atlas_data,
  n_vertices,
  group_col,
  na_group = NA_character_
) {
  vertex_groups <- rep(na_group, n_vertices)

  if (!group_col %in% names(atlas_data)) {
    cli::cli_abort(
      "Column {.val {group_col}} not found in atlas data."
    )
  }

  for (i in seq_len(nrow(atlas_data))) {
    region_vertices <- atlas_data$vertices[[i]]
    region_group <- as.character(atlas_data[[group_col]][i])

    if (length(region_vertices) > 0 && !is.na(region_group)) {
      idx <- region_vertices + 1L
      idx <- idx[idx >= 1 & idx <= n_vertices]
      vertex_groups[idx] <- region_group
    }
  }

  vertex_groups
}


#' Build mesh list for cortical atlases
#'
#' Creates mesh data structures for cortical ggseg_atlas objects using
#' shared brain meshes with vertex-based colouring.
#'
#' @param atlas_data Prepared atlas data frame
#' @param hemisphere Hemispheres to include
#' @param surface Surface type
#' @param na_colour Colour for NA values
#' @param edge_by Column for edge grouping (or NULL)
#' @param brain_meshes Optional user-supplied brain meshes
#'
#' @return List of mesh data structures
#' @importFrom rlang .data
#' @keywords internal
build_cortical_meshes <- function(
  atlas_data,
  hemisphere,
  surface,
  na_colour,
  edge_by,
  brain_meshes = NULL,
  text_by = NULL,
  label_by = "region"
) {
  hemi_map <- c("right" = "rh", "left" = "lh")

  meshes <- lapply(hemisphere, function(current_hemi) {
    hemi_short <- hemi_map[current_hemi]
    mesh <- resolve_brain_mesh(
      hemisphere = hemi_short,
      surface = surface,
      brain_meshes = brain_meshes
    )

    if (is.null(mesh)) {
      cli::cli_warn(
        "Brain mesh not found for {.val {hemi_short}} {.val {surface}}."
      )
      return(NULL)
    }

    hemi_data <- dplyr::filter(atlas_data, .data$hemi == current_hemi)
    if (nrow(hemi_data) == 0) {
      return(NULL)
    }

    n_vertices <- nrow(mesh$vertices)
    vertex_colors <- vertices_to_colors(hemi_data, n_vertices, na_colour)
    vertex_labels <- vertices_to_labels(hemi_data, n_vertices, na_label = "")

    vertex_texts <- if (!is.null(text_by)) {
      vertices_to_text(hemi_data, n_vertices, text_by)
    }

    if (!is.null(edge_by)) {
      edge_groups <- vertices_to_groups(hemi_data, n_vertices, edge_by)
      boundary <- find_boundary_edges(mesh$faces, edge_groups)
    } else {
      boundary <- find_boundary_edges(mesh$faces, vertex_colors)
    }

    edge_color <- if (!is.null(edge_by)) "#000000" else NULL

    make_mesh_entry(
      name = paste(current_hemi, surface),
      vertices = mesh$vertices,
      faces = mesh$faces,
      colors = vertex_colors,
      color_mode = "vertexcolor",
      boundary_edges = boundary,
      edge_color = edge_color,
      vertex_labels = vertex_labels,
      vertex_texts = vertex_texts
    )
  })

  Filter(Negate(is.null), meshes)
}


#' Build mesh list for cerebellar atlases
#'
#' Creates mesh data structures for cerebellar ggseg_atlas objects using
#' the shared SUIT cerebellar surface with vertex-based colouring.
#'
#' @param atlas_data Prepared atlas data frame with vertices column
#' @param na_colour Colour for NA values
#' @param text_by Column for hover text (or NULL)
#' @param label_by Column for vertex labels
#' @param opacity Numeric opacity for the mesh (0 = transparent, 1 = opaque)
#'
#' @return List of mesh data structures
#' @keywords internal
build_cerebellar_meshes <- function(
  atlas_data,
  na_colour,
  text_by = NULL,
  label_by = "region",
  opacity = 1
) {
  mesh <- ggseg.formats::get_cerebellar_mesh()

  if (is.null(mesh)) {
    cli::cli_abort("SUIT cerebellar mesh not available.")
  }

  if (min(mesh$faces$i) == 0) {
    mesh$faces$i <- mesh$faces$i + 1L
    mesh$faces$j <- mesh$faces$j + 1L
    mesh$faces$k <- mesh$faces$k + 1L
  }

  n_vertices <- nrow(mesh$vertices)
  vertex_colors <- vertices_to_colors(atlas_data, n_vertices, na_colour)
  vertex_labels <- vertices_to_labels(
    atlas_data, n_vertices, na_label = ""
  )

  vertex_texts <- if (!is.null(text_by)) {
    vertices_to_text(atlas_data, n_vertices, text_by)
  }

  boundary <- find_boundary_edges(mesh$faces, vertex_colors)

  entry <- make_mesh_entry(
    name = "cerebellum",
    vertices = mesh$vertices,
    faces = mesh$faces,
    colors = vertex_colors,
    color_mode = "vertexcolor",
    opacity = opacity,
    boundary_edges = boundary,
    vertex_labels = vertex_labels,
    vertex_texts = vertex_texts
  )

  if (is_flat_mesh(mesh$vertices)) {
    entry$isFlatmap <- TRUE
  }

  list(entry)
}


# A flatmap mesh has all vertices in a single plane (negligible extent on one
# axis). Used to detect cerebellar flatmaps that cannot be combined with
# other 3D meshes.
is_flat_mesh <- function(vertices, tol = 1) {
  any(
    c(
      diff(range(vertices$x)),
      diff(range(vertices$y)),
      diff(range(vertices$z))
    ) < tol
  )
}


#' Build mesh list for subcortical atlases
#'
#' Creates mesh data structures for subcortical atlases with per-region mesh
#' data using face-based colouring (each structure is a separate mesh).
#'
#' @param atlas_data Prepared atlas data frame with label, colour, and mesh
#'   columns
#' @param na_colour Colour for NA values
#'
#' @return List of mesh data structures
#' @keywords internal
build_subcortical_meshes <- function(
  atlas_data, na_colour,
  text_by = NULL, label_by = "region"
) {
  meshes <- lapply(seq_len(nrow(atlas_data)), function(i) {
    mesh_data <- atlas_data$mesh[[i]]
    if (is.null(mesh_data)) return(NULL)

    colour <- atlas_data$colour[i]
    if (is.na(colour)) colour <- na_colour
    colour <- unname(ifelse(grepl("^#", colour), colour, col2hex(colour)))

    region_name <- if (label_by %in% names(atlas_data)) {
      atlas_data[[label_by]][i]
    } else {
      atlas_data$label[i]
    }

    hover <- NULL
    if (!is.null(text_by) && text_by %in% names(atlas_data)) {
      val <- atlas_data[[text_by]][i]
      if (!is.na(val)) hover <- paste0(text_by, ": ", val)
    }

    make_mesh_entry(
      name = region_name,
      vertices = mesh_data$vertices,
      faces = mesh_data$faces,
      colors = rep(colour, nrow(mesh_data$faces)),
      color_mode = "facecolor",
      hover_text = hover
    )
  })

  Filter(Negate(is.null), meshes)
}


#' Build mesh list for tract atlases
#'
#' Creates mesh data structures for tract atlases with per-vertex colouring.
#' Supports palette colours (uniform per tract) or orientation-based RGB
#' colours computed from centerline tangent vectors.
#'
#' @param atlas_data Prepared atlas data frame with label, colour, and mesh
#'   columns
#' @param na_colour Colour for NA values
#' @param color_by How to colour tracts: "colour" (use colour column),
#'   "orientation" (direction-based RGB from tangents)
#'
#' @return List of mesh data structures
#' @keywords internal
build_tract_meshes <- function(
  atlas_data,
  na_colour,
  color_by = "colour",
  atlas_centerlines = NULL,
  text_by = NULL,
  label_by = "region"
) {
  has_centerlines <- !is.null(atlas_centerlines) &&
    !is.null(atlas_centerlines$centerlines)
  has_legacy_meshes <- "mesh" %in% names(atlas_data)

  if (!has_centerlines && !has_legacy_meshes) {
    cli::cli_warn("No centerlines or meshes found for tract atlas")
    return(list())
  }

  meshes <- lapply(seq_len(nrow(atlas_data)), function(i) {
    label <- atlas_data$label[i]

    if (has_centerlines) {
      cl_idx <- which(atlas_centerlines$centerlines$label == label)
      if (length(cl_idx) == 0) return(NULL)

      centerline <- atlas_centerlines$centerlines$points[[cl_idx]]
      tangents <- atlas_centerlines$centerlines$tangents[[cl_idx]]

      mesh_data <- generate_tube_mesh(
        centerline = centerline,
        radius = atlas_centerlines$tube_radius,
        segments = atlas_centerlines$tube_segments
      )
      mesh_data$metadata$tangents <- tangents
    } else {
      mesh_data <- atlas_data$mesh[[i]]
      if (is.null(mesh_data)) return(NULL)
    }

    n_vertices <- nrow(mesh_data$vertices)

    if (color_by == "orientation" && !is.null(mesh_data$metadata$tangents)) {
      vertex_colors <- tangents_to_colors(mesh_data)
    } else {
      colour <- atlas_data$colour[i]
      if (is.na(colour)) colour <- na_colour
      colour <- unname(ifelse(grepl("^#", colour), colour, col2hex(colour)))
      vertex_colors <- rep(colour, n_vertices)
    }

    tract_name <- if (label_by %in% names(atlas_data)) {
      atlas_data[[label_by]][i]
    } else {
      label
    }

    hover <- NULL
    if (!is.null(text_by) && text_by %in% names(atlas_data)) {
      val <- atlas_data[[text_by]][i]
      if (!is.na(val)) hover <- paste0(text_by, ": ", val)
    }

    make_mesh_entry(
      name = tract_name,
      vertices = mesh_data$vertices,
      faces = mesh_data$faces,
      colors = vertex_colors,
      color_mode = "vertexcolor",
      hover_text = hover
    )
  })

  Filter(Negate(is.null), meshes)
}


#' Convert tangent vectors to orientation RGB colours
#'
#' Computes direction-based RGB colours from centerline tangent vectors.
#' Standard tractography colouring: R = left-right (x),
#' G = anterior-posterior (y), B = superior-inferior (z).
#'
#' @param mesh_data Mesh data with vertices data.frame and metadata list
#' @return Character vector of hex colours (one per mesh vertex)
#' @keywords internal
tangents_to_colors <- function(mesh_data) {
  metadata <- mesh_data$metadata
  n_centerline <- metadata$n_centerline_points
  tangents <- metadata$tangents

  n_vertices <- nrow(mesh_data$vertices)
  n_segments <- as.integer(n_vertices / n_centerline)

  abs_tangents <- abs(tangents)
  row_max <- apply(abs_tangents, 1, max)
  row_max[row_max == 0] <- 1
  normalized <- abs_tangents / row_max

  centerline_colors <- grDevices::rgb(
    normalized[, 1],
    normalized[, 2],
    normalized[, 3]
  )

  rep(centerline_colors, each = n_segments)
}


# Tube mesh generation ----

#' Generate tube mesh from centerline
#'
#' Creates a 3D tube mesh around a centerline path using parallel transport
#' frames for smooth geometry without twisting artifacts.
#'
#' @param centerline Matrix with N rows and 3 columns (x, y, z coordinates)
#' @param radius Tube radius. Either a single value or vector of length N.
#' @param segments Number of segments around tube circumference.
#' @return List with vertices (data.frame), faces (data.frame), and metadata
#' @keywords internal
generate_tube_mesh <- function(centerline, radius = 0.5, segments = 8) {
  if (!is.matrix(centerline) || nrow(centerline) < 2) {
    cli::cli_abort("centerline must be a matrix with at least 2 rows")
  }

  n_points <- nrow(centerline)
  frames <- compute_parallel_transp_fr(centerline)

  if (length(radius) == 1) {
    radius <- rep(radius, n_points)
  }

  n_vertices <- n_points * segments
  n_faces <- (n_points - 1) * segments * 2

  vertices <- matrix(0, nrow = n_vertices, ncol = 3)
  faces <- matrix(0L, nrow = n_faces, ncol = 3)

  angles <- seq(0, 2 * pi, length.out = segments + 1)[1:segments]

  for (i in seq_len(n_points)) {
    center <- centerline[i, ]
    normal <- frames$normals[i, ]
    binormal <- frames$binormals[i, ]
    r <- radius[i]

    for (j in seq_len(segments)) {
      angle <- angles[j]
      offset <- r * (cos(angle) * normal + sin(angle) * binormal)
      vertex_idx <- (i - 1) * segments + j
      vertices[vertex_idx, ] <- center + offset
    }
  }

  face_idx <- 1
  for (i in seq_len(n_points - 1)) {
    for (j in seq_len(segments)) {
      j_next <- if (j == segments) 1L else j + 1L

      v1 <- (i - 1L) * segments + j
      v2 <- (i - 1L) * segments + j_next
      v3 <- i * segments + j
      v4 <- i * segments + j_next

      faces[face_idx, ] <- c(v1, v2, v3)
      faces[face_idx + 1L, ] <- c(v2, v4, v3)
      face_idx <- face_idx + 2L
    }
  }

  list(
    vertices = data.frame(
      x = vertices[, 1],
      y = vertices[, 2],
      z = vertices[, 3]
    ),
    faces = data.frame(i = faces[, 1], j = faces[, 2], k = faces[, 3]),
    metadata = list(
      n_centerline_points = n_points,
      centerline = centerline,
      tangents = frames$tangents
    )
  )
}


#' Compute parallel transport frames along curve
#' @keywords internal
compute_parallel_transp_fr <- function(curve) {
  # nolint: object_length_linter
  n <- nrow(curve)

  tangents <- matrix(0, nrow = n, ncol = 3)
  for (i in seq_len(n - 1)) {
    tangents[i, ] <- curve[i + 1, ] - curve[i, ]
    len <- sqrt(sum(tangents[i, ]^2))
    if (len > 0) tangents[i, ] <- tangents[i, ] / len
  }
  tangents[n, ] <- tangents[n - 1, ]

  t0 <- tangents[1, ]
  arbitrary <- if (abs(t0[1]) < 0.9) c(1, 0, 0) else c(0, 1, 0)
  n0 <- cross_product(t0, arbitrary)
  n0 <- n0 / sqrt(sum(n0^2))

  normals <- matrix(0, nrow = n, ncol = 3)
  binormals <- matrix(0, nrow = n, ncol = 3)

  normals[1, ] <- n0
  binormals[1, ] <- cross_product(t0, n0)

  for (i in seq_len(n - 1)) {
    t_curr <- tangents[i, ]
    t_next <- tangents[i + 1, ]

    cross_t <- cross_product(t_curr, t_next)
    cross_norm <- sqrt(sum(cross_t^2))

    if (cross_norm < 1e-10) {
      normals[i + 1, ] <- normals[i, ]
      binormals[i + 1, ] <- binormals[i, ]
    } else {
      axis <- cross_t / cross_norm
      angle <- acos(max(-1, min(1, sum(t_curr * t_next))))

      normals[i + 1, ] <- rotate_vector(normals[i, ], axis, angle)
      normals[i + 1, ] <- normals[i + 1, ] / sqrt(sum(normals[i + 1, ]^2))
      binormals[i + 1, ] <- cross_product(t_next, normals[i + 1, ])
    }
  }

  list(tangents = tangents, normals = normals, binormals = binormals)
}


#' Cross product of two 3D vectors
#' @keywords internal
cross_product <- function(a, b) {
  c(
    a[2] * b[3] - a[3] * b[2],
    a[3] * b[1] - a[1] * b[3],
    a[1] * b[2] - a[2] * b[1]
  )
}


#' Rotate vector around axis by angle (Rodrigues' formula)
#' @keywords internal
rotate_vector <- function(v, axis, angle) {
  cos_a <- cos(angle)
  sin_a <- sin(angle)
  v *
    cos_a +
    cross_product(axis, v) * sin_a +
    axis * sum(axis * v) * (1 - cos_a)
}

test_that("build_cortical_meshes skips hemisphere without data", {
  atlas_data <- data.frame(
    label = c("a"),
    region = c("region a"),
    hemi = c("left"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0, 1, 2))

  meshes <- build_cortical_meshes(
    atlas_data,
    c("left", "right"),
    "inflated",
    "#CCCCCC",
    NULL
  )

  expect_true(length(meshes) >= 1)
})

test_that("build_cortical_meshes with edge_by parameter", {
  atlas_data <- data.frame(
    label = c("a", "b"),
    region = c("region a", "region b"),
    hemi = c("left", "left"),
    colour = c("#FF0000", "#00FF00"),
    lobe = c("frontal", "parietal"),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(0:50, 51:100)

  meshes <- build_cortical_meshes(
    atlas_data,
    "left",
    "inflated",
    "#CCCCCC",
    "lobe"
  )

  expect_true(length(meshes) > 0)
  expect_true(!is.null(meshes[[1]]$edgeColor))
})

test_that("build_subcortical_meshes handles subcort data", {
  atlas_data <- data.frame(
    label = c("Left-Caudate"),
    region = c("caudate"),
    hemi = c("subcort"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(
    list(
      vertices = data.frame(x = 1:3, y = 1:3, z = 1:3),
      faces = data.frame(i = 1, j = 2, k = 3)
    )
  )

  meshes <- build_subcortical_meshes(atlas_data, "#CCCCCC")

  expect_true(length(meshes) > 0)
})

test_that("build_tract_meshes handles tract data with legacy meshes", {
  atlas_data <- data.frame(
    label = c("tract1"),
    region = c("tract 1"),
    hemi = c("subcort"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(
    list(
      vertices = data.frame(x = 1:4, y = 1:4, z = 1:4),
      faces = data.frame(i = c(1, 2), j = c(2, 3), k = c(3, 4))
    )
  )

  meshes <- build_tract_meshes(atlas_data, "#CCCCCC", "colour", NULL)

  expect_true(length(meshes) > 0)
})

test_that("make_mesh_entry creates correct structure", {
  vertices <- data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2))
  faces <- data.frame(i = 1, j = 2, k = 3)

  entry <- make_mesh_entry(
    name = "test mesh",
    vertices = vertices,
    faces = faces,
    colors = c("#FF0000", "#00FF00", "#0000FF"),
    color_mode = "vertexcolor",
    opacity = 0.8,
    hover_text = "hover info"
  )

  expect_equal(entry$name, "test mesh")
  expect_equal(entry$vertices$x, c(0, 1, 2))
  expect_equal(entry$vertices$y, c(0, 1, 2))
  expect_equal(entry$vertices$z, c(0, 1, 2))
  expect_equal(entry$faces$i, 0L)
  expect_equal(entry$faces$j, 1L)
  expect_equal(entry$faces$k, 2L)
  expect_equal(entry$colorMode, "vertexcolor")
  expect_equal(entry$opacity, 0.8)
  expect_equal(entry$hoverText, "hover info")
})

test_that("make_mesh_entry includes boundary edges when provided", {
  vertices <- data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2))
  faces <- data.frame(i = 1, j = 2, k = 3)
  boundary <- list(c(0, 1), c(1, 2))

  entry <- make_mesh_entry(
    name = "test",
    vertices = vertices,
    faces = faces,
    colors = c("#FF0000", "#00FF00", "#0000FF"),
    boundary_edges = boundary
  )

  expect_equal(entry$boundaryEdges, boundary)
})

test_that("make_mesh_entry includes edge color and width when provided", {
  vertices <- data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2))
  faces <- data.frame(i = 1, j = 2, k = 3)

  entry <- make_mesh_entry(
    name = "test",
    vertices = vertices,
    faces = faces,
    colors = c("#FF0000", "#00FF00", "#0000FF"),
    edge_color = "#000000",
    edge_width = 2
  )

  expect_equal(entry$edgeColor, "#000000")
  expect_equal(entry$edgeWidth, 2)
})

test_that("make_mesh_entry uses default edge width of 1", {
  vertices <- data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2))
  faces <- data.frame(i = 1, j = 2, k = 3)

  entry <- make_mesh_entry(
    name = "test",
    vertices = vertices,
    faces = faces,
    colors = c("#FF0000", "#00FF00", "#0000FF"),
    edge_color = "#000000"
  )

  expect_equal(entry$edgeWidth, 1)
})

test_that("build_subcortical_meshes creates mesh entries for each region", {
  atlas_data <- data.frame(
    label = c("Left-Caudate", "Right-Caudate"),
    colour = c("#FF0000", "#00FF00"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(
    list(
      vertices = data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2)),
      faces = data.frame(i = c(1, 2), j = c(2, 3), k = c(3, 1))
    ),
    list(
      vertices = data.frame(x = c(3, 4, 5), y = c(3, 4, 5), z = c(3, 4, 5)),
      faces = data.frame(i = c(1, 2), j = c(2, 3), k = c(3, 1))
    )
  )

  meshes <- build_subcortical_meshes(atlas_data, "#CCCCCC")

  expect_length(meshes, 2)
  expect_equal(meshes[[1]]$name, "Left-Caudate")
  expect_equal(meshes[[2]]$name, "Right-Caudate")
  expect_equal(meshes[[1]]$colorMode, "facecolor")
})

test_that("build_subcortical_meshes handles NA colours", {
  atlas_data <- data.frame(
    label = c("region_a"),
    colour = NA_character_,
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(
    list(
      vertices = data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2)),
      faces = data.frame(i = 1, j = 2, k = 3)
    )
  )

  meshes <- build_subcortical_meshes(atlas_data, "#AABBCC")

  expect_equal(unique(meshes[[1]]$colors), "#AABBCC")
})

test_that("build_subcortical_meshes skips NULL meshes", {
  atlas_data <- data.frame(
    label = c("region_a", "region_b"),
    colour = c("#FF0000", "#00FF00"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(
    NULL,
    list(
      vertices = data.frame(x = c(0, 1, 2), y = c(0, 1, 2), z = c(0, 1, 2)),
      faces = data.frame(i = 1, j = 2, k = 3)
    )
  )

  meshes <- build_subcortical_meshes(atlas_data, "#CCCCCC")

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$name, "region_b")
})

test_that("build_tract_meshes creates vertex-colored meshes", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(
    list(
      vertices = data.frame(
        x = c(0, 1, 2, 3),
        y = c(0, 1, 2, 3),
        z = c(0, 1, 2, 3)
      ),
      faces = data.frame(i = c(1, 2), j = c(2, 3), k = c(3, 4))
    )
  )

  meshes <- build_tract_meshes(atlas_data, "#CCCCCC", color_by = "colour")

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$colorMode, "vertexcolor")
  expect_equal(meshes[[1]]$name, "tract_a")
})

test_that("build_tract_meshes handles orientation coloring", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )

  tangents <- matrix(c(1, 0, 0, 0, 1, 0, 0, 0, 1), nrow = 3, byrow = TRUE)

  atlas_data$mesh <- list(
    list(
      vertices = data.frame(x = 1:6, y = 1:6, z = 1:6),
      faces = data.frame(i = c(1, 2), j = c(2, 3), k = c(3, 4)),
      metadata = list(
        n_centerline_points = 3,
        tangents = tangents
      )
    )
  )

  meshes <- build_tract_meshes(atlas_data, "#CCCCCC", color_by = "orientation")

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$colorMode, "vertexcolor")
  expect_length(meshes[[1]]$colors, 6)
})

test_that("tangents_to_colors computes direction-based colors", {
  mesh_data <- list(
    vertices = data.frame(x = 1:6, y = 1:6, z = 1:6),
    metadata = list(
      n_centerline_points = 3,
      tangents = matrix(
        c(
          1,
          0,
          0,
          0,
          1,
          0,
          0,
          0,
          1
        ),
        nrow = 3,
        byrow = TRUE
      )
    )
  )

  colors <- tangents_to_colors(mesh_data)

  expect_length(colors, 6)
  expect_equal(colors[1], grDevices::rgb(1, 0, 0))
  expect_equal(colors[2], grDevices::rgb(1, 0, 0))
  expect_equal(colors[3], grDevices::rgb(0, 1, 0))
  expect_equal(colors[4], grDevices::rgb(0, 1, 0))
  expect_equal(colors[5], grDevices::rgb(0, 0, 1))
  expect_equal(colors[6], grDevices::rgb(0, 0, 1))
})

test_that("tangents_to_colors handles mixed directions", {
  mesh_data <- list(
    vertices = data.frame(x = 1:4, y = 1:4, z = 1:4),
    metadata = list(
      n_centerline_points = 2,
      tangents = matrix(
        c(
          1,
          1,
          0,
          0,
          1,
          1
        ),
        nrow = 2,
        byrow = TRUE
      )
    )
  )

  colors <- tangents_to_colors(mesh_data)

  expect_length(colors, 4)
  expect_true(all(grepl("^#", colors)))
})

test_that("resolve_brain_mesh returns mesh for inflated surface", {
  mesh <- resolve_brain_mesh(hemisphere = "lh", surface = "inflated")

  expect_true(!is.null(mesh))
  expect_true("vertices" %in% names(mesh))
  expect_true("faces" %in% names(mesh))
  expect_true(nrow(mesh$vertices) > 0)
  expect_true(nrow(mesh$faces) > 0)
  expect_equal(ncol(mesh$vertices), 3)
  expect_equal(ncol(mesh$faces), 3)
})

test_that("resolve_brain_mesh returns both hemispheres", {
  lh <- resolve_brain_mesh(hemisphere = "lh", surface = "inflated")
  rh <- resolve_brain_mesh(hemisphere = "rh", surface = "inflated")

  expect_true(!is.null(lh))
  expect_true(!is.null(rh))
  expect_equal(nrow(lh$vertices), nrow(rh$vertices))
})

test_that("resolve_brain_mesh validates arguments", {
  expect_error(resolve_brain_mesh(hemisphere = "invalid"))
  expect_error(resolve_brain_mesh(surface = "invalid"))
})

test_that("is_unified_atlas identifies unified atlases correctly", {
  expect_true(is_unified_atlas(dk()))
  expect_true(is_unified_atlas(aseg()))

  expect_false(is_unified_atlas(list()))
  expect_false(is_unified_atlas(data.frame()))
  expect_false(is_unified_atlas(NULL))
  expect_false(is_unified_atlas("dk"))
})

test_that("class-based atlas checks work correctly", {
  expect_true(is_cortical_atlas(dk()))
  expect_false(is_subcortical_atlas(dk()))
  expect_false(is_tract_atlas(dk()))

  expect_true(is_subcortical_atlas(aseg()))
  expect_false(is_cortical_atlas(aseg()))
  expect_false(is_tract_atlas(aseg()))
})


test_that("vertices_to_colors creates correct color vector", {
  atlas_data <- data.frame(
    region = c("a", "b"),
    colour = c("#FF0000", "#00FF00"),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0, 1, 2), c(3, 4))

  colors <- vertices_to_colors(
    atlas_data,
    n_vertices = 6,
    na_colour = "#CCCCCC"
  )

  expect_length(colors, 6)
  expect_equal(colors[1:3], rep("#FF0000", 3))
  expect_equal(colors[4:5], rep("#00FF00", 2))
  expect_equal(colors[6], "#CCCCCC")
})

test_that("vertices_to_colors handles NA colours", {
  atlas_data <- data.frame(
    region = c("a", "b"),
    colour = c("#FF0000", NA),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0, 1), c(2, 3))

  colors <- vertices_to_colors(
    atlas_data,
    n_vertices = 5,
    na_colour = "#AAAAAA"
  )

  expect_equal(colors[1:2], rep("#FF0000", 2))
  expect_equal(colors[3:4], rep("#AAAAAA", 2))
  expect_equal(colors[5], "#AAAAAA")
})

test_that("vertices_to_groups creates correct group vector", {
  atlas_data <- data.frame(
    region = c("precentral", "postcentral"),
    lobe = c("frontal", "parietal"),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0, 1, 2), c(3, 4))

  groups <- vertices_to_groups(atlas_data, n_vertices = 6, group_col = "lobe")

  expect_length(groups, 6)
  expect_equal(groups[1:3], rep("frontal", 3))
  expect_equal(groups[4:5], rep("parietal", 2))
  expect_true(is.na(groups[6]))
})

test_that("vertices_to_groups errors on missing column", {
  atlas_data <- data.frame(region = "a")
  atlas_data$vertices <- list(c(0, 1))

  expect_error(
    vertices_to_groups(atlas_data, n_vertices = 3, group_col = "missing"),
    "not found"
  )
})

test_that("resolve_brain_mesh returns inflated surfaces", {
  lh <- resolve_brain_mesh(hemisphere = "lh", surface = "inflated")
  rh <- resolve_brain_mesh(hemisphere = "rh", surface = "inflated")

  expect_true(!is.null(lh))
  expect_true(!is.null(rh))
})

test_that("resolve_brain_mesh returns white surface", {
  mesh <- resolve_brain_mesh(hemisphere = "lh", surface = "white")
  expect_true(!is.null(mesh))
  expect_true("vertices" %in% names(mesh))
  expect_true("faces" %in% names(mesh))
})

test_that("vertices_to_colors handles empty vertices", {
  atlas_data <- data.frame(
    region = c("a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(integer(0))

  colors <- vertices_to_colors(
    atlas_data,
    n_vertices = 5,
    na_colour = "#CCCCCC"
  )

  expect_equal(colors, rep("#CCCCCC", 5))
})

test_that("vertices_to_colors handles out-of-bounds indices", {
  atlas_data <- data.frame(
    region = c("a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(-1, 0, 100))

  colors <- vertices_to_colors(
    atlas_data,
    n_vertices = 5,
    na_colour = "#CCCCCC"
  )

  expect_equal(colors[1], "#FF0000")
  expect_equal(colors[5], "#CCCCCC")
})

test_that("vertices_to_groups handles NA group values", {
  atlas_data <- data.frame(
    region = c("a", "b"),
    lobe = c("frontal", NA),
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0, 1), c(2, 3))

  groups <- vertices_to_groups(atlas_data, n_vertices = 5, group_col = "lobe")

  expect_equal(groups[1:2], rep("frontal", 2))
  expect_true(is.na(groups[3]))
  expect_true(is.na(groups[4]))
})

test_that("is_unified_atlas detects atlas with data component", {
  atlas <- structure(
    list(
      core = data.frame(label = "a", region = "r", hemi = "left"),
      data = structure(
        list(vertices = data.frame(label = "a")),
        class = "ggseg_atlas_data"
      )
    ),
    class = "ggseg_atlas"
  )
  atlas$data$vertices$vertices <- list(1:10)

  expect_true(is_unified_atlas(atlas))
})

test_that("is_unified_atlas returns FALSE for atlas without 3d data", {
  atlas <- structure(
    list(
      core = data.frame(label = "a", region = "r", hemi = "left"),
      data = structure(
        list(geometry = data.frame()),
        class = "ggseg_atlas_data"
      )
    ),
    class = "ggseg_atlas"
  )

  expect_false(is_unified_atlas(atlas))
})

test_that("is_subcortical_atlas detects subcortical atlases", {
  expect_true(is_subcortical_atlas(aseg()))
  expect_false(is_subcortical_atlas(dk()))
})

test_that("is_unified_atlas detects direct vertices", {
  atlas <- structure(
    list(
      core = data.frame(label = "a", region = "r", hemi = "left"),
      vertices = data.frame(label = "a")
    ),
    class = "ggseg_atlas"
  )
  atlas$vertices$vertices <- list(1:10)

  expect_true(is_unified_atlas(atlas))
})

test_that("is_unified_atlas detects direct meshes", {
  atlas <- structure(
    list(
      core = data.frame(label = "a", region = "r", hemi = "subcort"),
      meshes = data.frame(label = "a")
    ),
    class = "ggseg_atlas"
  )

  expect_true(is_unified_atlas(atlas))
})

test_that("cross_product computes correct cross products", {
  expect_equal(cross_product(c(1, 0, 0), c(0, 1, 0)), c(0, 0, 1))
  expect_equal(cross_product(c(0, 1, 0), c(0, 0, 1)), c(1, 0, 0))
  expect_equal(cross_product(c(1, 0, 0), c(1, 0, 0)), c(0, 0, 0))
})

test_that("rotate_vector rotates correctly", {
  rotated <- rotate_vector(c(1, 0, 0), c(0, 0, 1), pi / 2)
  expect_equal(rotated[1], 0, tolerance = 1e-10)
  expect_equal(rotated[2], 1, tolerance = 1e-10)
  expect_equal(rotated[3], 0, tolerance = 1e-10)

  no_rotation <- rotate_vector(c(1, 0, 0), c(0, 0, 1), 0)
  expect_equal(no_rotation, c(1, 0, 0), tolerance = 1e-10)
})

test_that("generate_tube_mesh creates correct mesh structure", {
  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0, 3, 0, 0),
    nrow = 4,
    byrow = TRUE
  )

  result <- generate_tube_mesh(centerline, radius = 0.5, segments = 6)

  expect_true("vertices" %in% names(result))
  expect_true("faces" %in% names(result))
  expect_true("metadata" %in% names(result))
  expect_equal(nrow(result$vertices), 4 * 6)
  expect_equal(nrow(result$faces), (4 - 1) * 6 * 2)
  expect_equal(result$metadata$n_centerline_points, 4)
})

test_that("generate_tube_mesh accepts per-point radius", {
  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0),
    nrow = 3,
    byrow = TRUE
  )

  result <- generate_tube_mesh(
    centerline,
    radius = c(0.5, 1.0, 0.5),
    segments = 4
  )

  expect_equal(nrow(result$vertices), 3 * 4)
})

test_that("generate_tube_mesh errors on bad input", {
  expect_error(generate_tube_mesh(matrix(1:3, nrow = 1)), "at least 2 rows")
  expect_error(generate_tube_mesh(c(1, 2, 3)), "matrix")
})

test_that("compute_parallel_transp_fr returns correct structure", {
  curve <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 1, 0, 3, 1, 1),
    nrow = 4,
    byrow = TRUE
  )

  frames <- compute_parallel_transp_fr(curve)

  expect_true("tangents" %in% names(frames))
  expect_true("normals" %in% names(frames))
  expect_true("binormals" %in% names(frames))
  expect_equal(nrow(frames$tangents), 4)
  expect_equal(nrow(frames$normals), 4)
  expect_equal(nrow(frames$binormals), 4)

  for (i in 1:4) {
    expect_equal(sqrt(sum(frames$tangents[i, ]^2)), 1, tolerance = 1e-10)
    expect_equal(sqrt(sum(frames$normals[i, ]^2)), 1, tolerance = 1e-10)
    expect_equal(sqrt(sum(frames$binormals[i, ]^2)), 1, tolerance = 1e-10)
  }
})

test_that("build_tract_meshes with centerlines creates tube meshes", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )

  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0),
    nrow = 3,
    byrow = TRUE
  )
  tangents <- matrix(
    c(1, 0, 0, 1, 0, 0, 1, 0, 0),
    nrow = 3,
    byrow = TRUE
  )

  atlas_centerlines <- list(
    centerlines = data.frame(label = "tract_a", stringsAsFactors = FALSE),
    tube_radius = 0.5,
    tube_segments = 4
  )
  atlas_centerlines$centerlines$points <- list(centerline)
  atlas_centerlines$centerlines$tangents <- list(tangents)

  meshes <- build_tract_meshes(
    atlas_data,
    "#CCCCCC",
    color_by = "colour",
    atlas_centerlines = atlas_centerlines
  )

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$name, "tract_a")
  expect_equal(meshes[[1]]$colorMode, "vertexcolor")
  expect_equal(length(meshes[[1]]$colors), 3 * 4)
})

test_that("build_tract_meshes warns with no data", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )

  expect_warning(
    meshes <- build_tract_meshes(atlas_data, "#CCCCCC"),
    "No centerlines or meshes"
  )
  expect_length(meshes, 0)
})

test_that("build_tract_meshes with centerlines and orientation coloring", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )

  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0),
    nrow = 3,
    byrow = TRUE
  )
  tangents <- matrix(
    c(1, 0, 0, 0, 1, 0, 0, 0, 1),
    nrow = 3,
    byrow = TRUE
  )

  atlas_centerlines <- list(
    centerlines = data.frame(label = "tract_a", stringsAsFactors = FALSE),
    tube_radius = 0.5,
    tube_segments = 4
  )
  atlas_centerlines$centerlines$points <- list(centerline)
  atlas_centerlines$centerlines$tangents <- list(tangents)

  meshes <- build_tract_meshes(
    atlas_data,
    "#CCCCCC",
    color_by = "orientation",
    atlas_centerlines = atlas_centerlines
  )

  expect_length(meshes, 1)
  expect_true(all(grepl("^#", meshes[[1]]$colors)))
})

test_that("resolve_brain_mesh returns NULL for empty brain_meshes", {
  result <- resolve_brain_mesh(
    hemisphere = "lh", surface = "pial", brain_meshes = list()
  )
  expect_null(result)
})

test_that("resolve_brain_mesh anatomically separates hemispheres on x", {
  lh <- resolve_brain_mesh("lh", "inflated")
  rh <- resolve_brain_mesh("rh", "inflated")

  expect_lte(max(lh$vertices$x), 1e-6)
  expect_gte(min(rh$vertices$x), -1e-6)
})

test_that("resolve_brain_mesh separates hemispheres for pial surface", {
  skip_if_not_installed("ggseg.meshes")
  lh <- resolve_brain_mesh("lh", "pial")
  rh <- resolve_brain_mesh("rh", "pial")

  expect_lte(max(lh$vertices$x), 1e-6)
  expect_gte(min(rh$vertices$x), -1e-6)
})

test_that("pial meshes use shared axis convention after normalization", {
  skip_if_not_installed("ggseg.meshes")
  pial_lh <- resolve_brain_mesh("lh", "pial")
  infl_lh <- resolve_brain_mesh("lh", "inflated")

  # Both surfaces should have LH medial edge at x = 0 and widest extent
  # on the AP axis (y).
  expect_equal(max(pial_lh$vertices$x), 0, tolerance = 1e-6)
  expect_equal(max(infl_lh$vertices$x), 0, tolerance = 1e-6)
  expect_gt(
    diff(range(pial_lh$vertices$y)),
    diff(range(pial_lh$vertices$x))
  )
})

test_that("is_flat_mesh detects flat surfaces", {
  flat <- data.frame(x = c(0, 1, 0), y = c(0, 0, 1), z = c(0, 0, 0))
  vol <- data.frame(x = c(0, 1, 0), y = c(0, 0, 1), z = c(0, 0, 1))
  expect_true(is_flat_mesh(flat))
  expect_false(is_flat_mesh(vol))
})

test_that("build_tract_meshes with mesh data (no centerlines)", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = c("#FF0000"),
    stringsAsFactors = FALSE
  )
  mesh <- list(
    vertices = data.frame(x = c(0, 1, 0), y = c(0, 0, 1), z = c(0, 0, 0)),
    faces = data.frame(i = 0, j = 1, k = 2)
  )
  atlas_data$mesh <- list(mesh)

  meshes <- build_tract_meshes(atlas_data, "#CCCCCC", color_by = "colour")

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$name, "tract_a")
  expect_equal(meshes[[1]]$colors, rep("#FF0000", 3))
})

test_that("build_tract_meshes skips NULL mesh entries", {
  atlas_data <- data.frame(
    label = c("tract_a", "tract_b"),
    colour = c("#FF0000", "#00FF00"),
    stringsAsFactors = FALSE
  )
  mesh <- list(
    vertices = data.frame(x = c(0, 1, 0), y = c(0, 0, 1), z = c(0, 0, 0)),
    faces = data.frame(i = 0, j = 1, k = 2)
  )
  atlas_data$mesh <- list(mesh, NULL)

  meshes <- build_tract_meshes(atlas_data, "#CCCCCC", color_by = "colour")

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$name, "tract_a")
})

test_that("build_cortical_meshes warns when brain mesh not found", {
  atlas_data <- data.frame(
    region = "precentral",
    hemi = "left",
    colour = "#FF0000",
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0, 1, 2))

  expect_warning(
    build_cortical_meshes(
      atlas_data,
      hemisphere = "left",
      surface = "pial",
      na_colour = "#CCCCCC",
      edge_by = NULL,
      brain_meshes = list()
    ),
    "Brain mesh not found"
  )
})

test_that("build_tract_meshes skips labels not in centerlines", {
  atlas_data <- data.frame(
    label = c("tract_a", "tract_missing"),
    colour = c("#FF0000", "#00FF00"),
    stringsAsFactors = FALSE
  )

  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0),
    nrow = 3, byrow = TRUE
  )
  tangents <- matrix(
    c(1, 0, 0, 1, 0, 0, 1, 0, 0),
    nrow = 3, byrow = TRUE
  )

  atlas_centerlines <- list(
    centerlines = data.frame(label = "tract_a", stringsAsFactors = FALSE),
    tube_radius = 0.5,
    tube_segments = 4
  )
  atlas_centerlines$centerlines$points <- list(centerline)
  atlas_centerlines$centerlines$tangents <- list(tangents)

  meshes <- build_tract_meshes(
    atlas_data, "#CCCCCC",
    color_by = "colour",
    atlas_centerlines = atlas_centerlines
  )

  expect_length(meshes, 1)
  expect_equal(meshes[[1]]$name, "tract_a")
})

test_that("build_tract_meshes applies na_colour for NA colour", {
  atlas_data <- data.frame(
    label = c("tract_a"),
    colour = NA_character_,
    stringsAsFactors = FALSE
  )
  mesh <- list(
    vertices = data.frame(x = c(0, 1, 0), y = c(0, 0, 1), z = c(0, 0, 0)),
    faces = data.frame(i = 0, j = 1, k = 2)
  )
  atlas_data$mesh <- list(mesh)

  meshes <- build_tract_meshes(
    atlas_data, "#CCCCCC", color_by = "colour"
  )

  expect_equal(meshes[[1]]$colors, rep("#CCCCCC", 3))
})

test_that("build_centerline_data returns NULL when no centerlines", {
  atlas <- structure(
    list(
      data = structure(
        list(centerlines = NULL),
        class = c("ggseg_data_tract", "ggseg_atlas_data")
      )
    ),
    class = c("tract_atlas", "ggseg_atlas", "list")
  )

  expect_null(build_centerline_data(atlas))
})

test_that("build_centerline_data skips NULL points in centerlines", {
  cl_data <- data.frame(
    label = c("tract_a", "tract_b"),
    stringsAsFactors = FALSE
  )
  cl_data$points <- list(
    matrix(c(0, 0, 0, 1, 0, 0), nrow = 2, byrow = TRUE),
    NULL
  )

  atlas <- structure(
    list(
      data = structure(
        list(centerlines = cl_data),
        class = c("ggseg_data_tract", "ggseg_atlas_data")
      )
    ),
    class = c("tract_atlas", "ggseg_atlas", "list")
  )

  result <- build_centerline_data(atlas)

  expect_type(result, "list")
  expect_null(result$centerlines$points[[2]])
  expect_equal(nrow(result$centerlines$points[[1]]), 2)
})

test_that("Check that ggseg3d is working", {
  p <- ggseg3d()
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true("meshes" %in% names(p$x))
  expect_true("options" %in% names(p$x))
  expect_true(length(p$x$meshes) > 0)
  rm(p)

  lifecycle::expect_deprecated(
    p <- ggseg3d(atlas = "aseg")
  )
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true(length(p$x$meshes) > 0)

  expect_error(ggseg3d(atlas = hhj), "object 'hhj")

  expect_warning(
    ggseg3d(
      .data = data.frame(
        region = c(
          "transverse tempral",
          "insula",
          "precentral",
          "superior parietal"
        ),
        p = sample(seq(0, .5, .001), 4),
        stringsAsFactors = FALSE
      ),
      colour_by = "p"
    )
  )

  some_data <- data.frame(
    region = c(
      "transverse temporal",
      "insula",
      "precentral",
      "superior parietal"
    ),
    p = sample(seq(0, .5, .001), 4),
    stringsAsFactors = FALSE
  )

  p <- ggseg3d(
    .data = some_data,
    colour_by = "p",
    text_by = "p",
    palette = c("black", "white")
  )
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  p <- ggseg3d(
    .data = some_data,
    colour_by = "p",
    text_by = "p",
    palette = c("black", "white")
  )
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true(!is.null(p$x$colorbar))
  expect_true(p$x$options$showLegend)

  p_hidden <- p |> set_legend(FALSE)
  expect_false(p_hidden$x$options$showLegend)

  p_sized <- ggseg3d() |> set_dimensions(width = 800, height = 600)
  expect_equal(p_sized$width, 800)
  expect_equal(p_sized$height, 600)
})

test_that("ggseg3d works with aseg subcortical atlas", {
  p <- ggseg3d(atlas = aseg())
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true(length(p$x$meshes) > 0)
})

test_that("ggseg3d with left hemisphere only", {
  p <- ggseg3d(hemisphere = "left")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true(length(p$x$meshes) > 0)
})

test_that("ggseg3d with inflated surface", {
  p <- ggseg3d(hemisphere = "left", surface = "inflated")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("ggseg3d handles edge_by parameter", {
  some_data <- data.frame(
    region = c(
      "transverse temporal",
      "insula",
      "precentral",
      "superior parietal"
    ),
    lobe = c("temporal", "insular", "frontal", "parietal"),
    stringsAsFactors = FALSE
  )

  p <- ggseg3d(
    .data = some_data,
    hemisphere = "left",
    edge_by = "lobe"
  )
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("ggseg3d default colorbar is present", {
  p <- ggseg3d()
  expect_true(!is.null(p$x$colorbar) || p$x$colorbar$type == "discrete")
})

test_that("ggseg3d with custom palette", {
  some_data <- data.frame(
    region = c("transverse temporal", "insula"),
    p = c(0.1, 0.9),
    stringsAsFactors = FALSE
  )

  p <- ggseg3d(
    .data = some_data,
    hemisphere = "left",
    colour_by = "p",
    palette = c("blue" = 0, "white" = 0.5, "red" = 1)
  )
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_equal(p$x$colorbar$type, "continuous")
})

test_that("ggseg3d with na_colour and na_alpha", {
  p <- ggseg3d(hemisphere = "left", na_colour = "red", na_alpha = 0.5)
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("ggseg3d with label_by parameter", {
  p <- ggseg3d(hemisphere = "left", label_by = "label")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("deprecated params trigger warnings", {
  some_data <- data.frame(
    region = c("precentral", "insula"),
    p = c(0.1, 0.5),
    stringsAsFactors = FALSE
  )

  lifecycle::expect_deprecated(
    ggseg3d(hemisphere = "left", colour = "colour")
  )
  lifecycle::expect_deprecated(
    ggseg3d(hemisphere = "left", label = "label")
  )
  lifecycle::expect_deprecated(
    ggseg3d(.data = some_data, text = "p")
  )
})

test_that("ggseg3d with both hemispheres", {
  p <- ggseg3d(hemisphere = c("left", "right"))
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true(length(p$x$meshes) >= 2)
})

test_that("ggseg3d with atlas object instead of string", {
  p <- ggseg3d(atlas = dk(), hemisphere = "left")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("ggseg3d unified atlas without user data", {
  p <- ggseg3d(atlas = dk(), hemisphere = "left", .data = NULL)
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
})

test_that("ggseg3d with aseg mesh atlas", {
  p <- ggseg3d(atlas = aseg(), hemisphere = "subcort")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_true(length(p$x$meshes) > 0)
})

test_that("ggseg3d errors on invalid atlas object", {
  expect_error(ggseg3d(atlas = list()), "ggseg_atlas")
  expect_error(ggseg3d(atlas = data.frame()), "ggseg_atlas")
})

test_that("prepare_brain_meshes handles atlas with centerlines", {
  atlas_data <- data.frame(
    label = "tract_a",
    stringsAsFactors = FALSE
  )

  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0),
    nrow = 3,
    byrow = TRUE
  )
  tangents <- matrix(
    c(1, 0, 0, 1, 0, 0, 1, 0, 0),
    nrow = 3,
    byrow = TRUE
  )

  cl_data <- data.frame(label = "tract_a", stringsAsFactors = FALSE)
  cl_data$points <- list(centerline)
  cl_data$tangents <- list(tangents)

  atlas <- structure(
    list(
      atlas = "test_tract",
      type = "tract",
      core = data.frame(
        label = "tract_a",
        region = "tract a",
        hemi = "subcort",
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(centerlines = cl_data),
        class = c("ggseg_data_tract", "ggseg_atlas_data")
      ),
      palette = c("tract_a" = "#FF0000")
    ),
    class = c("tract_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas)

  expect_type(prepared, "list")
  expect_true(length(prepared$meshes) > 0)
})

test_that("prepare_brain_meshes handles atlas$data$meshes path", {
  meshes_data <- data.frame(
    label = "Left-Caudate",
    stringsAsFactors = FALSE
  )
  meshes_data$mesh <- list(
    list(
      vertices = data.frame(x = 1:3, y = 1:3, z = 1:3),
      faces = data.frame(i = 1L, j = 2L, k = 3L)
    )
  )

  atlas <- structure(
    list(
      atlas = "test_subcort",
      type = "subcortical",
      core = data.frame(
        label = "Left-Caudate",
        region = "caudate",
        hemi = "subcort",
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(meshes = meshes_data),
        class = c("ggseg_data_subcortical", "ggseg_atlas_data")
      ),
      palette = c("Left-Caudate" = "#FF0000")
    ),
    class = c("subcortical_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas)

  expect_type(prepared, "list")
  expect_true(length(prepared$meshes) > 0)
})

test_that("prepare_brain_meshes uses orientation coloring for tracts", {
  centerline <- matrix(
    c(0, 0, 0, 1, 0, 0, 2, 0, 0),
    nrow = 3, byrow = TRUE
  )
  tangents <- matrix(
    c(1, 0, 0, 0, 1, 0, 0, 0, 1),
    nrow = 3, byrow = TRUE
  )

  cl_data <- data.frame(label = "tract_a", stringsAsFactors = FALSE)
  cl_data$points <- list(centerline)
  cl_data$tangents <- list(tangents)

  meshes_data <- data.frame(
    label = "tract_a",
    stringsAsFactors = FALSE
  )
  meshes_data$mesh <- list(NULL)

  atlas <- structure(
    list(
      atlas = "test_tract",
      type = "tract",
      core = data.frame(
        label = "tract_a", region = "tract a", hemi = "subcort",
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(
          meshes = meshes_data,
          centerlines = cl_data
        ),
        class = c("ggseg_data_tract", "ggseg_atlas_data")
      ),
      palette = c("tract_a" = "#FF0000")
    ),
    class = c("tract_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas, tract_color = "orientation")

  expect_true(length(prepared$meshes) > 0)
  expect_true(all(grepl("^#", prepared$meshes[[1]]$colors)))
})

test_that("prepare_brain_meshes handles cerebellar atlas with vertices", {
  vertices_data <- data.frame(
    label = "left_I-IV",
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(0L:99L)

  atlas <- structure(
    list(
      atlas = "suit_lobules",
      type = "cerebellar",
      core = data.frame(
        label = "left_I-IV",
        region = "I-IV",
        hemi = "left",
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data),
        class = c("ggseg_data_cerebellar", "ggseg_atlas_data")
      ),
      palette = c("left_I-IV" = "#FF0000")
    ),
    class = c("cerebellar_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas)

  expect_type(prepared, "list")
  expect_true(length(prepared$meshes) > 0)
  expect_equal(prepared$meshes[[1]]$colorMode, "vertexcolor")
  expect_equal(prepared$meshes[[1]]$name, "cerebellum")
  expect_equal(
    length(prepared$meshes[[1]]$vertices$x),
    nrow(ggseg.formats::get_cerebellar_mesh()$vertices)
  )
})

test_that("cerebellar atlas colors correct vertices", {
  vertices_data <- data.frame(
    label = c("left_I-IV", "right_V"),
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(0L:4L, 100L:104L)

  atlas <- structure(
    list(
      atlas = "test_cer",
      type = "cerebellar",
      core = data.frame(
        label = c("left_I-IV", "right_V"),
        region = c("I-IV", "V"),
        hemi = c("left", "right"),
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data),
        class = c("ggseg_data_cerebellar", "ggseg_atlas_data")
      ),
      palette = c(
        "left_I-IV" = "#FF0000",
        "right_V" = "#00FF00"
      )
    ),
    class = c("cerebellar_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas)
  colors <- prepared$meshes[[1]]$colors

  expect_equal(colors[1:5], rep("#FF0000", 5))
  expect_equal(colors[101:105], rep("#00FF00", 5))
  expect_equal(colors[50], "darkgrey")
})

test_that("cerebellar atlas with deep nuclei renders mixed surface + meshes", {
  vertices_data <- data.frame(
    label = "left_I-IV",
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(0L:4L)

  deep_meshes <- data.frame(
    label = "Left-Dentate",
    stringsAsFactors = FALSE
  )
  deep_meshes$mesh <- list(
    list(
      vertices = data.frame(x = 1:4, y = 1:4, z = 1:4),
      faces = data.frame(i = 1L, j = 2L, k = 3L)
    )
  )

  atlas <- structure(
    list(
      atlas = "suit_deep",
      type = "cerebellar",
      core = data.frame(
        label = c("left_I-IV", "Left-Dentate"),
        region = c("I-IV", "Dentate"),
        hemi = c("left", "left"),
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data, meshes = deep_meshes),
        class = c("ggseg_data_cerebellar", "ggseg_atlas_data")
      ),
      palette = c(
        "left_I-IV" = "#FF0000",
        "Left-Dentate" = "#0000FF"
      )
    ),
    class = c("cerebellar_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas)

  expect_true(length(prepared$meshes) >= 2)
  surface <- prepared$meshes[[1]]
  expect_equal(surface$name, "cerebellum")
  expect_equal(surface$colorMode, "vertexcolor")
  expect_equal(surface$opacity, 0.3)

  deep <- prepared$meshes[[2]]
  expect_equal(deep$name, "Dentate")
  expect_equal(deep$colorMode, "facecolor")
})

test_that("cerebellar surface_opacity can be overridden", {
  vertices_data <- data.frame(
    label = "left_I-IV",
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(0L:4L)

  deep_meshes <- data.frame(
    label = "Left-Dentate",
    stringsAsFactors = FALSE
  )
  deep_meshes$mesh <- list(
    list(
      vertices = data.frame(x = 1:4, y = 1:4, z = 1:4),
      faces = data.frame(i = 1L, j = 2L, k = 3L)
    )
  )

  atlas <- structure(
    list(
      atlas = "suit_deep",
      type = "cerebellar",
      core = data.frame(
        label = c("left_I-IV", "Left-Dentate"),
        region = c("I-IV", "Dentate"),
        hemi = c("left", "left"),
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data, meshes = deep_meshes),
        class = c("ggseg_data_cerebellar", "ggseg_atlas_data")
      ),
      palette = c(
        "left_I-IV" = "#FF0000",
        "Left-Dentate" = "#0000FF"
      )
    ),
    class = c("cerebellar_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas, surface_opacity = 0.5)
  expect_equal(prepared$meshes[[1]]$opacity, 0.5)
})

test_that("merge_legend_data handles NULL inputs", {
  legend <- data.frame(label = "a", colour = "#FF0000")

  expect_null(merge_legend_data(NULL, NULL))
  expect_equal(merge_legend_data(NULL, legend), legend)
  expect_equal(merge_legend_data(legend, NULL), legend)

  combined <- merge_legend_data(legend, legend)
  expect_equal(nrow(combined), 1)
})

test_that("build_cerebellar_meshes errors when mesh unavailable", {
  local_mocked_bindings(
    get_cerebellar_mesh = function(...) NULL,
    .package = "ggseg.formats"
  )
  expect_error(
    build_cerebellar_meshes(data.frame(), "darkgrey"),
    "SUIT cerebellar mesh"
  )
})

test_that("build_cerebellar_meshes flags flatmap meshes", {
  flat_mesh <- list(
    vertices = data.frame(
      x = c(0, 1, 2, 3),
      y = c(0, 1, 0, 1),
      z = c(0, 0, 0, 0)
    ),
    faces = data.frame(i = c(1L, 2L), j = c(2L, 3L), k = c(3L, 4L))
  )
  local_mocked_bindings(
    get_cerebellar_mesh = function(...) flat_mesh,
    .package = "ggseg.formats"
  )

  atlas_data <- data.frame(
    label = "region1",
    region = "region1",
    colour = "#FF0000"
  )
  atlas_data$vertices <- list(0:3)

  meshes <- build_cerebellar_meshes(atlas_data, "darkgrey")

  expect_length(meshes, 1)
  expect_true(isTRUE(meshes[[1]]$isFlatmap))
})

test_that("check_ggseg_meshes errors when package missing", {
  local_mocked_bindings(
    requireNamespace = function(...) FALSE,
    .package = "base"
  )
  expect_error(check_ggseg_meshes("pial"), "ggseg.meshes")
})

test_that("cerebellar atlas with text_by populates vertex texts", {
  vertices_data <- data.frame(
    label = "left_I-IV",
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(0L:4L)

  atlas <- structure(
    list(
      atlas = "suit_text",
      type = "cerebellar",
      core = data.frame(
        label = "left_I-IV",
        region = "I-IV",
        hemi = "left",
        score = 0.75,
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data),
        class = c("ggseg_data_cerebellar", "ggseg_atlas_data")
      ),
      palette = c("left_I-IV" = "#FF0000")
    ),
    class = c("cerebellar_atlas", "ggseg_atlas", "list")
  )

  prepared <- prepare_brain_meshes(atlas, text_by = "score")
  texts <- prepared$meshes[[1]]$vertexTexts
  expect_true(!is.null(texts))
  expect_match(texts[1], "score")
})

test_that("prepare_brain_meshes.default errors on unknown atlas class", {
  fake <- structure(list(), class = "weird_atlas")
  expect_error(prepare_brain_meshes(fake), "No method")
})

test_that("ggseg3d errors on nonexistent string atlas name", {
  expect_error(
    lifecycle::expect_deprecated(
      ggseg3d(atlas = "nonexistent_atlas_xyz")
    ),
    "Could not find atlas"
  )
})

test_that("vertices_to_text returns NA vector when column is missing", {
  atlas_data <- data.frame(
    label = "a",
    region = "r",
    stringsAsFactors = FALSE
  )
  atlas_data$vertices <- list(c(0L, 1L))

  result <- vertices_to_text(atlas_data, 3, "nonexistent")

  expect_equal(length(result), 3)
  expect_true(all(is.na(result)))
})

test_that("text_by works with subcortical atlas", {
  some_data <- data.frame(
    region = c("Thalamus", "Caudate"),
    p = c(0.1, 0.5),
    stringsAsFactors = FALSE
  )

  p <- ggseg3d(.data = some_data, atlas = aseg(), text_by = "p")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  hover_texts <- vapply(
    p$x$meshes,
    function(m) m$hoverText %||% "",
    character(1)
  )
  expect_true(any(grepl("p:", hover_texts)))
})

test_that("text_by works with tract atlas", {
  some_data <- data.frame(
    region = c("arcuate fasciculus", "corticospinal tract"),
    fa = c(0.45, 0.55),
    stringsAsFactors = FALSE
  )

  p <- ggseg3d(.data = some_data, atlas = tracula(), text_by = "fa")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  hover_texts <- vapply(
    p$x$meshes,
    function(m) m$hoverText %||% "",
    character(1)
  )
  expect_true(any(grepl("fa:", hover_texts)))
})

#' Add glass brain to ggseg3d plot
#'
#' Adds a translucent brain surface to a ggseg3d plot for anatomical reference.
#' Particularly useful for subcortical and tract visualizations where spatial
#' context helps interpretation. Works with both htmlwidget (`ggseg3d`) and
#' rgl (`ggsegray`) objects.
#'
#' @param p A `ggseg3d` widget or `ggsegray` rgl object.
#' @param hemisphere Character vector. Hemispheres to add: "left", "right",
#'   or both.
#' @param surface Character. Surface type. Defaults to `"inflated"`, which
#'   is supplied by `ggseg.formats` and requires no additional dependency.
#'   Other surfaces (`"pial"`, `"white"`, etc.) are drawn from `ggseg.meshes`.
#' @param colour Character. Colour for the glass brain surface (hex or named).
#' @param opacity Numeric. Transparency of the glass brain (0-1).
#' @param brain_meshes Optional user-supplied brain meshes. See
#'   [ggseg.formats::get_brain_mesh()] for format details.
#'
#' @return The input object (modified), for piping.
#' @export
#'
#' @examples
#' \donttest{
#' ggseg3d(atlas = aseg()) |>
#'   add_glassbrain("left", opacity = 0.2)
#' }
#' \dontrun{
#' # rgl requires OpenGL; not run in check environments.
#' ggsegray(atlas = aseg()) |>
#'   add_glassbrain(opacity = 0.15) |>
#'   pan_camera("right lateral")
#' }
add_glassbrain <- function(
  p,
  hemisphere = c("left", "right"),
  surface = "inflated",
  colour = "#CCCCCC",
  opacity = 0.3,
  brain_meshes = NULL
) {
  is_rgl <- inherits(p, "ggsegray")

  if (is_rgl) {
    check_ggsegray(p)
  } else {
    check_ggseg3d(p)
    existing <- p$x$meshes %||% list()
    if (any(vapply(existing, function(m) isTRUE(m$isFlatmap), logical(1)))) {
      cli::cli_warn(c(
        "Skipping glassbrain: widget contains a flat (2D) mesh.",
        "i" = "Flatmap surfaces share no coordinate frame with anatomical
          meshes; combining them produces misaligned geometry."
      ))
      return(p)
    }
  }

  colour <- if (grepl("^#", colour)) colour else col2hex(colour)
  hemi_map <- c("left" = "lh", "right" = "rh")
  cortical_hemis <- intersect(hemisphere, c("left", "right"))

  entries <- lapply(cortical_hemis, function(hemi) {
    hemi_short <- hemi_map[hemi]
    mesh <- resolve_brain_mesh(
      hemisphere = hemi_short,
      surface = surface,
      brain_meshes = brain_meshes
    )

    if (is.null(mesh)) {
      cli::cli_warn(
        "Brain mesh not available for {.val {hemi}} {.val {surface}}. Skipping."
      )
      return(NULL)
    }

    make_mesh_entry(
      name = paste("glass brain", hemi),
      vertices = mesh$vertices,
      faces = mesh$faces,
      colors = rep(colour, nrow(mesh$vertices)),
      color_mode = "vertexcolor",
      opacity = opacity
    )
  })
  entries <- Filter(Negate(is.null), entries)

  if (is_rgl) {
    lapply(entries, function(entry) {
      mesh3d <- mesh_entry_to_mesh3d(entry) # nolint [object_usage_linter]
      rgl::shade3d(mesh3d)
    })
  } else {
    p$x$meshes <- c(entries, p$x$meshes)
  }

  p
}


#' Pan camera position of ggseg3d plot
#'
#' Sets the camera position for a ggseg3d widget or ggsegray rgl scene
#' to standard anatomical views or custom positions.
#'
#' @param p A `ggseg3d` widget or `ggsegray` rgl object.
#' @param camera string, list, or numeric vector. Camera position preset
#'   name, custom eye position list, or `c(x, y, z)` for rgl.
#'
#' \strong{Available camera presets:}
#' \itemize{
#' \item `left lateral` or `left_lateral`
#' \item `left medial` or `left_medial`
#' \item `right lateral` or `right_lateral`
#' \item `right medial` or `right_medial`
#' \item `left superior` or `left_superior`
#' \item `right superior` or `right_superior`
#' \item `left inferior` or `left_inferior`
#' \item `right inferior` or `right_inferior`
#' \item `left anterior` or `left_anterior`
#' \item `right anterior` or `right_anterior`
#' \item `left posterior` or `left_posterior`
#' \item `right posterior` or `right_posterior`
#' }
#'
#' @return The input object (modified), for piping.
#' @export
#'
#' @examples
#' ggseg3d() |> pan_camera("right lateral")
#' \dontrun{
#' # rgl requires OpenGL; not run in check environments.
#' ggsegray(atlas = dk(), hemisphere = "left") |>
#'   pan_camera("left lateral")
#' }
pan_camera <- function(p, camera) {
  if (inherits(p, "ggsegray")) {
    check_ggsegray(p)

    cam_pos <- if (is.numeric(camera)) {
      camera
    } else if (is.character(camera)) {
      camera_preset_to_position(camera) # nolint [object_usage_linter]
    } else {
      cli::cli_abort(c(
        "{.arg camera} must be a character string or numeric vector,",
        "not {.obj_type_friendly {camera}}."
      ))
    }

    um <- look_at_origin(cam_pos) # nolint [object_usage_linter]
    rgl::view3d(userMatrix = um, fov = 0)
    return(p)
  }

  check_ggseg3d(p)
  if (!is.character(camera) && !is.list(camera)) {
    cli::cli_abort(c(
      "{.arg camera} must be a character string or list,",
      "not {.obj_type_friendly {camera}}."
    ))
  }

  p$x$options$camera <- camera
  p
}


#' Set background color of ggseg3d plot
#'
#' Changes the background color of a ggseg3d widget or ggsegray rgl scene.
#'
#' @param p A `ggseg3d` widget or `ggsegray` rgl object.
#' @param colour string. Background color (hex or named color)
#'
#' @return The input object (modified), for piping.
#' @export
#'
#' @examples
#' ggseg3d() |> set_background("black")
#' \dontrun{
#' # rgl requires OpenGL; not run in check environments.
#' ggsegray(atlas = dk()) |> set_background("black")
#' }
set_background <- function(p, colour = "#ffffff") {
  if (inherits(p, "ggsegray")) {
    check_ggsegray(p)
    if (!grepl("^#", colour)) {
      colour <- col2hex(colour)
    }
    rgl::bg3d(color = colour)
    return(p)
  }

  check_ggseg3d(p)

  if (!grepl("^#", colour)) {
    colour <- col2hex(colour)
  }

  p$x$options$backgroundColor <- colour
  p
}


#' Set legend visibility
#'
#' For htmlwidget output, toggles legend visibility. For rgl output,
#' draws or removes the legend overlay.
#'
#' @param p A ggseg3d or ggsegray object
#' @param show logical. Whether to show the legend (default: TRUE)
#'
#' @return The input object, modified
#' @export
#'
#' @examples
#' ggseg3d() |> set_legend(FALSE)
#' \dontrun{
#' # rgl requires OpenGL; not run in check environments.
#' ggsegray(hemisphere = "left") |> set_legend()
#' }
set_legend <- function(p, show = TRUE) {
  if (inherits(p, "ggsegray")) {
    check_ggsegray(p)
    if (show && !is.null(p$legend_data)) {
      render_legend_rgl(p$legend_data)
    }
    return(p)
  }

  check_ggseg3d(p)
  p$x$options$showLegend <- show
  p
}


#' Set widget dimensions
#'
#' Changes the width and height of a ggseg3d widget.
#'
#' @param p ggseg3d widget object
#' @param width numeric. Widget width in pixels (NULL for default)
#' @param height numeric. Widget height in pixels (NULL for default)
#'
#' @return ggseg3d widget object with updated dimensions
#' @export
#'
#' @examples
#' ggseg3d() |>
#'   set_dimensions(width = 800, height = 600)
set_dimensions <- function(p, width = NULL, height = NULL) {
  check_ggseg3d(p)
  if (!is.null(width)) {
    p$width <- width
  }
  if (!is.null(height)) {
    p$height <- height
  }
  p
}


#' Set region boundary edges
#'
#' Adds coloured outlines around brain regions. This is useful for
#' highlighting region boundaries in figures. Works with both
#' htmlwidget (`ggseg3d`) and rgl (`ggsegray`) objects. For rgl,
#' edges must have been computed at creation time via `edge_by`.
#'
#' @param p A `ggseg3d` widget or `ggsegray` rgl object.
#' @param colour string. Edge colour (hex or named color). Set to NULL to
#'   hide edges.
#' @param width numeric. Width of edge lines (default: 1). Note: line width > 1
#'   may not render on all systems due to WebGL limitations.
#'
#' @return The input object (modified), for piping.
#' @export
#'
#' @section Lifecycle:
#' `r lifecycle::badge("experimental")`
#'
#' @examples
#' ggseg3d(hemisphere = "left", edge_by = "region") |>
#'   set_edges("black") |>
#'   pan_camera("left lateral")
#' \dontrun{
#' # rgl requires OpenGL; not run in check environments.
#' ggsegray(hemisphere = "left", edge_by = "region") |>
#'   set_edges("red", width = 2) |>
#'   pan_camera("left lateral")
#' }
set_edges <- function(p, colour = "black", width = 1) {
  lifecycle::signal_stage("experimental", "set_edges()")

  if (inherits(p, "ggsegray")) {
    check_ggsegray(p)

    if (length(p$edge_ids) > 0) {
      lapply(p$edge_ids, function(eid) rgl::pop3d(id = eid))
      p$edge_ids <- integer(0)
    }

    if (!is.null(colour)) {
      edge_col <- if (grepl("^#", colour)) colour else col2hex(colour)
      p$edge_ids <- unlist(lapply(p$meshes, function(mesh_entry) {
        render_edges_rgl(mesh_entry, colour = edge_col, width = width)
      }))
    }

    return(p)
  }

  check_ggseg3d(p)

  if (is.null(colour)) {
    p$x$meshes <- lapply(p$x$meshes, function(mesh) {
      mesh$edgeColor <- NULL
      mesh$edgeWidth <- NULL
      mesh
    })
  } else {
    edge_col <- if (grepl("^#", colour)) colour else col2hex(colour)
    p$x$meshes <- lapply(p$x$meshes, function(mesh) {
      mesh$edgeColor <- edge_col
      mesh$edgeWidth <- width
      mesh
    })
  }
  p
}


#' Enable flat shading for ggseg3d plot
#'
#' Disables lighting effects to show colors exactly as specified.
#' Useful for screenshots where accurate color reproduction is needed,
#' such as atlas creation pipelines that extract contours from images.
#'
#' @param p ggseg3d widget object
#' @param flat logical. Enable flat shading (default: TRUE)
#'
#' @return ggseg3d widget object with updated shading
#' @export
#'
#' @examples
#' ggseg3d() |>
#'   set_flat_shading()
set_flat_shading <- function(p, flat = TRUE) {
  check_ggseg3d(p)
  p$x$options$flatShading <- flat
  p
}


#' Enable orthographic camera for ggseg3d plot
#'
#' Uses orthographic projection instead of perspective. This eliminates
#' perspective distortion and ensures consistent sizing across all views.
#'
#' @param p ggseg3d widget object
#' @param ortho logical. Enable orthographic mode (default: TRUE)
#' @param frustum_size numeric. Size of the orthographic frustum. Controls
#'   how much of the scene is visible. Default 220 works well for brain meshes.
#'   Use the same value across all views for consistent sizing.
#'
#' @return ggseg3d widget object with updated camera mode
#' @export
#'
#' @examples
#' ggseg3d() |>
#'   set_orthographic()
set_orthographic <- function(p, ortho = TRUE, frustum_size = 220) {
  check_ggseg3d(p)
  p$x$options$orthographic <- ortho
  p$x$options$frustumSize <- frustum_size
  p
}


#' Set hemisphere positioning mode
#'
#' Repositions cortical hemisphere meshes in a ggseg3d widget. Meshes are
#' produced with anatomical positioning by default (medial edges at x = 0).
#' Use this to centre each hemisphere at the origin for atlas-creation
#' snapshots, or to reapply anatomical positioning after manual edits.
#'
#' @param p ggseg3d widget object
#' @param positioning How to position hemispheres:
#'   - "anatomical": Medial surfaces adjacent at midline. Left at negative
#'     x, right at positive x. Default for displaying both hemispheres
#'     together.
#'   - "centered": Centre each hemisphere at the origin. Best for
#'     single-hemisphere snapshots where consistent sizing is needed.
#'
#' @details Only cortical hemisphere meshes are repositioned (those named
#'   `"<hemi> <surface>"`, e.g. `"left inflated"`). Subcortical, cerebellar
#'   and glassbrain meshes are left untouched.
#'
#' @return ggseg3d widget object with repositioned meshes
#' @export
#'
#' @examples
#' # View both hemispheres anatomically positioned (default)
#' ggseg3d(hemisphere = c("left", "right")) |>
#'   pan_camera("left lateral")
#'
#' # Atlas creation: centered for consistent sizing
#' ggseg3d(hemisphere = "left") |>
#'   set_positioning("centered") |>
#'   set_orthographic() |>
#'   pan_camera("left lateral")
set_positioning <- function(p, positioning = c("anatomical", "centered")) {
  check_ggseg3d(p)
  positioning <- match.arg(positioning)

  surface_pattern <- paste0(
    "^(left|right) ",
    "(inflated|semi-inflated|white|pial|midthickness|",
    "sphere|smoothwm|orig)$"
  )

  p$x$meshes <- lapply(p$x$meshes, function(mesh) {
    name <- mesh$name %||% ""
    if (!grepl(surface_pattern, name)) return(mesh)

    is_left <- startsWith(name, "left ")
    vertices <- mesh$vertices
    x_range <- range(vertices$x)
    x_center <- mean(x_range)
    half_width <- (x_range[2] - x_range[1]) / 2

    vertices$x <- vertices$x - x_center
    if (positioning == "anatomical") {
      vertices$x <- vertices$x +
        if (is_left) -half_width else half_width
    }

    mesh$vertices <- vertices
    mesh
  })

  p
}


#' Save ggseg3d widget as image
#'
#' Takes a screenshot of a ggseg3d widget and saves it as a PNG image.
#' Requires a Chrome-based browser to be installed.
#'
#' @param p ggseg3d widget object
#' @param file string. Output file path (should end in .png)
#' @param width numeric. Image width in pixels (default: 600)
#' @param height numeric. Image height in pixels (default: 500)
#' @param delay numeric. Seconds to wait for widget to render before capture
#'   (default: 1)
#' @param zoom numeric. Zoom factor for higher resolution (default: 2)
#' @param ... Additional arguments passed to webshot2::webshot
#'
#' @return The file path (invisibly)
#' @export
#' @importFrom webshot2 webshot
#'
#' @examples
#' \dontrun{
#' ggseg3d() |>
#'   pan_camera("left lateral") |>
#'   snapshot_brain("brain.png")
#' }
# nocov start
snapshot_brain <- function(
  p,
  file,
  width = 600,
  height = 500,
  delay = 1,
  zoom = 2,
  ...
) {
  check_ggseg3d(p)

  tmpfile <- tempfile(fileext = ".html")
  on.exit(unlink(tmpfile), add = TRUE)

  htmlwidgets::saveWidget(p, tmpfile, selfcontained = TRUE)

  webshot2::webshot(
    url = tmpfile,
    file = file,
    vwidth = width,
    vheight = height,
    delay = delay,
    zoom = zoom,
    ...
  )

  invisible(file)
}
# nocov end

test_that("set_background works", {
  p <- ggseg3d() |> set_background("black")
  expect_equal(p$x$options$backgroundColor, "#000000")

  p <- ggseg3d() |> set_background("#FF5733")
  expect_equal(p$x$options$backgroundColor, "#FF5733")

  p <- ggseg3d() |> set_background("white")
  expect_equal(p$x$options$backgroundColor, "#FFFFFF")
})

test_that("set_legend works", {
  p <- ggseg3d() |> set_legend(TRUE)
  expect_true(p$x$options$showLegend)

  p <- ggseg3d() |> set_legend(FALSE)
  expect_false(p$x$options$showLegend)
})

test_that("set_dimensions works", {
  p <- ggseg3d() |> set_dimensions(width = 800, height = 600)
  expect_equal(p$width, 800)
  expect_equal(p$height, 600)

  p <- ggseg3d() |> set_dimensions(width = 1200)
  expect_equal(p$width, 1200)
  expect_null(p$height)

  p <- ggseg3d() |> set_dimensions(height = 400)
  expect_null(p$width)
  expect_equal(p$height, 400)
})

test_that("set_edges works", {
  p <- ggseg3d() |> set_edges("red")
  expect_equal(p$x$meshes[[1]]$edgeColor, "#FF0000")
  expect_equal(p$x$meshes[[1]]$edgeWidth, 1)

  p <- ggseg3d() |> set_edges("#00FF00", width = 2)
  expect_equal(p$x$meshes[[1]]$edgeColor, "#00FF00")
  expect_equal(p$x$meshes[[1]]$edgeWidth, 2)

  p <- ggseg3d() |>
    set_edges("black") |>
    set_edges(NULL)
  expect_null(p$x$meshes[[1]]$edgeColor)
  expect_null(p$x$meshes[[1]]$edgeWidth)
})

test_that("set_flat_shading works", {
  p <- ggseg3d() |> set_flat_shading(TRUE)
  expect_true(p$x$options$flatShading)

  p <- ggseg3d() |> set_flat_shading(FALSE)
  expect_false(p$x$options$flatShading)

  p <- ggseg3d() |> set_flat_shading()
  expect_true(p$x$options$flatShading)
})

test_that("set_orthographic works", {
  p <- ggseg3d() |> set_orthographic(TRUE)
  expect_true(p$x$options$orthographic)
  expect_equal(p$x$options$frustumSize, 220)

  p <- ggseg3d() |> set_orthographic(FALSE)
  expect_false(p$x$options$orthographic)

  p <- ggseg3d() |> set_orthographic(TRUE, frustum_size = 300)
  expect_true(p$x$options$orthographic)
  expect_equal(p$x$options$frustumSize, 300)
})

test_that("additions reject non-ggseg3d objects", {
  fake_widget <- list(x = list())
  class(fake_widget) <- "htmlwidget"

  expect_error(set_background(fake_widget, "black"), "ggseg3d")
  expect_error(set_legend(fake_widget, TRUE), "ggseg3d")
  expect_error(set_dimensions(fake_widget, 800, 600), "ggseg3d")
  expect_error(set_edges(fake_widget, "black"), "ggseg3d")
  expect_error(set_flat_shading(fake_widget), "ggseg3d")
  expect_error(set_orthographic(fake_widget), "ggseg3d")
  expect_error(pan_camera(fake_widget, "left lateral"), "ggseg3d")
})

test_that("pan_camera validates input", {
  expect_error(
    ggseg3d() |> pan_camera(123),
    "character string or list"
  )
})

test_that("set_positioning centers hemispheres", {
  p <- ggseg3d(hemisphere = c("left", "right")) |>
    set_positioning("centered")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  for (m in p$x$meshes) {
    if (!grepl("left|right", m$name, ignore.case = TRUE)) {
      next
    }
    x_range <- range(m$vertices$x)
    x_center <- mean(x_range)
    expect_true(abs(x_center) < 1)
  }
})

test_that("set_positioning anatomical offsets hemispheres", {
  p <- ggseg3d(hemisphere = c("left", "right")) |>
    set_positioning("anatomical")
  expect_s3_class(p, c("ggseg3d", "htmlwidget"))

  left_mesh <- NULL
  right_mesh <- NULL
  for (m in p$x$meshes) {
    if (grepl("left", m$name, ignore.case = TRUE)) {
      left_mesh <- m
    }
    if (grepl("right", m$name, ignore.case = TRUE)) right_mesh <- m
  }
  expect_true(mean(range(left_mesh$vertices$x)) < 0)
  expect_true(mean(range(right_mesh$vertices$x)) > 0)
})

test_that("set_positioning leaves subcortical region meshes untouched", {
  p <- ggseg3d(atlas = aseg())
  original_meshes <- p$x$meshes

  p_positioned <- p |> set_positioning("centered")
  for (i in seq_along(p_positioned$x$meshes)) {
    expect_equal(
      p_positioned$x$meshes[[i]]$vertices,
      original_meshes[[i]]$vertices
    )
  }
})

test_that("set_positioning rejects non-ggseg3d objects", {
  expect_error(set_positioning(list()), "ggseg3d")
})

test_that("ggseg3d produces anatomically positioned hemispheres by default", {
  p <- ggseg3d(hemisphere = c("left", "right"))
  left <- NULL
  right <- NULL
  for (m in p$x$meshes) {
    if (m$name == "left inflated") left <- m
    if (m$name == "right inflated") right <- m
  }
  expect_lte(max(left$vertices$x), 1e-6)
  expect_gte(min(right$vertices$x), -1e-6)
})

test_that("add_glassbrain warns and returns unchanged when widget is flat", {
  p <- ggseg3d(hemisphere = "left")
  p$x$meshes[[1]]$isFlatmap <- TRUE
  n <- length(p$x$meshes)
  expect_warning(result <- add_glassbrain(p), "flatmap|flat")
  expect_equal(length(result$x$meshes), n)
})

test_that("add_glassbrain warns for unavailable mesh", {
  expect_warning(
    ggseg3d(atlas = aseg()) |>
      add_glassbrain(
        hemisphere = "left",
        surface = "pial",
        brain_meshes = list()
      ),
    "not available"
  )
})

test_that("additions can be chained", {
  p <- ggseg3d() |>
    set_background("black") |>
    set_legend(FALSE) |>
    set_flat_shading(TRUE) |>
    set_orthographic(TRUE) |>
    pan_camera("left lateral")

  expect_s3_class(p, c("ggseg3d", "htmlwidget"))
  expect_equal(p$x$options$backgroundColor, "#000000")
  expect_false(p$x$options$showLegend)
  expect_true(p$x$options$flatShading)
  expect_true(p$x$options$orthographic)
  expect_equal(p$x$options$camera, "left lateral")
})

#' Render brain atlas with rgl
#'
#' Creates an rgl 3D scene from a brain atlas. Uses the same atlas
#' preparation pipeline as [ggseg3d()] but outputs to rgl instead of
#' htmlwidgets. The resulting scene can be piped into [pan_camera()],
#' [add_glassbrain()], and [set_background()], then rendered with
#' rayshader's `render_highquality()` or captured with `rgl::snapshot3d()`.
#'
#' @inheritParams ggseg3d
#' @param material Named list of rgl material properties passed to
#'   [rgl::tmesh3d()]. Controls how the mesh surface is shaded.
#'
#' @section Material properties:
#' Useful material list entries:
#' \describe{
#'   \item{`specular`}{`"black"` (matte) or `"white"` (glossy).}
#'   \item{`shininess`}{Specular exponent. Higher = tighter highlights.}
#'   \item{`lit`}{`FALSE` disables lighting.}
#'   \item{`alpha`}{Transparency, 0 (invisible) to 1 (opaque).}
#'   \item{`smooth`}{`TRUE` for Gouraud shading, `FALSE` for flat.}
#' }
#'
#' See [rgl::material3d()] for the full list.
#'
#' @template type-specific-args
#'
#' @return An object of class `ggsegray` (invisibly), which wraps the
#'   rgl device ID. Pipe into [pan_camera()], [add_glassbrain()], or
#'   [set_background()] to modify the scene.
#'
#' @importFrom graphics par plot.new plot.window rect text
#'
#' @examples
#' \dontrun{
#' # rgl requires OpenGL; not run in check environments.
#' ggsegray(hemisphere = "left") |>
#'   pan_camera("left lateral")
#'
#' ggsegray(atlas = aseg()) |>
#'   add_glassbrain(opacity = 0.15) |>
#'   pan_camera("right lateral") |>
#'   set_background("black")
#' }
#'
#' @export
ggsegray <- function(
  .data = NULL,
  atlas = dk(), # nolint [object_usage_linter]
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  material = list(),
  ...,
  label = deprecated(),
  text = deprecated(),
  colour = deprecated()
) {
  if (lifecycle::is_present(label)) {
    lifecycle::deprecate_warn(
      "2.1.0",
      "ggsegray(label=)",
      "ggsegray(label_by=)"
    )
    label_by <- label
  }
  if (lifecycle::is_present(text)) {
    lifecycle::deprecate_warn("2.1.0", "ggsegray(text=)", "ggsegray(text_by=)")
    text_by <- text
  }
  if (lifecycle::is_present(colour)) {
    lifecycle::deprecate_warn(
      "2.1.0",
      "ggsegray(colour=)",
      "ggsegray(colour_by=)"
    )
    colour_by <- colour
  }
  rlang::check_installed("rgl", reason = "to render 3D brain scenes with rgl")

  if (!inherits(atlas, "ggseg_atlas") && !inherits(atlas, "brain_atlas")) {
    cli::cli_abort(
      "{.arg atlas} must be a {.cls ggseg_atlas} object,
      not {.obj_type_friendly {atlas}}."
    )
  }

  if (!is_unified_atlas(atlas)) {
    cli::cli_abort(c(
      "{.arg atlas} is a {.cls ggseg_atlas} but has no 3D data.",
      "i" = "Use atlases from {.pkg ggseg.formats}
      that include vertex or mesh data."
    ))
  }

  prepared <- prepare_brain_meshes(
    atlas,
    .data = .data,
    label_by = label_by,
    text_by = text_by,
    colour_by = colour_by,
    palette = palette,
    na_colour = na_colour,
    na_alpha = na_alpha,
    ...
  )

  rgl::open3d()

  edge_ids <- unlist(lapply(prepared$meshes, function(mesh_entry) {
    mesh3d <- do.call(mesh_entry_to_mesh3d, c(list(mesh_entry), material))
    rgl::shade3d(mesh3d)
    render_edges_rgl(mesh_entry)
  }))

  structure(
    list(
      device = rgl::cur3d(),
      legend_data = prepared$legend_data,
      meshes = prepared$meshes,
      edge_ids = edge_ids
    ),
    class = "ggsegray"
  )
}


#' Convert mesh entry to rgl mesh3d object
#'
#' Converts the internal mesh_entry list structure (as built by
#' [make_mesh_entry()]) into an [rgl::tmesh3d()] object for rgl rendering.
#'
#' @param mesh_entry A mesh entry list with vertices, faces, colors,
#'   colorMode, and opacity.
#' @param ... Material properties merged into the `material` list of
#'   [rgl::tmesh3d()]. Overrides defaults (`specular = "black"`,
#'   `shininess = 128`). See [rgl::material3d()] for all options.
#'
#' @return An rgl `mesh3d` object
#' @keywords internal
mesh_entry_to_mesh3d <- function(mesh_entry, ...) {
  rlang::check_installed("rgl", reason = "to convert meshes")

  vb <- rbind(
    mesh_entry$vertices$x,
    mesh_entry$vertices$y,
    mesh_entry$vertices$z,
    1
  )

  it <- rbind(
    mesh_entry$faces$i + 1L,
    mesh_entry$faces$j + 1L,
    mesh_entry$faces$k + 1L
  )

  mesh_color <- if (mesh_entry$colorMode == "vertexcolor") {
    "vertices"
  } else {
    "faces"
  }

  alpha <- mesh_entry$opacity %||% 1

  material <- list(
    color = mesh_entry$colors,
    alpha = alpha,
    specular = "black",
    shininess = 128
  )
  extra <- list(...)
  material[names(extra)] <- extra

  rgl::tmesh3d(
    vertices = vb,
    indices = it,
    material = material,
    meshColor = mesh_color
  )
}


#' Map camera preset name to position vector
#'
#' Converts a camera preset string to an xyz position vector matching
#' the same presets used in the Three.js viewer.
#'
#' @param preset Character string naming the camera preset.
#'
#' @return Numeric vector of length 3 (x, y, z).
#' @keywords internal
camera_preset_to_position <- function(preset) {
  presets <- list(
    "left lateral" = c(-350, 0, 0),
    "left_lateral" = c(-350, 0, 0),
    "left medial" = c(350, 0, 0),
    "left_medial" = c(350, 0, 0),
    "right lateral" = c(350, 0, 0),
    "right_lateral" = c(350, 0, 0),
    "right medial" = c(-350, 0, 0),
    "right_medial" = c(-350, 0, 0),
    "left superior" = c(-120, 0, 330),
    "left_superior" = c(-120, 0, 330),
    "right superior" = c(120, 0, 330),
    "right_superior" = c(120, 0, 330),
    "left inferior" = c(-120, 0, -330),
    "left_inferior" = c(-120, 0, -330),
    "right inferior" = c(120, 0, -330),
    "right_inferior" = c(120, 0, -330),
    "left anterior" = c(-120, 330, 0),
    "left_anterior" = c(-120, 330, 0),
    "right anterior" = c(120, 330, 0),
    "right_anterior" = c(120, 330, 0),
    "left posterior" = c(-120, -330, 0),
    "left_posterior" = c(-120, -330, 0),
    "right posterior" = c(120, -330, 0),
    "right_posterior" = c(120, -330, 0)
  )

  pos <- presets[[preset]]
  if (is.null(pos)) {
    available <- unique(gsub("_", " ", names(presets))) # nolint: object_usage_linter
    cli::cli_abort(c(
      "Unknown camera preset {.val {preset}}.",
      "i" = "Available presets: {.val {available}}"
    ))
  }

  pos
}


#' Compute rgl rotation matrix to look at the origin from a given position
#'
#' Builds a 4x4 rotation matrix suitable for `rgl::view3d(userMatrix = ...)`
#' that orients the scene as if the camera is at `eye` looking toward the
#' origin with z pointing up.
#'
#' @param eye Numeric vector of length 3 (x, y, z) — camera position.
#'
#' @return A 4x4 rotation matrix.
#' @keywords internal
look_at_origin <- function(eye) {
  eye_n <- eye / sqrt(sum(eye^2))

  up <- c(0, 0, 1)
  if (abs(eye_n[3]) > 0.999) {
    up <- c(0, 1, 0)
  }

  fwd <- -eye_n
  right <- c(
    fwd[2] * up[3] - fwd[3] * up[2],
    fwd[3] * up[1] - fwd[1] * up[3],
    fwd[1] * up[2] - fwd[2] * up[1]
  )
  right <- right / sqrt(sum(right^2))

  actual_up <- c(
    right[2] * fwd[3] - right[3] * fwd[2],
    right[3] * fwd[1] - right[1] * fwd[3],
    right[1] * fwd[2] - right[2] * fwd[1]
  )

  m <- diag(4)
  m[1, 1:3] <- right
  m[2, 1:3] <- actual_up
  m[3, 1:3] <- eye_n
  m
}


render_edges_rgl <- function(mesh_entry, colour = NULL, width = NULL) {
  edges <- mesh_entry$boundaryEdges
  if (is.null(edges) || length(edges) == 0) {
    return(invisible(NULL))
  }

  edge_color <- colour %||% mesh_entry$edgeColor
  if (is.null(edge_color)) {
    return(invisible(NULL))
  }

  edge_width <- width %||% mesh_entry$edgeWidth %||% 1
  verts <- mesh_entry$vertices

  edge_matrix <- do.call(rbind, edges)
  idx1 <- edge_matrix[, 1] + 1L
  idx2 <- edge_matrix[, 2] + 1L

  cx <- mean(verts$x)
  cy <- mean(verts$y)
  cz <- mean(verts$z)
  nudge <- 0.3

  edge_x <- c(verts$x[idx1], verts$x[idx2])
  edge_y <- c(verts$y[idx1], verts$y[idx2])
  edge_z <- c(verts$z[idx1], verts$z[idx2])
  dx <- edge_x - cx
  dy <- edge_y - cy
  dz <- edge_z - cz
  d <- sqrt(dx^2 + dy^2 + dz^2)
  d[d == 0] <- 1
  edge_x <- edge_x + nudge * dx / d
  edge_y <- edge_y + nudge * dy / d
  edge_z <- edge_z + nudge * dz / d

  n <- length(idx1)
  x <- as.vector(rbind(edge_x[seq_len(n)], edge_x[n + seq_len(n)]))
  y <- as.vector(rbind(edge_y[seq_len(n)], edge_y[n + seq_len(n)]))
  z <- as.vector(rbind(edge_z[seq_len(n)], edge_z[n + seq_len(n)]))

  id <- rgl::segments3d(x, y, z, color = edge_color, lwd = edge_width)
  invisible(id)
}


#' @noRd
#' @export
print.ggsegray <- function(x, ...) {
  rgl::set3d(x$device)
  print(rgl::rglwidget())
}

#' @noRd
#' @importFrom knitr knit_print
#' @export
knit_print.ggsegray <- function(x, ...) {
  # nocov start
  invisible(x)
} # nocov end


check_ggsegray <- function(
  p,
  arg = rlang::caller_arg(p),
  call = rlang::caller_env()
) {
  if (!inherits(p, "ggsegray")) {
    cli::cli_abort(
      "{.arg {arg}} must be a {.cls ggsegray} object, not {.obj_type_friendly {p}}.", # nolint: line_length_linter
      call = call
    )
  }
  rlang::check_installed("rgl", reason = "to modify rgl scenes")
  rgl::set3d(p$device)
}


render_legend_rgl <- function(legend_data) {
  if (legend_data$type == "continuous") {
    render_continuous_legend_rgl(legend_data)
  } else if (legend_data$type == "discrete") {
    render_discrete_legend_rgl(legend_data)
  }
}


render_continuous_legend_rgl <- function(legend_data) {
  colors <- legend_data$colors
  title <- legend_data$title
  data_min <- legend_data$min
  data_max <- legend_data$max

  rgl::bgplot3d({
    par(mar = c(0, 0, 0, 0))
    plot.new()
    plot.window(xlim = c(0, 1), ylim = c(0, 1))

    bar_x <- 0.88
    bar_w <- 0.03
    bar_y0 <- 0.25
    bar_y1 <- 0.75
    n_steps <- 100

    y_seq <- seq(bar_y0, bar_y1, length.out = n_steps + 1)
    col_fn <- grDevices::colorRampPalette(colors)
    bar_cols <- col_fn(n_steps)

    invisible(lapply(seq_len(n_steps), function(i) {
      rect(
        bar_x,
        y_seq[i],
        bar_x + bar_w,
        y_seq[i + 1],
        col = bar_cols[i],
        border = NA
      )
    }))
    rect(bar_x, bar_y0, bar_x + bar_w, bar_y1, border = "black", lwd = 0.5)

    tick_x <- bar_x + bar_w + 0.005
    text(
      tick_x,
      bar_y0,
      format(data_min, digits = 2),
      adj = c(0, 0.5),
      cex = 0.7
    )
    text(
      tick_x,
      bar_y1,
      format(data_max, digits = 2),
      adj = c(0, 0.5),
      cex = 0.7
    )

    text(
      bar_x + bar_w / 2,
      bar_y1 + 0.03,
      title,
      adj = c(0.5, 0),
      cex = 0.8,
      font = 2
    )
  })
}


render_discrete_legend_rgl <- function(legend_data) {
  labels <- legend_data$labels
  colors <- legend_data$colors
  title <- legend_data$title

  n <- length(labels)
  max_show <- min(n, 30)
  labels <- labels[seq_len(max_show)]
  colors <- colors[seq_len(max_show)]

  rgl::bgplot3d({
    par(mar = c(0, 0, 0, 0))
    plot.new()
    plot.window(xlim = c(0, 1), ylim = c(0, 1))

    box_size <- 0.015
    line_h <- 0.025
    col_x <- 0.82
    top_y <- 0.95

    text(
      col_x + box_size + 0.005,
      top_y,
      title,
      adj = c(0, 0.5),
      cex = 0.7,
      font = 2
    )

    invisible(lapply(seq_len(max_show), function(i) {
      y <- top_y - i * line_h
      rect(
        col_x,
        y - box_size / 2,
        col_x + box_size,
        y + box_size / 2,
        col = colors[i],
        border = "grey50",
        lwd = 0.3
      )
      text(
        col_x + box_size + 0.005,
        y,
        labels[i],
        adj = c(0, 0.5),
        cex = 0.55
      )
    }))
  })
}

options(rgl.useNULL = TRUE)

test_that("camera_preset_to_position returns correct vectors", {
  expect_equal(camera_preset_to_position("left lateral"), c(-350, 0, 0))
  expect_equal(camera_preset_to_position("right lateral"), c(350, 0, 0))
  expect_equal(camera_preset_to_position("left medial"), c(350, 0, 0))
  expect_equal(camera_preset_to_position("right medial"), c(-350, 0, 0))
  expect_equal(camera_preset_to_position("left superior"), c(-120, 0, 330))
  expect_equal(camera_preset_to_position("right inferior"), c(120, 0, -330))
  expect_equal(camera_preset_to_position("left anterior"), c(-120, 330, 0))
  expect_equal(camera_preset_to_position("right posterior"), c(120, -330, 0))
})

test_that("camera_preset_to_position accepts underscore variants", {
  expect_equal(
    camera_preset_to_position("left_lateral"),
    camera_preset_to_position("left lateral")
  )
  expect_equal(
    camera_preset_to_position("right_superior"),
    camera_preset_to_position("right superior")
  )
})

test_that("camera_preset_to_position errors on unknown preset", {
  expect_error(
    camera_preset_to_position("top down"),
    "Unknown camera preset"
  )
})

test_that("look_at_origin produces orthonormal rotation matrix", {
  m <- look_at_origin(c(-350, 0, 0))
  expect_equal(dim(m), c(4, 4))
  expect_equal(m[4, ], c(0, 0, 0, 1))
  expect_equal(m[1:3, 4], c(0, 0, 0))

  rows <- m[1:3, 1:3]
  expect_equal(rows %*% t(rows), diag(3), tolerance = 1e-10)
})

test_that("look_at_origin handles vertical eye positions", {
  m <- look_at_origin(c(0, 0, 330))
  expect_equal(dim(m), c(4, 4))

  rows <- m[1:3, 1:3]
  expect_equal(rows %*% t(rows), diag(3), tolerance = 1e-10)
})

test_that("mesh_entry_to_mesh3d converts vertex-colored mesh", {
  skip_if_not_installed("rgl")

  entry <- make_mesh_entry(
    name = "test",
    vertices = data.frame(
      x = c(0, 1, 0, 1),
      y = c(0, 0, 1, 1),
      z = c(0, 0, 0, 1)
    ),
    faces = data.frame(i = c(1L, 2L), j = c(2L, 3L), k = c(3L, 4L)),
    colors = c("#FF0000", "#00FF00", "#0000FF", "#FFFFFF"),
    color_mode = "vertexcolor",
    opacity = 0.8
  )

  mesh3d <- mesh_entry_to_mesh3d(entry)

  expect_s3_class(mesh3d, "mesh3d")
  expect_equal(ncol(mesh3d$vb), 4)
  expect_equal(ncol(mesh3d$it), 2)
  expect_equal(mesh3d$material$alpha, 0.8)
  expect_equal(length(mesh3d$material$color), 4)
  expect_equal(mesh3d$meshColor, "vertices")
})

test_that("mesh_entry_to_mesh3d handles 0-indexed faces from make_mesh_entry", {
  skip_if_not_installed("rgl")

  entry <- make_mesh_entry(
    name = "test",
    vertices = data.frame(
      x = c(0, 1, 0),
      y = c(0, 0, 1),
      z = c(0, 0, 0)
    ),
    faces = data.frame(i = 1L, j = 2L, k = 3L),
    colors = c("#FF0000", "#00FF00", "#0000FF"),
    color_mode = "vertexcolor"
  )

  expect_equal(entry$faces$i, 0L)
  expect_equal(entry$faces$j, 1L)
  expect_equal(entry$faces$k, 2L)

  mesh3d <- mesh_entry_to_mesh3d(entry)
  expect_equal(mesh3d$it[1, 1], 1L)
  expect_equal(mesh3d$it[2, 1], 2L)
  expect_equal(mesh3d$it[3, 1], 3L)
})

test_that("mesh_entry_to_mesh3d converts face-colored mesh", {
  skip_if_not_installed("rgl")

  entry <- make_mesh_entry(
    name = "test",
    vertices = data.frame(
      x = c(0, 1, 0, 1),
      y = c(0, 0, 1, 1),
      z = c(0, 0, 0, 1)
    ),
    faces = data.frame(i = c(1L, 2L), j = c(2L, 3L), k = c(3L, 4L)),
    colors = c("#FF0000", "#00FF00"),
    color_mode = "facecolor"
  )

  mesh3d <- mesh_entry_to_mesh3d(entry)
  expect_equal(mesh3d$meshColor, "faces")
  expect_equal(length(mesh3d$material$color), 2)
})

test_that("prepare_brain_meshes returns meshes and legend_data", {
  prepared <- prepare_brain_meshes(
    dk(), hemisphere = "left", surface = "inflated"
  )

  expect_type(prepared, "list")
  expect_true("meshes" %in% names(prepared))
  expect_true("legend_data" %in% names(prepared))
  expect_true(length(prepared$meshes) > 0)
})

test_that("ggsegray errors on invalid atlas", {
  skip_if_not_installed("rgl")

  expect_error(
    ggsegray(atlas = list()),
    "ggseg_atlas"
  )
})

test_that("ggsegray errors when rgl not installed", {
  local_mocked_bindings(
    check_installed = function(pkg, ...) {
      if ("rgl" %in% pkg) {
        stop("rgl is required")
      }
    },
    .package = "rlang"
  )

  expect_error(ggsegray(hemisphere = "left"), "rgl")
})

test_that("ggsegray creates rgl scene", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk())

  expect_s3_class(p, "ggsegray")
  expect_true(is.integer(p$device))
  rgl::close3d()
})

test_that("ggsegray works with aseg atlas", {
  skip_if_not_installed("rgl")

  p <- ggsegray(atlas = aseg())

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("pan_camera works with ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk()) |>
    pan_camera("left lateral")

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("pan_camera with numeric vector works for ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk()) |>
    pan_camera(c(-400, 0, 0))

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("set_background works with ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk()) |>
    set_background("black")

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("add_glassbrain works with ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk()) |>
    add_glassbrain(hemisphere = "left")

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("ggsegray piping chain works", {
  skip_if_not_installed("rgl")

  p <- ggsegray(atlas = aseg()) |>
    add_glassbrain(opacity = 0.15) |>
    pan_camera("right lateral") |>
    set_background("black")

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("render_edges_rgl draws boundary edges", {
  skip_if_not_installed("rgl")

  entry <- make_mesh_entry(
    name = "test",
    vertices = data.frame(
      x = c(0, 1, 0, 1),
      y = c(0, 0, 1, 1),
      z = c(0, 0, 0, 0)
    ),
    faces = data.frame(i = c(1L, 2L), j = c(2L, 3L), k = c(3L, 4L)),
    colors = c("#FF0000", "#00FF00", "#0000FF", "#FF0000"),
    color_mode = "vertexcolor",
    boundary_edges = list(c(0L, 1L), c(1L, 2L)),
    edge_color = "#000000"
  )

  rgl::open3d()
  expect_silent(render_edges_rgl(entry))
  rgl::close3d()
})

test_that("render_edges_rgl skips when no edges", {
  entry <- make_mesh_entry(
    name = "test",
    vertices = data.frame(x = c(0, 1, 0), y = c(0, 0, 1), z = c(0, 0, 0)),
    faces = data.frame(i = 1L, j = 2L, k = 3L),
    colors = c("#FF0000", "#00FF00", "#0000FF"),
    color_mode = "vertexcolor"
  )

  expect_silent(render_edges_rgl(entry))
})

test_that("ggsegray renders edges when edge_by is set", {
  skip_if_not_installed("rgl")

  p <- ggsegray(
    atlas = dk(),
    hemisphere = "left",
    edge_by = "region"
  )

  expect_s3_class(p, "ggsegray")
  rgl::close3d()
})

test_that("check_ggsegray rejects non-ggsegray objects", {
  expect_error(check_ggsegray(list()), "ggsegray")
  expect_error(check_ggsegray("not a scene"), "ggsegray")
})

test_that("ggsegray errors on non-unified atlas", {
  skip_if_not_installed("rgl")

  fake_atlas <- structure(
    list(
      core = data.frame(label = "a", region = "r", hemi = "left"),
      data = structure(list(), class = "ggseg_atlas_data")
    ),
    class = "ggseg_atlas"
  )

  expect_error(ggsegray(atlas = fake_atlas), "no 3D data")
})

test_that("pan_camera rejects bad type for ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk())

  expect_error(
    pan_camera(p, list(x = 1)),
    "character string or numeric"
  )
  rgl::close3d()
})

test_that("set_legend works with ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk())
  result <- set_legend(p)
  expect_s3_class(result, "ggsegray")
  rgl::close3d()
})

test_that("set_edges works with ggsegray", {
  skip_if_not_installed("rgl")

  p <- ggsegray(
    atlas = dk(),
    hemisphere = "left",
    edge_by = "region"
  )

  p2 <- set_edges(p, "red", width = 2)
  expect_s3_class(p2, "ggsegray")

  p3 <- set_edges(p2, NULL)
  expect_s3_class(p3, "ggsegray")
  expect_equal(length(p3$edge_ids), 0)

  rgl::close3d()
})

test_that("ggsegray stores meshes and edge_ids", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk())
  expect_true("meshes" %in% names(p))
  expect_true("edge_ids" %in% names(p))
  rgl::close3d()
})

test_that("print.ggsegray returns rglwidget", {
  skip_if_not_installed("rgl")

  p <- ggsegray(hemisphere = "left", atlas = dk())
  widget <- print(p)
  expect_s3_class(widget, "htmlwidget")
  rgl::close3d()
})


test_that("set_legend renders continuous legend for numeric data", {
  skip_if_not_installed("rgl")

  some_data <- dplyr::tibble(
    region = c("precentral", "postcentral"),
    p = c(0.1, 0.5)
  )

  p <- ggsegray(
    .data = some_data,
    atlas = dk(),
    colour_by = "p",
    hemisphere = "left"
  )
  expect_equal(p$legend_data$type, "continuous")
  result <- set_legend(p)
  expect_s3_class(result, "ggsegray")
  rgl::close3d()
})

test_that("ggsegray deprecated params trigger warnings", {
  skip_if_not_installed("rgl")

  lifecycle::expect_deprecated(
    p <- ggsegray(hemisphere = "left", atlas = dk(), colour = "colour")
  )
  rgl::close3d()

  lifecycle::expect_deprecated(
    p <- ggsegray(hemisphere = "left", atlas = dk(), label = "label")
  )
  rgl::close3d()

  some_data <- data.frame(
    region = c("precentral", "insula"),
    p = c(0.1, 0.5),
    stringsAsFactors = FALSE
  )
  lifecycle::expect_deprecated(
    p <- ggsegray(.data = some_data, atlas = dk(), text = "p")
  )
  rgl::close3d()
})

#' Build legend data structure
#'
#' Creates the appropriate legend data structure based on whether the
#' colour variable is numeric (continuous colorbar) or categorical
#' (discrete legend).
#'
#' @param is_numeric Whether the colour variable is numeric
#' @param data_min Minimum data value (for continuous)
#' @param data_max Maximum data value (for continuous)
#' @param palette Original palette specification
#' @param pal_colours Processed palette colours
#' @param colour_col Name of the colour column
#' @param label_col Name of the label column
#' @param fill_col Name of the fill column
#' @param data Atlas data
#'
#' @return List with legend specification or NULL
#' @keywords internal
build_legend_data <- function(
  is_numeric,
  data_min,
  data_max,
  palette,
  pal_colours,
  colour_col,
  label_col,
  fill_col,
  data
) {
  if (is_numeric && !is.na(data_min) && data_min != data_max) {
    return(build_continuous_legend(
      palette,
      pal_colours,
      colour_col,
      data_min,
      data_max
    ))
  }

  if (!is_numeric) {
    return(build_discrete_legend(data, fill_col, label_col))
  }

  NULL
}


#' Build continuous legend
#'
#' Creates a continuous colorbar legend specification for numeric data.
#'
#' @param palette Original palette specification
#' @param pal_colours Processed palette colours
#' @param colour_col Name of the colour column (used as title)
#' @param data_min Minimum data value
#' @param data_max Maximum data value
#'
#' @return List with continuous legend specification
#' @keywords internal
build_continuous_legend <- function(
  palette,
  pal_colours,
  colour_col,
  data_min,
  data_max
) {
  if (!is.null(names(palette))) {
    list(
      type = "continuous",
      title = colour_col,
      min = min(pal_colours$values),
      max = max(pal_colours$values),
      colors = unname(vapply(pal_colours$orig, col2hex, character(1))),
      breakpoints = unname(pal_colours$values)
    )
  } else {
    colorbar_values <- seq(data_min, data_max, length.out = 10)
    colorbar_colors <- scales::gradient_n_pal(
      pal_colours$orig,
      pal_colours$values,
      "Lab"
    )(colorbar_values)

    list(
      type = "continuous",
      title = colour_col,
      min = data_min,
      max = data_max,
      colors = unname(colorbar_colors),
      values = unname(colorbar_values)
    )
  }
}


#' Build discrete legend
#'
#' Creates a discrete legend specification for categorical data.
#'
#' @param data Atlas data
#' @param fill_col Name of the fill column
#' @param label_col Name of the label column
#'
#' @return List with discrete legend specification or NULL if too many
#'   categories
#' @keywords internal
build_discrete_legend <- function(data, fill_col, label_col) {
  if (is.data.frame(data)) {
    unique_values <- unique(data[[fill_col]])
  } else {
    unique_values <- unique(unlist(data[, fill_col]))
  }
  unique_values <- unique_values[!is.na(unique_values)]

  if (length(unique_values) > 50) {
    return(NULL)
  }

  if (is.data.frame(data)) {
    color_label_map <- stats::setNames(
      as.character(data[[fill_col]]),
      as.character(data[[label_col]])
    )
  } else {
    color_label_map <- stats::setNames(
      as.character(unlist(data[, fill_col])),
      as.character(unlist(data[, label_col]))
    )
  }
  color_label_map <- color_label_map[!is.na(names(color_label_map))]
  color_label_map <- color_label_map[!duplicated(names(color_label_map))]

  list(
    type = "discrete",
    title = label_col,
    labels = unname(names(color_label_map)),
    colors = unname(color_label_map)
  )
}

test_that("build_legend_data returns continuous legend for numeric data", {
  pal_colours <- get_palette(c("blue", "red"))

  result <- build_legend_data(
    is_numeric = TRUE,
    data_min = 0,
    data_max = 10,
    palette = c("blue", "red"),
    pal_colours = pal_colours,
    colour_col = "value",
    label_col = "region",
    fill_col = "new_col",
    data = data.frame()
  )

  expect_equal(result$type, "continuous")
  expect_equal(result$title, "value")
  expect_equal(result$min, 0)
  expect_equal(result$max, 10)
  expect_true(length(result$colors) > 0)
})

test_that("build_legend_data returns discrete legend for categorical data", {
  data <- data.frame(
    region = c("a", "b", "c"),
    colour = c("#FF0000", "#00FF00", "#0000FF"),
    stringsAsFactors = FALSE
  )

  result <- build_legend_data(
    is_numeric = FALSE,
    data_min = NA,
    data_max = NA,
    palette = NULL,
    pal_colours = NULL,
    colour_col = "colour",
    label_col = "region",
    fill_col = "colour",
    data = data
  )

  expect_equal(result$type, "discrete")
  expect_equal(result$title, "region")
  expect_equal(length(result$labels), 3)
  expect_equal(length(result$colors), 3)
})

test_that("build_legend_data returns NULL when data_min equals data_max", {
  result <- build_legend_data(
    is_numeric = TRUE,
    data_min = 5,
    data_max = 5,
    palette = NULL,
    pal_colours = NULL,
    colour_col = "value",
    label_col = "region",
    fill_col = "new_col",
    data = data.frame()
  )

  expect_null(result)
})

test_that("build_continuous_legend works with named palette", {
  pal_colours <- data.frame(
    values = c(0, 5, 10),
    norm = c(0, 0.5, 1),
    orig = c("blue", "white", "red"),
    stringsAsFactors = FALSE
  )

  result <- build_continuous_legend(
    palette = c("blue" = 0, "white" = 5, "red" = 10),
    pal_colours = pal_colours,
    colour_col = "my_value",
    data_min = 0,
    data_max = 10
  )

  expect_equal(result$type, "continuous")
  expect_equal(result$min, 0)
  expect_equal(result$max, 10)
  expect_true(!is.null(result$breakpoints))
})

test_that("build_continuous_legend works without named palette", {
  pal_colours <- data.frame(
    values = c(0, 0.5, 1),
    norm = c(0, 0.5, 1),
    orig = c("blue", "white", "red"),
    stringsAsFactors = FALSE
  )

  result <- build_continuous_legend(
    palette = c("blue", "white", "red"),
    pal_colours = pal_colours,
    colour_col = "my_value",
    data_min = 0,
    data_max = 100
  )

  expect_equal(result$type, "continuous")
  expect_equal(result$min, 0)
  expect_equal(result$max, 100)
  expect_true(!is.null(result$values))
  expect_equal(length(result$values), 10)
})

test_that("build_discrete_legend handles data.frame input", {
  data <- data.frame(
    region = c("frontal", "temporal", "parietal"),
    colour = c("#FF0000", "#00FF00", "#0000FF"),
    stringsAsFactors = FALSE
  )

  result <- build_discrete_legend(data, "colour", "region")

  expect_equal(result$type, "discrete")
  expect_equal(result$title, "region")
  expect_equal(length(result$labels), 3)
})

test_that("build_discrete_legend handles many unique values", {
  data <- data.frame(
    region = paste0("r", 1:60),
    colour = paste0("#", sprintf("%06X", 1:60)),
    stringsAsFactors = FALSE
  )

  result <- build_discrete_legend(data, "colour", "region")

  expect_null(result)
})

test_that("build_discrete_legend handles NA values", {
  data <- data.frame(
    region = c("a", "b", NA),
    colour = c("#FF0000", "#00FF00", NA),
    stringsAsFactors = FALSE
  )

  result <- build_discrete_legend(data, "colour", "region")

  expect_equal(result$type, "discrete")
  expect_true(all(!is.na(result$colors)))
})

test_that("build_discrete_legend handles tibble input", {
  data <- dplyr::tibble(
    region = c("a", "b"),
    colour = c("#FF0000", "#00FF00")
  )

  result <- build_discrete_legend(data, "colour", "region")

  expect_equal(result$type, "discrete")
  expect_equal(length(result$labels), 2)
})


test_that("build_discrete_legend handles non-dataframe input", {
  data <- matrix(
    c("a", "b", "#FF0000", "#00FF00"),
    ncol = 2,
    dimnames = list(NULL, c("region", "colour"))
  )

  result <- build_discrete_legend(data, "colour", "region")
  expect_equal(result$type, "discrete")
  expect_equal(length(result$labels), 2)
})

test_that("build_discrete_legend removes duplicate labels", {
  data <- data.frame(
    region = c("a", "a", "b"),
    colour = c("#FF0000", "#FF0000", "#00FF00"),
    stringsAsFactors = FALSE
  )

  result <- build_discrete_legend(data, "colour", "region")

  expect_equal(length(result$labels), 2)
})

#' ggseg3d: Plot brain segmentations in 3D
#'
#' Plotting results from analyses done on data derived from brain
#' segmentations is a common need, but may be quite laborious.
#' Results from such analyses are usually easier to interpret if
#' the plot can mimmick the shape and position in the brain it
#' represents.
#'
#' This package contains data from various brain parcellations,
#' with convenient functions to inspect the results directly on
#' a brain-plot using interactive 3D mesh visualizations powered
#' by Three.js via htmlwidgets.
#'
#' The package uses `ggseg_atlas` objects from ggseg.formats
#' that contain 3D vertex mappings.
#'
#' @name ggseg3d-package
#' @docType package
#' @keywords internal
#' @import ggseg.formats
"_PACKAGE"

#' @export
ggseg.formats::dk

#' @export
ggseg.formats::aseg

#' @export
ggseg.formats::tracula

#' Check if atlas is a unified ggseg_atlas
#'
#' Checks whether an atlas object is a ggseg_atlas with the unified format
#' (containing vertices or meshes component for 3D rendering).
#'
#' Unified atlases (ggseg_atlas class from ggseg.formats) contain:
#' - core: region info (hemi, region, label)
#' - data: type-specific data object
#'   (ggseg_data_cortical, ggseg_data_subcortical, ggseg_data_tract)
#' - palette: colours keyed by label
#'
#' @param atlas An atlas object to check
#'
#' @return Logical indicating if this is a unified ggseg_atlas
#' @noRd
#' @keywords internal
#' \dontrun{
#' is_unified_atlas(dk())
#' is_unified_atlas(ggseg.formats::dk())
#' }
is_unified_atlas <- function(atlas) {
  if (!inherits(atlas, "ggseg_atlas") && !inherits(atlas, "brain_atlas")) {
    return(FALSE)
  }

  has_core <- !is.null(atlas$core)

  if (!is.null(atlas$data) &&
        (inherits(atlas$data, "ggseg_atlas_data") ||
           inherits(atlas$data, "brain_atlas_data"))) {
    has_3d <- !is.null(atlas$data$vertices) ||
      !is.null(atlas$data$meshes) ||
      !is.null(atlas$data$centerlines)
    return(has_core && has_3d)
  }

  has_3d_data <- !is.null(atlas$vertices) || !is.null(atlas$meshes)
  has_core && has_3d_data
}


#' Prepare atlas data
#'
#' Extracts and prepares data from a ggseg_atlas object for rendering.
#' Joins vertices with core region info and palette colours.
#'
#' @param atlas A ggseg_atlas object
#' @param .data Optional user data to merge
#'
#' @return Prepared data frame with hemi, region, label, colour, and vertices
#' @keywords internal
prepare_atlas_data <- function(atlas, .data) {
  vertices <- if (!is.null(atlas$data$vertices)) {
    atlas$data$vertices
  } else {
    atlas$vertices
  }
  atlas_data <- dplyr::left_join(
    vertices,
    atlas$core,
    by = "label",
    relationship = "many-to-many"
  )

  if (!is.null(atlas$palette)) {
    atlas_data$colour <- atlas$palette[atlas_data$label]
  } else {
    atlas_data$colour <- NA_character_
  }

  if (!is.null(.data)) {
    atlas_data <- merge_atlas_data(.data, atlas_data)
  }

  atlas_data
}


#' Prepare mesh-based atlas data
#'
#' Extracts and prepares data from a mesh-based ggseg_atlas object
#' (subcortical/tract) for rendering. Joins meshes with core region info
#' and palette colours.
#'
#' @param atlas A mesh-based ggseg_atlas object
#' @param .data Optional user data to merge
#'
#' @return Prepared data frame with hemi, region, label, colour, and mesh
#' @keywords internal
prepare_mesh_atlas_data <- function(atlas, .data) {
  if (!is.null(atlas$data$centerlines)) {
    base_data <- atlas$data$centerlines[, "label", drop = FALSE]
  } else {
    base_data <- if (!is.null(atlas$data$meshes)) {
      atlas$data$meshes
    } else {
      atlas$meshes
    }
  }

  atlas_data <- dplyr::left_join(
    base_data,
    atlas$core,
    by = "label",
    relationship = "many-to-many"
  )

  if (!is.null(atlas$palette)) {
    atlas_data$colour <- atlas$palette[atlas_data$label]
  } else {
    atlas_data$colour <- NA_character_
  }

  if (!is.null(.data)) {
    atlas_data <- data_merge_mesh(.data, atlas_data)
  }

  atlas_data
}


#' Merge user data with mesh atlas data
#'
#' @param .data User-provided data frame
#' @param atlas_data Atlas data frame
#' @return Merged data frame
#' @keywords internal
data_merge_mesh <- function(.data, atlas_data) {
  join_cols <- intersect(c("region", "label", "hemi"), names(.data))

  if (length(join_cols) == 0) {
    cli::cli_warn("No common columns found for merging data with atlas")
    return(atlas_data)
  }

  dplyr::left_join(
    atlas_data,
    .data,
    by = join_cols,
    relationship = "many-to-many"
  )
}


#' Merge legend data from surface and deep cerebellar components
#' @noRd
merge_legend_data <- function(surface_legend, deep_legend) {
  if (is.null(surface_legend) && is.null(deep_legend)) return(NULL)
  if (is.null(surface_legend)) return(deep_legend)
  if (is.null(deep_legend)) return(surface_legend)

  combined <- rbind(surface_legend, deep_legend)
  combined[!duplicated(combined$label), , drop = FALSE]
}

test_that("prepare_atlas_data extracts vertices and joins core", {
  vertices_data <- data.frame(
    label = c("bankssts", "caudalanteriorcingulate"),
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(c(1, 2, 3), c(4, 5, 6))

  atlas <- structure(
    list(
      core = data.frame(
        label = c("bankssts", "caudalanteriorcingulate"),
        region = c("banks sts", "caudal anterior cingulate"),
        hemi = c("left", "left"),
        stringsAsFactors = FALSE
      ),
      vertices = vertices_data,
      palette = c("bankssts" = "#FF0000", "caudalanteriorcingulate" = "#00FF00")
    ),
    class = "ggseg_atlas"
  )

  result <- prepare_atlas_data(atlas, NULL)

  expect_true("label" %in% names(result))
  expect_true("region" %in% names(result))
  expect_true("hemi" %in% names(result))
  expect_true("colour" %in% names(result))
  expect_true("vertices" %in% names(result))
  expect_equal(result$colour[result$label == "bankssts"], "#FF0000")
})

test_that("prepare_atlas_data works with data component", {
  vertices_data <- data.frame(
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(c(1, 2), c(3, 4))

  atlas <- structure(
    list(
      core = data.frame(
        label = c("a", "b"),
        region = c("region a", "region b"),
        hemi = c("left", "right"),
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(vertices = vertices_data),
        class = "ggseg_atlas_data"
      ),
      palette = c("a" = "#AAAAAA", "b" = "#BBBBBB")
    ),
    class = "ggseg_atlas"
  )

  result <- prepare_atlas_data(atlas, NULL)

  expect_equal(nrow(result), 2)
  expect_true("vertices" %in% names(result))
})

test_that("prepare_atlas_data merges user data", {
  vertices_data <- data.frame(
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )
  vertices_data$vertices <- list(c(1, 2), c(3, 4))

  atlas <- structure(
    list(
      core = data.frame(
        label = c("a", "b"),
        region = c("region a", "region b"),
        hemi = c("left", "right"),
        stringsAsFactors = FALSE
      ),
      vertices = vertices_data,
      palette = NULL
    ),
    class = "ggseg_atlas"
  )

  user_data <- data.frame(
    label = c("a", "b"),
    my_value = c(10, 20),
    stringsAsFactors = FALSE
  )

  result <- prepare_atlas_data(atlas, user_data)

  expect_true("my_value" %in% names(result))
  expect_equal(result$my_value[result$label == "a"], 10)
})

test_that("prepare_mesh_atlas_data extracts meshes and joins core", {
  meshes_data <- data.frame(
    label = c("Left-Caudate", "Right-Caudate"),
    stringsAsFactors = FALSE
  )
  meshes_data$mesh <- list(
    list(
      vertices = data.frame(x = 1, y = 2, z = 3),
      faces = data.frame(i = 1, j = 2, k = 3)
    ),
    list(
      vertices = data.frame(x = 4, y = 5, z = 6),
      faces = data.frame(i = 4, j = 5, k = 6)
    )
  )

  atlas <- structure(
    list(
      core = data.frame(
        label = c("Left-Caudate", "Right-Caudate"),
        region = c("caudate", "caudate"),
        hemi = c("subcort", "subcort"),
        stringsAsFactors = FALSE
      ),
      meshes = meshes_data,
      palette = c("Left-Caudate" = "#123456", "Right-Caudate" = "#654321")
    ),
    class = "ggseg_atlas"
  )

  result <- prepare_mesh_atlas_data(atlas, NULL)

  expect_true("label" %in% names(result))
  expect_true("mesh" %in% names(result))
  expect_true("colour" %in% names(result))
  expect_equal(result$colour[result$label == "Left-Caudate"], "#123456")
})

test_that("prepare_mesh_atlas_data works with data component", {
  meshes_data <- data.frame(
    label = c("a"),
    stringsAsFactors = FALSE
  )
  meshes_data$mesh <- list(
    list(
      vertices = data.frame(x = 1, y = 2, z = 3),
      faces = data.frame(i = 1, j = 2, k = 3)
    )
  )

  atlas <- structure(
    list(
      core = data.frame(
        label = "a",
        region = "region a",
        hemi = "subcort",
        stringsAsFactors = FALSE
      ),
      data = structure(
        list(meshes = meshes_data),
        class = "ggseg_atlas_data"
      ),
      palette = NULL
    ),
    class = "ggseg_atlas"
  )

  result <- prepare_mesh_atlas_data(atlas, NULL)

  expect_equal(nrow(result), 1)
  expect_true("mesh" %in% names(result))
})

test_that("data_merge_mesh joins user data with atlas", {
  atlas_data <- data.frame(
    label = c("a", "b"),
    region = c("region a", "region b"),
    hemi = c("subcort", "subcort"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(list(), list())

  user_data <- data.frame(
    label = c("a", "b"),
    value = c(100, 200),
    stringsAsFactors = FALSE
  )

  result <- data_merge_mesh(user_data, atlas_data)

  expect_true("value" %in% names(result))
  expect_equal(result$value[result$label == "a"], 100)
})

test_that("data_merge_mesh warns when no common columns", {
  atlas_data <- data.frame(
    label = c("a"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(list())

  user_data <- data.frame(
    unrelated_col = c("x"),
    stringsAsFactors = FALSE
  )

  expect_warning(
    result <- data_merge_mesh(user_data, atlas_data),
    "No common columns"
  )

  expect_equal(result, atlas_data)
})

test_that("data_merge_mesh joins on region column", {
  atlas_data <- data.frame(
    label = c("a", "b"),
    region = c("region a", "region b"),
    stringsAsFactors = FALSE
  )
  atlas_data$mesh <- list(list(), list())

  user_data <- data.frame(
    region = c("region a", "region b"),
    score = c(1.5, 2.5),
    stringsAsFactors = FALSE
  )

  result <- data_merge_mesh(user_data, atlas_data)

  expect_true("score" %in% names(result))
  expect_equal(result$score[result$region == "region a"], 1.5)
})

test_that("prepare_mesh_atlas_data merges user data", {
  meshes_data <- data.frame(
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )
  meshes_data$mesh <- list(
    list(
      vertices = data.frame(x = 1, y = 2, z = 3),
      faces = data.frame(i = 1, j = 2, k = 3)
    ),
    list(
      vertices = data.frame(x = 4, y = 5, z = 6),
      faces = data.frame(i = 4, j = 5, k = 6)
    )
  )

  atlas <- structure(
    list(
      core = data.frame(
        label = c("a", "b"),
        region = c("region a", "region b"),
        hemi = c("subcort", "subcort"),
        stringsAsFactors = FALSE
      ),
      meshes = meshes_data,
      palette = NULL
    ),
    class = "ggseg_atlas"
  )

  user_data <- data.frame(
    label = c("a", "b"),
    value = c(100, 200),
    stringsAsFactors = FALSE
  )

  result <- prepare_mesh_atlas_data(atlas, user_data)

  expect_true("value" %in% names(result))
  expect_equal(result$value[result$label == "a"], 100)
})

#' Shiny bindings for ggseg3d
#'
#' Output and render functions for using ggseg3d within Shiny
#' applications and interactive R Markdown documents.
#'
#' @param outputId output variable to read from
#' @param width,height Must be a valid CSS unit (like \code{'100\%'},
#'   \code{'400px'}, \code{'auto'}) or a number, which will be coerced to a
#'   string and have \code{'px'} appended.
#' @param expr An expression that generates a ggseg3d
#' @param env The environment in which to evaluate \code{expr}.
#' @param quoted Is \code{expr} a quoted expression (with \code{quote()})? This
#'   is useful if you want to save an expression in a variable.
#'
#' @return `ggseg3dOutput` returns an HTML widget output element for use in
#'   a Shiny UI. `renderGgseg3d` returns a render function for use in a
#'   Shiny server.
#'
#' @name ggseg3d-shiny
#'
#' @importFrom htmlwidgets shinyWidgetOutput shinyRenderWidget
#'
#' @examplesIf interactive() && rlang::is_installed("shiny")
#' library(shiny)
#' ui <- fluidPage(ggseg3dOutput("brain"))
#' server <- function(input, output) {
#'   output$brain <- renderGgseg3d(ggseg3d())
#' }
#'
#' @export
# nolint start: object_name_linter
ggseg3dOutput <- function(outputId, width = "100%", height = "400px") {
  htmlwidgets::shinyWidgetOutput(
    outputId,
    "ggseg3d",
    width,
    height,
    package = "ggseg3d"
  )
}

#' @rdname ggseg3d-shiny
#' @export
renderGgseg3d <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!quoted) {
    expr <- substitute(expr)
  }
  htmlwidgets::shinyRenderWidget(expr, ggseg3dOutput, env, quoted = TRUE)
}


#' Update camera in Shiny
#'
#' Sends a message to update the camera position of a ggseg3d widget
#' in a Shiny app.
#'
#' @param session The Shiny session object
#' @param outputId The output ID of the ggseg3d widget
#' @param camera Camera position preset or custom position
#'
#' @return None, called for side effects (sends message to client)
#'
#' @examplesIf interactive() && rlang::is_installed("shiny")
#' \dontrun{
#' updateGgseg3dCamera(session, "brain", "left lateral")
#' }
#'
#' @export
updateGgseg3dCamera <- function(session, outputId, camera) {
  session$sendCustomMessage(paste0("ggseg3d-camera-", outputId), camera)
}


#' Update background in Shiny
#'
#' Sends a message to update the background color of a ggseg3d widget
#' in a Shiny app.
#'
#' @param session The Shiny session object
#' @param outputId The output ID of the ggseg3d widget
#' @param colour Background color (hex or named color)
#'
#' @return None, called for side effects (sends message to client)
#'
#' @examplesIf interactive() && rlang::is_installed("shiny")
#' \dontrun{
#' updateGgseg3dBackground(session, "brain", "black")
#' }
#'
#' @export
updateGgseg3dBackground <- function(session, outputId, colour) {
  if (!grepl("^#", colour)) {
    colour <- col2hex(colour)
  }
  session$sendCustomMessage(paste0("ggseg3d-background-", outputId), colour)
}
# nolint end

test_that("ggseg3dOutput creates shiny output", {
  output <- ggseg3dOutput("brain_plot")

  expect_s3_class(output, "shiny.tag.list")
})

test_that("ggseg3dOutput accepts custom dimensions", {
  output <- ggseg3dOutput("brain_plot", width = "800px", height = "600px")

  expect_s3_class(output, "shiny.tag.list")
})

test_that("renderGgseg3d creates render function", {
  render_fn <- renderGgseg3d({
    ggseg3d()
  })

  expect_true(is.function(render_fn))
})

test_that("renderGgseg3d with quoted expression", {
  expr <- quote(ggseg3d())
  render_fn <- renderGgseg3d(expr, quoted = TRUE)

  expect_true(is.function(render_fn))
})

test_that("updateGgseg3dCamera builds correct message type", {
  mock_session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )

  result <- mock_session$sendCustomMessage(
    paste0("ggseg3d-camera-", "brain_plot"),
    "left lateral"
  )

  expect_equal(result$type, "ggseg3d-camera-brain_plot")
  expect_equal(result$message, "left lateral")
})

test_that("updateGgseg3dBackground converts named color to hex", {
  converted <- col2hex("red")
  expect_equal(converted, "#FF0000")

  converted <- col2hex("blue")
  expect_equal(converted, "#0000FF")
})

test_that("updateGgseg3dCamera function exists and is exported", {
  expect_true(is.function(updateGgseg3dCamera))
})

test_that("updateGgseg3dBackground function exists and is exported", {
  expect_true(is.function(updateGgseg3dBackground))
})

test_that("updateGgseg3dCamera calls sendCustomMessage", {
  messages <- list()
  mock_session <- list(
    sendCustomMessage = function(type, message) {
      messages <<- list(type = type, message = message)
    }
  )

  updateGgseg3dCamera(mock_session, "test_output", "left lateral")

  expect_equal(messages$type, "ggseg3d-camera-test_output")
  expect_equal(messages$message, "left lateral")
})

test_that("updateGgseg3dBackground calls sendCustomMessage with hex color", {
  messages <- list()
  mock_session <- list(
    sendCustomMessage = function(type, message) {
      messages <<- list(type = type, message = message)
    }
  )

  updateGgseg3dBackground(mock_session, "test_output", "red")

  expect_equal(messages$type, "ggseg3d-background-test_output")
  expect_equal(messages$message, "#FF0000")
})

test_that("updateGgseg3dBackground passes hex color through", {
  messages <- list()
  mock_session <- list(
    sendCustomMessage = function(type, message) {
      messages <<- list(type = type, message = message)
    }
  )

  updateGgseg3dBackground(mock_session, "test_output", "#AABBCC")

  expect_equal(messages$message, "#AABBCC")
})

#' Apply colour palette
#'
#' Processes colour mapping for ggseg_atlas objects using
#' vertex-based colouring.
#'
#' @param atlas_data Atlas data frame with vertices column
#' @param colour Column name for colour values
#' @param palette Colour palette specification
#' @param na_colour Colour for NA values
#'
#' @return List with data, fill column name, palette, and colour metadata
#' @keywords internal
apply_colour_palette <- function(
  atlas_data,
  colour,
  palette,
  na_colour
) {
  pal_colours <- get_palette(palette)
  is_numeric <- colour %in%
    names(atlas_data) &&
    is.numeric(atlas_data[[colour]])
  data_min <- NA
  data_max <- NA

  if (is_numeric) {
    data_min <- min(atlas_data[[colour]], na.rm = TRUE)
    data_max <- max(atlas_data[[colour]], na.rm = TRUE)

    if (data_min == data_max) {
      atlas_data$new_col <- pal_colours$orig[1]
    } else {
      if (is.null(names(palette))) {
        pal_colours$values <- seq(
          data_min,
          data_max,
          length.out = nrow(pal_colours)
        )
      }
      atlas_data$new_col <- scales::gradient_n_pal(
        pal_colours$orig,
        pal_colours$values,
        "Lab"
      )(atlas_data[[colour]])
    }
    fill <- "new_col"
  } else {
    fill <- colour
  }

  atlas_data$colour <- vapply(
    atlas_data[[fill]],
    function(c) {
      if (is.na(c)) {
        na_colour
      } else if (grepl("^#", c)) {
        c
      } else {
        col2hex(c)
      }
    },
    character(1)
  )

  list(
    data = atlas_data,
    fill = fill,
    palette = pal_colours,
    is_numeric = is_numeric,
    data_min = data_min,
    data_max = data_max
  )
}

test_that("apply_colour_palette handles numeric data", {
  atlas_data <- data.frame(
    label = c("a", "b", "c"),
    my_value = c(1, 2, 3),
    stringsAsFactors = FALSE
  )

  result <- apply_colour_palette(
    atlas_data,
    "my_value",
    c("blue", "red"),
    "#CCCCCC"
  )

  expect_true(result$is_numeric)
  expect_equal(result$fill, "new_col")
  expect_equal(result$data_min, 1)
  expect_equal(result$data_max, 3)
  expect_true(all(grepl("^#", result$data$colour)))
})

test_that("apply_colour_palette handles categorical data", {
  atlas_data <- data.frame(
    label = c("a", "b", "c"),
    colour = c("#FF0000", "#00FF00", "#0000FF"),
    stringsAsFactors = FALSE
  )

  result <- apply_colour_palette(atlas_data, "colour", NULL, "#CCCCCC")

  expect_false(result$is_numeric)
  expect_equal(result$fill, "colour")
})

test_that("apply_colour_palette handles NA values", {
  atlas_data <- data.frame(
    label = c("a", "b", "c"),
    colour = c("#FF0000", NA, "#0000FF"),
    stringsAsFactors = FALSE
  )

  result <- apply_colour_palette(atlas_data, "colour", NULL, "#AABBCC")

  expect_equal(result$data$colour[2], "#AABBCC")
})

test_that("apply_colour_palette converts named colors to hex", {
  atlas_data <- data.frame(
    label = c("a", "b"),
    colour = c("red", "blue"),
    stringsAsFactors = FALSE
  )

  result <- apply_colour_palette(atlas_data, "colour", NULL, "#CCCCCC")

  expect_equal(result$data$colour[1], "#FF0000")
  expect_equal(result$data$colour[2], "#0000FF")
})

test_that("apply_colour_palette handles single constant value", {
  atlas_data <- data.frame(
    label = c("a", "b"),
    my_value = c(5, 5),
    stringsAsFactors = FALSE
  )

  result <- apply_colour_palette(
    atlas_data,
    "my_value",
    c("blue", "red"),
    "#CCCCCC"
  )

  expect_true(result$is_numeric)
  expect_equal(result$data_min, result$data_max)
})

test_that("data_merge works with unified atlas", {
  atlas_data <- ggseg.formats::atlas_sf(dk())

  some_data <- data.frame(
    region = c(
      "transverse temporal",
      "insula",
      "precentral",
      "superior parietal"
    ),
    p = sample(seq(0, .5, .001), 4),
    stringsAsFactors = FALSE
  )

  merged <- dplyr::left_join(atlas_data, some_data, by = "region")

  expect_true("p" %in% names(merged))
  expect_true("region" %in% names(merged))
  expect_true("hemi" %in% names(merged))
})

#!/bin/sh

# Print usage if less than two arguments given
if [ "$#" -ne 1 ]; then
cat <<EOU
Batch convert surface and DPV (curvature) files from a given
subject to ASCII format. FreeSurfer must be correctly
configured in the memory, with both \${FREESURFER_HOME}
and \${SUBJECTS_DIR} variables set.

Usage:
  fsaverage5_2ascii <outdir>

  _____________________________________
Anderson M. Winkler
Yale University / Institute of Living
Jan/2011 (original version)
Jul/2013 (this version)
http://brainder.org

adapted by: Athanasia Monika Mowinckel for ggseg3d R-package
(March/2020)
EOU
exit
fi

# outdir
OUTDIR=$1
mkdir -p $OUTDIR

# Hemispheres
HEMI="lh rh"

# Surfaces (up to v5.3.0; increase the list as needed for future versions)
SURF="inflated orig pial sphere"

# Simplify a bit with a shorter variable
SDIR=${SUBJECTS_DIR}/fsaverage5

# For each hemisphere
for h in ${HEMI} ; do

# For each surface file
for s in ${SURF} ; do

if [ -e ${SDIR}/surf/${h}.${s} ] ; then
echo "${SDIR}/surf/${h}.${s} -> ${OUTDIR}/${h}.${s}.srf"
${FREESURFER_HOME}/bin/mris_convert ${SDIR}/surf/${h}.${s} ${OUTDIR}/${h}.${s}.asc
fi # end if surf exists
done # done s

done # done h

---
title: "Introduction to ggseg3d"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to ggseg3d}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: intro-setup
#| include: false
knitr::opts_chunk$set(
  message = FALSE,
  eval = interactive()
)
library(ggseg3d)
library(dplyr)
dir.create("img", showWarnings = FALSE)
```

Brain segmentation results are easier to interpret when they look like a brain.
ggseg3d takes a data frame with region names and renders it as an interactive 3D mesh — rotate, zoom, hover for labels — right in the browser.

## Getting started

```{r}
#| label: intro-basic
#| eval: false
library(ggseg3d)

ggseg3d(hemisphere = "left") |>
  pan_camera("left lateral")
```

```{r}
#| label: intro-snap-basic
#| echo: false
ggseg3d(hemisphere = "left") |>
  pan_camera("left lateral") |>
  snapshot_brain("img/intro-basic.png")
```

![Left hemisphere of the Desikan-Killiany atlas, each cortical region in a distinct colour, viewed from the lateral side.](img/intro-basic.png "Default DK atlas, left lateral view")

With no arguments, `ggseg3d()` plots the Desikan-Killiany atlas.
The output is an htmlwidget: click and drag to rotate, scroll to zoom, hover to see region names.

## Mapping data

Match your data to the atlas by region name, then point `colour_by` at the variable you care about:

```{r}
#| label: intro-plot-data
#| eval: false
library(dplyr)

some_data <- tibble(
  region = c("precentral", "postcentral", "insula", "superior parietal"),
  p = c(0.01, 0.04, 0.2, 0.5)
)

ggseg3d(.data = some_data, atlas = dk(), colour_by = "p", text_by = "p") |>
  pan_camera("right lateral")
```

```{r}
#| label: intro-snap-plot-data
#| echo: false
some_data <- tibble(
  region = c("precentral", "postcentral", "insula", "superior parietal"),
  p = c(0.01, 0.04, 0.2, 0.5)
)

ggseg3d(.data = some_data, atlas = dk(), colour_by = "p", text_by = "p") |>
  pan_camera("right lateral") |>
  snapshot_brain("img/intro-plot-data.png")
```

![Right hemisphere with four cortical regions coloured on a blue-to-red scale by p-value; unmatched regions shown in grey.](img/intro-plot-data.png "Four regions mapped by p-value")

`text_by = "p"` adds the p-value to the hover tooltip so you can inspect individual regions without a separate table.

## Subcortical atlases

Cortical surfaces are only half the story.
The `aseg` atlas covers subcortical structures, and `add_glassbrain()` wraps them in a translucent cortex for anatomical context:

```{r}
#| label: intro-subcortical
#| eval: false
subcort_data <- tibble(
  region = c("Thalamus", "Caudate", "Hippocampus"),
  p = c(0.2, 0.5, 0.8)
)

ggseg3d(.data = subcort_data, atlas = aseg(), colour_by = "p", na_alpha = .5) |>
  add_glassbrain()
```

```{r}
#| label: intro-snap-subcortical
#| echo: false
subcort_data <- tibble(
  region = c("Thalamus", "Caudate", "Hippocampus"),
  p = c(0.2, 0.5, 0.8)
)

ggseg3d(.data = subcort_data, atlas = aseg(), colour_by = "p", na_alpha = .5) |>
  add_glassbrain() |>
  snapshot_brain("img/intro-subcortical.png")
```

![Thalamus, caudate, and hippocampus coloured by p-value, visible through a semi-transparent cortical shell.](img/intro-subcortical.png "Subcortical structures inside a glass brain")

## Camera and background

`pan_camera()` sets the viewing angle.
`set_background()` changes the canvas colour — handy for dark-themed slides:

```{r}
#| label: intro-camera
#| eval: false
ggseg3d(hemisphere = "left") |>
  pan_camera("left lateral") |>
  set_background("black")
```

```{r}
#| label: intro-snap-camera
#| echo: false
ggseg3d(hemisphere = "left") |>
  pan_camera("left lateral") |>
  set_background("black") |>
  snapshot_brain("img/intro-background.png")
```

![Left hemisphere on a black background, cortical regions visible in their default atlas colours.](img/intro-background.png "Dark background for slides or posters")

Camera presets cover the standard anatomical views: `"left lateral"`, `"right medial"`, `"left superior"`, and so on.
For anything else, pass a list with `eye` coordinates.

---
title: "Using ggseg3d in Shiny"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using ggseg3d in Shiny}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: setup
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

ggseg3d widgets work in Shiny applications.
This vignette covers the output and render functions, plus reactive updates.

```{r}
#| label: load-packages
library(shiny)
library(ggseg3d)
```

## Basic Shiny integration

Use `ggseg3dOutput()` in the UI and `renderGgseg3d()` in the server:

```{r}
#| label: basic-app
ui <- fluidPage(
  titlePanel("Brain Atlas Viewer"),
  sidebarLayout(
    sidebarPanel(
      selectInput(
        "hemi",
        "Hemisphere",
        choices = c("left", "right", "both"),
        selected = "left"
      )
    ),
    mainPanel(
      ggseg3dOutput("brain", height = "500px")
    )
  )
)

server <- function(input, output, session) {
  output$brain <- renderGgseg3d({
    hemi <- if (input$hemi == "both") NULL else input$hemi

    ggseg3d(hemisphere = hemi) |>
      pan_camera("left lateral")
  })
}

shinyApp(ui, server)
```

## Output sizing

Set explicit dimensions in `ggseg3dOutput()`:

```{r}
#| label: output-sizing
ggseg3dOutput("brain", width = "100%", height = "600px")
```

Or use `set_dimensions()` in the render function:

```{r}
#| label: render-dimensions
renderGgseg3d({
  ggseg3d() |>
    set_dimensions(width = 800, height = 600)
})
```

## Reactive data

Update the brain plot when data changes:

```{r}
#| label: reactive-data
server <- function(input, output, session) {
  brain_data <- reactive({
    tibble(
      region = c("precentral", "postcentral", "insula"),
      value = runif(3)
    )
  })

  output$brain <- renderGgseg3d({
    ggseg3d(
      .data = brain_data(),
      atlas = dk(), # nolint [object_usage_linter]
      colour_by = "value"
    ) |>
      pan_camera("left lateral")
  })
}
```

## Updating camera and background

Use `updateGgseg3dCamera()` and `updateGgseg3dBackground()` to modify an existing widget without re-rendering:

```{r}
#| label: reactive-updates
ui <- fluidPage(
  sidebarLayout(
    sidebarPanel(
      selectInput(
        "view",
        "View",
        choices = c(
          "left lateral",
          "left medial",
          "right lateral",
          "right medial"
        )
      ),
      selectInput("bg", "Background", choices = c("white", "black", "grey"))
    ),
    mainPanel(
      ggseg3dOutput("brain")
    )
  )
)

server <- function(input, output, session) {
  output$brain <- renderGgseg3d({
    ggseg3d() |>
      pan_camera("left lateral")
  })

  observeEvent(input$view, {
    updateGgseg3dCamera("brain", input$view)
  })

  observeEvent(input$bg, {
    updateGgseg3dBackground("brain", input$bg)
  })
}
```

These updates are faster than re-rendering the entire widget.

## Complete example

Here's a full app with data selection and dynamic updates:

```{r}
#| label: complete-app
library(shiny)
library(ggseg3d)
library(dplyr)

example_data <- tibble(
  region = c(
    "precentral",
    "postcentral",
    "insula",
    "superior parietal",
    "inferior parietal",
    "supramarginal",
    "cuneus",
    "pericalcarine"
  ),
  thickness = c(2.5, 2.3, 3.1, 2.2, 2.4, 2.6, 1.8, 1.9),
  volume = c(8500, 7200, 6800, 9100, 8800, 7500, 4200, 3800)
)

ui <- fluidPage(
  titlePanel("Brain Metrics Explorer"),
  sidebarLayout(
    sidebarPanel(
      selectInput("metric", "Metric", choices = c("thickness", "volume")),
      selectInput(
        "view",
        "Camera View",
        choices = c(
          "left lateral",
          "left medial",
          "right lateral",
          "right medial",
          "left superior",
          "left inferior"
        )
      ),
      checkboxInput("edges", "Show edges", value = FALSE),
      selectInput("bg", "Background", choices = c("white", "black", "grey90"))
    ),
    mainPanel(
      ggseg3dOutput("brain", height = "600px")
    )
  )
)

server <- function(input, output, session) {
  output$brain <- renderGgseg3d({
    p <- ggseg3d(
      .data = example_data,
      atlas = dk(), # nolint [object_usage_linter]
      colour_by = input$metric,
      text_by = input$metric
    ) |>
      pan_camera(input$view) |>
      set_background(input$bg)

    if (input$edges) {
      p <- p |> set_edges("black")
    }

    p
  })
}

shinyApp(ui, server)
```

## Performance tips

**Minimize re-renders**: Use `updateGgseg3dCamera()` and `updateGgseg3dBackground()` for camera and background changes instead of re-rendering.

**Debounce reactive data**: If data updates frequently, use `debounce()` to avoid excessive re-renders.

**Pre-compute data**: Do data transformations outside the render function when possible.

**Limit regions**: Showing fewer regions renders faster.
Filter your atlas if you only need specific structures.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggseg3d.R
\name{prepare_brain_meshes}
\alias{prepare_brain_meshes}
\alias{prepare_brain_meshes.cortical_atlas}
\alias{prepare_brain_meshes.subcortical_atlas}
\alias{prepare_brain_meshes.cerebellar_atlas}
\alias{prepare_brain_meshes.tract_atlas}
\title{Prepare brain meshes and legend data}
\usage{
prepare_brain_meshes(atlas, ...)

\method{prepare_brain_meshes}{cortical_atlas}(
  atlas,
  .data = NULL,
  surface = "LCBC",
  hemisphere = c("right", "left"),
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  edge_by = NULL,
  brain_meshes = NULL,
  ...
)

\method{prepare_brain_meshes}{subcortical_atlas}(
  atlas,
  .data = NULL,
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  ...
)

\method{prepare_brain_meshes}{cerebellar_atlas}(
  atlas,
  .data = NULL,
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  surface_opacity = NULL,
  ...
)

\method{prepare_brain_meshes}{tract_atlas}(
  atlas,
  .data = NULL,
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  tract_color = c("palette", "orientation"),
  tube_radius = 2,
  tube_segments = 10,
  ...
)
}
\arguments{
\item{atlas}{A `ggseg_atlas` object}

\item{...}{Type-specific arguments passed to methods}

\item{.data}{Optional user data to merge}

\item{surface}{Surface type: `"inflated"` (default), `"semi-inflated"`,
`"white"`, `"pial"`. Use `"LCBC"` as alias for `"inflated"`.}

\item{hemisphere}{Character vector of hemispheres: `"right"`, `"left"`.}

\item{label_by}{Column name for region hover labels}

\item{text_by}{Column name for extra hover text}

\item{colour_by}{Column name for colour values}

\item{palette}{Colour palette specification}

\item{na_colour}{Colour for NA values}

\item{na_alpha}{Transparency for NA regions}

\item{edge_by}{Column name for region boundary edge grouping}

\item{brain_meshes}{Optional user-supplied brain meshes}

\item{surface_opacity}{Numeric opacity for the cerebellar surface mesh.
Defaults to `0.3` when deep nuclei are present and `1` otherwise.}

\item{tract_color}{`"palette"` (default) or `"orientation"`
(direction-based RGB colouring)}

\item{tube_radius}{Numeric tube radius (default 5 when `NULL`).}

\item{tube_segments}{Integer tube segment count (default 8 when `NULL`).}
}
\value{
List with `meshes` (list of mesh entries) and `legend_data`
}
\description{
S3 generic that dispatches to atlas-type-specific preparation methods.
Builds mesh data structures and legend data from a `ggseg_atlas`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{build_discrete_legend}
\alias{build_discrete_legend}
\title{Build discrete legend}
\usage{
build_discrete_legend(data, fill_col, label_col)
}
\arguments{
\item{data}{Atlas data}

\item{fill_col}{Name of the fill column}

\item{label_col}{Name of the label column}
}
\value{
List with discrete legend specification or NULL if too many
  categories
}
\description{
Creates a discrete legend specification for categorical data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{build_tract_meshes}
\alias{build_tract_meshes}
\title{Build mesh list for tract atlases}
\usage{
build_tract_meshes(
  atlas_data,
  na_colour,
  color_by = "colour",
  atlas_centerlines = NULL,
  text_by = NULL,
  label_by = "region"
)
}
\arguments{
\item{atlas_data}{Prepared atlas data frame with label, colour, and mesh
columns}

\item{na_colour}{Colour for NA values}

\item{color_by}{How to colour tracts: "colour" (use colour column),
"orientation" (direction-based RGB from tangents)}
}
\value{
List of mesh data structures
}
\description{
Creates mesh data structures for tract atlases with per-vertex colouring.
Supports palette colours (uniform per tract) or orientation-based RGB
colours computed from centerline tangent vectors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_edges}
\alias{set_edges}
\title{Set region boundary edges}
\usage{
set_edges(p, colour = "black", width = 1)
}
\arguments{
\item{p}{A `ggseg3d` widget or `ggsegray` rgl object.}

\item{colour}{string. Edge colour (hex or named color). Set to NULL to
hide edges.}

\item{width}{numeric. Width of edge lines (default: 1). Note: line width > 1
may not render on all systems due to WebGL limitations.}
}
\value{
The input object (modified), for piping.
}
\description{
Adds coloured outlines around brain regions. This is useful for
highlighting region boundaries in figures. Works with both
htmlwidget (`ggseg3d`) and rgl (`ggsegray`) objects. For rgl,
edges must have been computed at creation time via `edge_by`.
}
\section{Lifecycle}{

`r lifecycle::badge("experimental")`
}

\examples{
ggseg3d(hemisphere = "left", edge_by = "region") |>
  set_edges("black") |>
  pan_camera("left lateral")
\dontrun{
# rgl requires OpenGL; not run in check environments.
ggsegray(hemisphere = "left", edge_by = "region") |>
  set_edges("red", width = 2) |>
  pan_camera("left lateral")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atlas-types.R
\name{data_merge_mesh}
\alias{data_merge_mesh}
\title{Merge user data with mesh atlas data}
\usage{
data_merge_mesh(.data, atlas_data)
}
\arguments{
\item{.data}{User-provided data frame}

\item{atlas_data}{Atlas data frame}
}
\value{
Merged data frame
}
\description{
Merge user data with mesh atlas data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_positioning}
\alias{set_positioning}
\title{Set hemisphere positioning mode}
\usage{
set_positioning(p, positioning = c("anatomical", "centered"))
}
\arguments{
\item{p}{ggseg3d widget object}

\item{positioning}{How to position hemispheres:
- "anatomical": Medial surfaces adjacent at midline. Left at negative
  x, right at positive x. Default for displaying both hemispheres
  together.
- "centered": Centre each hemisphere at the origin. Best for
  single-hemisphere snapshots where consistent sizing is needed.}
}
\value{
ggseg3d widget object with repositioned meshes
}
\description{
Repositions cortical hemisphere meshes in a ggseg3d widget. Meshes are
produced with anatomical positioning by default (medial edges at x = 0).
Use this to centre each hemisphere at the origin for atlas-creation
snapshots, or to reapply anatomical positioning after manual edits.
}
\details{
Only cortical hemisphere meshes are repositioned (those named
  `"<hemi> <surface>"`, e.g. `"left inflated"`). Subcortical, cerebellar
  and glassbrain meshes are left untouched.
}
\examples{
# View both hemispheres anatomically positioned (default)
ggseg3d(hemisphere = c("left", "right")) |>
  pan_camera("left lateral")

# Atlas creation: centered for consistent sizing
ggseg3d(hemisphere = "left") |>
  set_positioning("centered") |>
  set_orthographic() |>
  pan_camera("left lateral")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{cross_product}
\alias{cross_product}
\title{Cross product of two 3D vectors}
\usage{
cross_product(a, b)
}
\description{
Cross product of two 3D vectors
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_dimensions}
\alias{set_dimensions}
\title{Set widget dimensions}
\usage{
set_dimensions(p, width = NULL, height = NULL)
}
\arguments{
\item{p}{ggseg3d widget object}

\item{width}{numeric. Widget width in pixels (NULL for default)}

\item{height}{numeric. Widget height in pixels (NULL for default)}
}
\value{
ggseg3d widget object with updated dimensions
}
\description{
Changes the width and height of a ggseg3d widget.
}
\examples{
ggseg3d() |>
  set_dimensions(width = 800, height = 600)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggseg3d.R
\name{ggseg3d}
\alias{ggseg3d}
\title{Plot 3D brain parcellations}
\usage{
ggseg3d(
  .data = NULL,
  atlas = dk(),
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  ...,
  label = deprecated(),
  text = deprecated(),
  colour = deprecated()
)
}
\arguments{
\item{.data}{A data.frame to use for plot aesthetics. Must include a
column called "region" corresponding to regions.}

\item{atlas}{A `ggseg_atlas` object containing 3D vertex mappings, or a
string naming an atlas function (deprecated).}

\item{label_by}{String. Column name used as hover label for each region.}

\item{text_by}{String. Column name for extra hover text shown below the
region label.}

\item{colour_by}{String. Column name mapped to mesh colours.}

\item{palette}{String. Vector of colour names or HEX colours. Can also
be a named numeric vector, with colours as names, and breakpoint for
that colour as the value}

\item{na_colour}{String. Either name, hex of RGB for colour of NA in
colour.}

\item{na_alpha}{Numeric. A number between 0 and 1 to control
transparency of NA-regions.}

\item{...}{Type-specific arguments passed to the atlas method.
See section **Type-specific arguments** below.}

\item{label, text, colour}{`r lifecycle::badge("deprecated")` Use
`label_by`, `text_by`, and `colour_by` instead.}
}
\value{
an htmlwidget object for interactive 3D brain visualization
}
\description{
`ggseg3d` creates and returns an interactive Three.js brain mesh
visualization. Dispatches to atlas-type-specific methods via
[prepare_brain_meshes()].
}
\section{Type-specific arguments}{

Cortical atlases (`cortical_atlas`):
\describe{
  \item{`surface`}{Surface type: `"LCBC"` (default, alias for
    inflated), `"inflated"`, `"semi-inflated"`, `"white"`, `"pial"`.}
  \item{`hemisphere`}{Character vector of hemispheres: `"right"`,
    `"left"`.}
  \item{`edge_by`}{Column name for region boundary edges.}
  \item{`brain_meshes`}{Custom brain mesh data.}
}

Tract atlases (`tract_atlas`):
\describe{
  \item{`tract_color`}{`"palette"` (default) or `"orientation"`
    (direction-based RGB).}
  \item{`tube_radius`}{Tube radius (numeric, default 5).}
  \item{`tube_segments`}{Tube segment count (integer, default 8).}
}
}

\examples{
ggseg3d()
ggseg3d(hemisphere = "left") |> pan_camera("left lateral")
ggseg3d() |> set_legend(FALSE)
ggseg3d() |> set_background("black")

}
\seealso{
[pan_camera()] for camera position, [set_background()] for
  background colour, [set_legend()] for legend visibility
}
\author{
Athanasia Mowinckel and Didac Piñeiro
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{build_cerebellar_meshes}
\alias{build_cerebellar_meshes}
\title{Build mesh list for cerebellar atlases}
\usage{
build_cerebellar_meshes(
  atlas_data,
  na_colour,
  text_by = NULL,
  label_by = "region",
  opacity = 1
)
}
\arguments{
\item{atlas_data}{Prepared atlas data frame with vertices column}

\item{na_colour}{Colour for NA values}

\item{text_by}{Column for hover text (or NULL)}

\item{label_by}{Column for vertex labels}

\item{opacity}{Numeric opacity for the mesh (0 = transparent, 1 = opaque)}
}
\value{
List of mesh data structures
}
\description{
Creates mesh data structures for cerebellar ggseg_atlas objects using
the shared SUIT cerebellar surface with vertex-based colouring.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rayshader.R
\name{camera_preset_to_position}
\alias{camera_preset_to_position}
\title{Map camera preset name to position vector}
\usage{
camera_preset_to_position(preset)
}
\arguments{
\item{preset}{Character string naming the camera preset.}
}
\value{
Numeric vector of length 3 (x, y, z).
}
\description{
Converts a camera preset string to an xyz position vector matching
the same presets used in the Three.js viewer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rayshader.R
\name{look_at_origin}
\alias{look_at_origin}
\title{Compute rgl rotation matrix to look at the origin from a given position}
\usage{
look_at_origin(eye)
}
\arguments{
\item{eye}{Numeric vector of length 3 (x, y, z) — camera position.}
}
\value{
A 4x4 rotation matrix.
}
\description{
Builds a 4x4 rotation matrix suitable for `rgl::view3d(userMatrix = ...)`
that orients the scene as if the camera is at `eye` looking toward the
origin with z pointing up.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggseg3d.R
\name{apply_colours_and_legend}
\alias{apply_colours_and_legend}
\title{Apply colour palette and build legend data}
\usage{
apply_colours_and_legend(atlas_data, colour_by, palette, na_colour, label_by)
}
\arguments{
\item{atlas_data}{Prepared atlas data frame}

\item{colour_by}{Column name for colour values}

\item{palette}{Colour palette specification}

\item{na_colour}{Colour for NA values}

\item{label_by}{Column name for labels}
}
\value{
List with `atlas_data` and `legend_data`
}
\description{
Shared pipeline step for all atlas types: applies colour palette to
atlas data and builds the legend data structure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{build_legend_data}
\alias{build_legend_data}
\title{Build legend data structure}
\usage{
build_legend_data(
  is_numeric,
  data_min,
  data_max,
  palette,
  pal_colours,
  colour_col,
  label_col,
  fill_col,
  data
)
}
\arguments{
\item{is_numeric}{Whether the colour variable is numeric}

\item{data_min}{Minimum data value (for continuous)}

\item{data_max}{Maximum data value (for continuous)}

\item{palette}{Original palette specification}

\item{pal_colours}{Processed palette colours}

\item{colour_col}{Name of the colour column}

\item{label_col}{Name of the label column}

\item{fill_col}{Name of the fill column}

\item{data}{Atlas data}
}
\value{
List with legend specification or NULL
}
\description{
Creates the appropriate legend data structure based on whether the
colour variable is numeric (continuous colorbar) or categorical
(discrete legend).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{compute_parallel_transp_fr}
\alias{compute_parallel_transp_fr}
\title{Compute parallel transport frames along curve}
\usage{
compute_parallel_transp_fr(curve)
}
\description{
Compute parallel transport frames along curve
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rayshader.R
\name{ggsegray}
\alias{ggsegray}
\title{Render brain atlas with rgl}
\usage{
ggsegray(
  .data = NULL,
  atlas = dk(),
  label_by = "region",
  text_by = NULL,
  colour_by = "colour",
  palette = NULL,
  na_colour = "darkgrey",
  na_alpha = 1,
  material = list(),
  ...,
  label = deprecated(),
  text = deprecated(),
  colour = deprecated()
)
}
\arguments{
\item{.data}{A data.frame to use for plot aesthetics. Must include a
column called "region" corresponding to regions.}

\item{atlas}{A `ggseg_atlas` object containing 3D vertex mappings, or a
string naming an atlas function (deprecated).}

\item{label_by}{String. Column name used as hover label for each region.}

\item{text_by}{String. Column name for extra hover text shown below the
region label.}

\item{colour_by}{String. Column name mapped to mesh colours.}

\item{palette}{String. Vector of colour names or HEX colours. Can also
be a named numeric vector, with colours as names, and breakpoint for
that colour as the value}

\item{na_colour}{String. Either name, hex of RGB for colour of NA in
colour.}

\item{na_alpha}{Numeric. A number between 0 and 1 to control
transparency of NA-regions.}

\item{material}{Named list of rgl material properties passed to
[rgl::tmesh3d()]. Controls how the mesh surface is shaded.}

\item{...}{Type-specific arguments passed to the atlas method.
See section **Type-specific arguments** below.}

\item{label, text, colour}{`r lifecycle::badge("deprecated")` Use
`label_by`, `text_by`, and `colour_by` instead.}
}
\value{
An object of class `ggsegray` (invisibly), which wraps the
  rgl device ID. Pipe into [pan_camera()], [add_glassbrain()], or
  [set_background()] to modify the scene.
}
\description{
Creates an rgl 3D scene from a brain atlas. Uses the same atlas
preparation pipeline as [ggseg3d()] but outputs to rgl instead of
htmlwidgets. The resulting scene can be piped into [pan_camera()],
[add_glassbrain()], and [set_background()], then rendered with
rayshader's `render_highquality()` or captured with `rgl::snapshot3d()`.
}
\section{Material properties}{

Useful material list entries:
\describe{
  \item{`specular`}{`"black"` (matte) or `"white"` (glossy).}
  \item{`shininess`}{Specular exponent. Higher = tighter highlights.}
  \item{`lit`}{`FALSE` disables lighting.}
  \item{`alpha`}{Transparency, 0 (invisible) to 1 (opaque).}
  \item{`smooth`}{`TRUE` for Gouraud shading, `FALSE` for flat.}
}

See [rgl::material3d()] for the full list.
}

\section{Type-specific arguments}{

Cortical atlases (`cortical_atlas`):
\describe{
  \item{`surface`}{Surface type: `"LCBC"` (default, alias for
    inflated), `"inflated"`, `"semi-inflated"`, `"white"`, `"pial"`.}
  \item{`hemisphere`}{Character vector of hemispheres: `"right"`,
    `"left"`.}
  \item{`edge_by`}{Column name for region boundary edges.}
  \item{`brain_meshes`}{Custom brain mesh data.}
}

Tract atlases (`tract_atlas`):
\describe{
  \item{`tract_color`}{`"palette"` (default) or `"orientation"`
    (direction-based RGB).}
  \item{`tube_radius`}{Tube radius (numeric, default 5).}
  \item{`tube_segments`}{Tube segment count (integer, default 8).}
}
}

\examples{
\dontrun{
# rgl requires OpenGL; not run in check environments.
ggsegray(hemisphere = "left") |>
  pan_camera("left lateral")

ggsegray(atlas = aseg()) |>
  add_glassbrain(opacity = 0.15) |>
  pan_camera("right lateral") |>
  set_background("black")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rayshader.R
\name{mesh_entry_to_mesh3d}
\alias{mesh_entry_to_mesh3d}
\title{Convert mesh entry to rgl mesh3d object}
\usage{
mesh_entry_to_mesh3d(mesh_entry, ...)
}
\arguments{
\item{mesh_entry}{A mesh entry list with vertices, faces, colors,
colorMode, and opacity.}

\item{...}{Material properties merged into the `material` list of
[rgl::tmesh3d()]. Overrides defaults (`specular = "black"`,
`shininess = 128`). See [rgl::material3d()] for all options.}
}
\value{
An rgl `mesh3d` object
}
\description{
Converts the internal mesh_entry list structure (as built by
[make_mesh_entry()]) into an [rgl::tmesh3d()] object for rgl rendering.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_flat_shading}
\alias{set_flat_shading}
\title{Enable flat shading for ggseg3d plot}
\usage{
set_flat_shading(p, flat = TRUE)
}
\arguments{
\item{p}{ggseg3d widget object}

\item{flat}{logical. Enable flat shading (default: TRUE)}
}
\value{
ggseg3d widget object with updated shading
}
\description{
Disables lighting effects to show colors exactly as specified.
Useful for screenshots where accurate color reproduction is needed,
such as atlas creation pipelines that extract contours from images.
}
\examples{
ggseg3d() |>
  set_flat_shading()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atlas-types.R
\name{prepare_mesh_atlas_data}
\alias{prepare_mesh_atlas_data}
\title{Prepare mesh-based atlas data}
\usage{
prepare_mesh_atlas_data(atlas, .data)
}
\arguments{
\item{atlas}{A mesh-based ggseg_atlas object}

\item{.data}{Optional user data to merge}
}
\value{
Prepared data frame with hemi, region, label, colour, and mesh
}
\description{
Extracts and prepares data from a mesh-based ggseg_atlas object
(subcortical/tract) for rendering. Joins meshes with core region info
and palette colours.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atlas-types.R
\name{prepare_atlas_data}
\alias{prepare_atlas_data}
\title{Prepare atlas data}
\usage{
prepare_atlas_data(atlas, .data)
}
\arguments{
\item{atlas}{A ggseg_atlas object}

\item{.data}{Optional user data to merge}
}
\value{
Prepared data frame with hemi, region, label, colour, and vertices
}
\description{
Extracts and prepares data from a ggseg_atlas object for rendering.
Joins vertices with core region info and palette colours.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_legend}
\alias{set_legend}
\title{Set legend visibility}
\usage{
set_legend(p, show = TRUE)
}
\arguments{
\item{p}{A ggseg3d or ggsegray object}

\item{show}{logical. Whether to show the legend (default: TRUE)}
}
\value{
The input object, modified
}
\description{
For htmlwidget output, toggles legend visibility. For rgl output,
draws or removes the legend overlay.
}
\examples{
ggseg3d() |> set_legend(FALSE)
\dontrun{
# rgl requires OpenGL; not run in check environments.
ggsegray(hemisphere = "left") |> set_legend()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widget-output.R
\name{ggseg3d-shiny}
\alias{ggseg3d-shiny}
\alias{ggseg3dOutput}
\alias{renderGgseg3d}
\title{Shiny bindings for ggseg3d}
\usage{
ggseg3dOutput(outputId, width = "100\%", height = "400px")

renderGgseg3d(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{output variable to read from}

\item{width, height}{Must be a valid CSS unit (like \code{'100\%'},
\code{'400px'}, \code{'auto'}) or a number, which will be coerced to a
string and have \code{'px'} appended.}

\item{expr}{An expression that generates a ggseg3d}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Is \code{expr} a quoted expression (with \code{quote()})? This
is useful if you want to save an expression in a variable.}
}
\value{
`ggseg3dOutput` returns an HTML widget output element for use in
  a Shiny UI. `renderGgseg3d` returns a render function for use in a
  Shiny server.
}
\description{
Output and render functions for using ggseg3d within Shiny
applications and interactive R Markdown documents.
}
\examples{
\dontshow{if (interactive() && rlang::is_installed("shiny")) withAutoprint(\{ # examplesIf}
library(shiny)
ui <- fluidPage(ggseg3dOutput("brain"))
server <- function(input, output) {
  output$brain <- renderGgseg3d(ggseg3d())
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{resolve_brain_mesh}
\alias{resolve_brain_mesh}
\title{Resolve brain surface mesh}
\usage{
resolve_brain_mesh(
  hemisphere = c("lh", "rh"),
  surface = c("inflated", "semi-inflated", "white", "pial", "sphere", "smoothwm", "orig"),
  brain_meshes = NULL
)
}
\arguments{
\item{hemisphere}{`"lh"` or `"rh"`}

\item{surface}{Surface type: `"inflated"`, `"semi-inflated"`, `"white"`,
`"pial"`, `"sphere"`, `"smoothwm"`, `"orig"`}

\item{brain_meshes}{Optional user-supplied mesh data. Passed through to
[ggseg.formats::get_brain_mesh()] for format details.}
}
\value{
list with vertices (data.frame with x, y, z) and faces
  (data.frame with i, j, k), or NULL if mesh not found
}
\description{
Resolves and prepares a brain surface mesh for rendering. Delegates to
[ggseg.formats::get_brain_mesh()] for inflated surfaces and to
[ggseg.meshes::get_cortical_mesh()] for pial, white, semi-inflated, and
other surfaces. Returned meshes use the shared anatomical axis convention
(`x` = left-right, `y` = anterior-posterior, `z` = superior-inferior) with
`lh` positioned at `x <= 0` and `rh` at `x >= 0`, medial edges meeting at
the midline (`x = 0`).
}
\examples{
mesh <- resolve_brain_mesh("lh", "inflated")
str(mesh, max.level = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggseg3d.R
\name{build_centerline_data}
\alias{build_centerline_data}
\title{Build centerline data for tract atlases}
\usage{
build_centerline_data(atlas, tube_radius = NULL, tube_segments = NULL)
}
\arguments{
\item{atlas}{A `tract_atlas` object}

\item{tube_radius}{Optional radius override}

\item{tube_segments}{Optional segment count override}
}
\value{
List with centerlines, tube_radius, tube_segments, or NULL
}
\description{
Extracts centerline data from a tract atlas and assembles the tube
generation parameters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_orthographic}
\alias{set_orthographic}
\title{Enable orthographic camera for ggseg3d plot}
\usage{
set_orthographic(p, ortho = TRUE, frustum_size = 220)
}
\arguments{
\item{p}{ggseg3d widget object}

\item{ortho}{logical. Enable orthographic mode (default: TRUE)}

\item{frustum_size}{numeric. Size of the orthographic frustum. Controls
how much of the scene is visible. Default 220 works well for brain meshes.
Use the same value across all views for consistent sizing.}
}
\value{
ggseg3d widget object with updated camera mode
}
\description{
Uses orthographic projection instead of perspective. This eliminates
perspective distortion and ensures consistent sizing across all views.
}
\examples{
ggseg3d() |>
  set_orthographic()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{rotate_vector}
\alias{rotate_vector}
\title{Rotate vector around axis by angle (Rodrigues' formula)}
\usage{
rotate_vector(v, axis, angle)
}
\description{
Rotate vector around axis by angle (Rodrigues' formula)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{build_cortical_meshes}
\alias{build_cortical_meshes}
\title{Build mesh list for cortical atlases}
\usage{
build_cortical_meshes(
  atlas_data,
  hemisphere,
  surface,
  na_colour,
  edge_by,
  brain_meshes = NULL,
  text_by = NULL,
  label_by = "region"
)
}
\arguments{
\item{atlas_data}{Prepared atlas data frame}

\item{hemisphere}{Hemispheres to include}

\item{surface}{Surface type}

\item{na_colour}{Colour for NA values}

\item{edge_by}{Column for edge grouping (or NULL)}

\item{brain_meshes}{Optional user-supplied brain meshes}
}
\value{
List of mesh data structures
}
\description{
Creates mesh data structures for cortical ggseg_atlas objects using
shared brain meshes with vertex-based colouring.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{vertices_to_text}
\alias{vertices_to_text}
\title{Map vertices to text values for hover display}
\usage{
vertices_to_text(atlas_data, n_vertices, text_col)
}
\arguments{
\item{atlas_data}{Data frame with vertices list column}

\item{n_vertices}{Number of vertices in the mesh}

\item{text_col}{Name of the column containing text values}
}
\value{
Character vector of text values, one per mesh vertex
}
\description{
Assigns text values to mesh vertices based on a column in atlas data.
Used for per-vertex hover text in the Three.js tooltip.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{vertices_to_labels}
\alias{vertices_to_labels}
\title{Map atlas vertex indices to region labels}
\usage{
vertices_to_labels(atlas_data, n_vertices, na_label = NA_character_)
}
\arguments{
\item{atlas_data}{Data frame with region and vertices columns}

\item{n_vertices}{Number of vertices in the mesh}

\item{na_label}{Label for vertices not in any region}
}
\value{
Character vector of labels, one per mesh vertex
}
\description{
Given a ggseg_atlas with vertices column and a brain mesh, creates a label
vector for each mesh vertex based on which region it belongs to.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{build_continuous_legend}
\alias{build_continuous_legend}
\title{Build continuous legend}
\usage{
build_continuous_legend(palette, pal_colours, colour_col, data_min, data_max)
}
\arguments{
\item{palette}{Original palette specification}

\item{pal_colours}{Processed palette colours}

\item{colour_col}{Name of the colour column (used as title)}

\item{data_min}{Minimum data value}

\item{data_max}{Maximum data value}
}
\value{
List with continuous legend specification
}
\description{
Creates a continuous colorbar legend specification for numeric data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{pan_camera}
\alias{pan_camera}
\title{Pan camera position of ggseg3d plot}
\usage{
pan_camera(p, camera)
}
\arguments{
\item{p}{A `ggseg3d` widget or `ggsegray` rgl object.}

\item{camera}{string, list, or numeric vector. Camera position preset
  name, custom eye position list, or `c(x, y, z)` for rgl.

\strong{Available camera presets:}
\itemize{
\item `left lateral` or `left_lateral`
\item `left medial` or `left_medial`
\item `right lateral` or `right_lateral`
\item `right medial` or `right_medial`
\item `left superior` or `left_superior`
\item `right superior` or `right_superior`
\item `left inferior` or `left_inferior`
\item `right inferior` or `right_inferior`
\item `left anterior` or `left_anterior`
\item `right anterior` or `right_anterior`
\item `left posterior` or `left_posterior`
\item `right posterior` or `right_posterior`
}}
}
\value{
The input object (modified), for piping.
}
\description{
Sets the camera position for a ggseg3d widget or ggsegray rgl scene
to standard anatomical views or custom positions.
}
\examples{
ggseg3d() |> pan_camera("right lateral")
\dontrun{
# rgl requires OpenGL; not run in check environments.
ggsegray(atlas = dk(), hemisphere = "left") |>
  pan_camera("left lateral")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{add_glassbrain}
\alias{add_glassbrain}
\title{Add glass brain to ggseg3d plot}
\usage{
add_glassbrain(
  p,
  hemisphere = c("left", "right"),
  surface = "inflated",
  colour = "#CCCCCC",
  opacity = 0.3,
  brain_meshes = NULL
)
}
\arguments{
\item{p}{A `ggseg3d` widget or `ggsegray` rgl object.}

\item{hemisphere}{Character vector. Hemispheres to add: "left", "right",
or both.}

\item{surface}{Character. Surface type. Defaults to `"inflated"`, which
is supplied by `ggseg.formats` and requires no additional dependency.
Other surfaces (`"pial"`, `"white"`, etc.) are drawn from `ggseg.meshes`.}

\item{colour}{Character. Colour for the glass brain surface (hex or named).}

\item{opacity}{Numeric. Transparency of the glass brain (0-1).}

\item{brain_meshes}{Optional user-supplied brain meshes. See
[ggseg.formats::get_brain_mesh()] for format details.}
}
\value{
The input object (modified), for piping.
}
\description{
Adds a translucent brain surface to a ggseg3d plot for anatomical reference.
Particularly useful for subcortical and tract visualizations where spatial
context helps interpretation. Works with both htmlwidget (`ggseg3d`) and
rgl (`ggsegray`) objects.
}
\examples{
\donttest{
ggseg3d(atlas = aseg()) |>
  add_glassbrain("left", opacity = 0.2)
}
\dontrun{
# rgl requires OpenGL; not run in check environments.
ggsegray(atlas = aseg()) |>
  add_glassbrain(opacity = 0.15) |>
  pan_camera("right lateral")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colour.R
\name{apply_colour_palette}
\alias{apply_colour_palette}
\title{Apply colour palette}
\usage{
apply_colour_palette(atlas_data, colour, palette, na_colour)
}
\arguments{
\item{atlas_data}{Atlas data frame with vertices column}

\item{colour}{Column name for colour values}

\item{palette}{Colour palette specification}

\item{na_colour}{Colour for NA values}
}
\value{
List with data, fill column name, palette, and colour metadata
}
\description{
Processes colour mapping for ggseg_atlas objects using
vertex-based colouring.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{vertices_to_colors}
\alias{vertices_to_colors}
\title{Map atlas vertex indices to mesh colors}
\usage{
vertices_to_colors(atlas_data, n_vertices, na_colour = "#CCCCCC")
}
\arguments{
\item{atlas_data}{Data frame with region, colour, and vertices columns}

\item{n_vertices}{Number of vertices in the mesh}

\item{na_colour}{Color for vertices not in any region}
}
\value{
Character vector of colors, one per mesh vertex
}
\description{
Given a ggseg_atlas with vertices column and a brain mesh, creates a color
vector for each mesh vertex based on which region it belongs to.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggseg3d-package.R
\docType{package}
\name{ggseg3d-package}
\alias{ggseg3d-package}
\title{ggseg3d: Plot brain segmentations in 3D}
\description{
Plotting results from analyses done on data derived from brain
segmentations is a common need, but may be quite laborious.
Results from such analyses are usually easier to interpret if
the plot can mimmick the shape and position in the brain it
represents.
}
\details{
This package contains data from various brain parcellations,
with convenient functions to inspect the results directly on
a brain-plot using interactive 3D mesh visualizations powered
by Three.js via htmlwidgets.

The package uses `ggseg_atlas` objects from ggseg.formats
that contain 3D vertex mappings.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ggsegverse/ggseg3d}
  \item \url{https://ggsegverse.github.io/ggseg3d/}
  \item Report bugs at \url{https://github.com/ggsegverse/ggseg3d/issues}
}

}
\author{
\strong{Maintainer}: Athanasia Mo Mowinckel \email{a.m.mowinckel@psykologi.uio.no} (\href{https://orcid.org/0000-0002-5756-0223}{ORCID})

Authors:
\itemize{
  \item Didac Vidal-Piñeiro \email{d.v.pineiro@psykologi.uio.no} (\href{https://orcid.org/0000-0001-9997-9156}{ORCID})
}

Other contributors:
\itemize{
  \item Center for Lifespan Changes in Brain and Cognition (LCBC), University of Oslo [copyright holder]
  \item three.js authors (Three.js and OrbitControls libraries, see LICENSE.note) [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{set_background}
\alias{set_background}
\title{Set background color of ggseg3d plot}
\usage{
set_background(p, colour = "#ffffff")
}
\arguments{
\item{p}{A `ggseg3d` widget or `ggsegray` rgl object.}

\item{colour}{string. Background color (hex or named color)}
}
\value{
The input object (modified), for piping.
}
\description{
Changes the background color of a ggseg3d widget or ggsegray rgl scene.
}
\examples{
ggseg3d() |> set_background("black")
\dontrun{
# rgl requires OpenGL; not run in check environments.
ggsegray(atlas = dk()) |> set_background("black")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{generate_tube_mesh}
\alias{generate_tube_mesh}
\title{Generate tube mesh from centerline}
\usage{
generate_tube_mesh(centerline, radius = 0.5, segments = 8)
}
\arguments{
\item{centerline}{Matrix with N rows and 3 columns (x, y, z coordinates)}

\item{radius}{Tube radius. Either a single value or vector of length N.}

\item{segments}{Number of segments around tube circumference.}
}
\value{
List with vertices (data.frame), faces (data.frame), and metadata
}
\description{
Creates a 3D tube mesh around a centerline path using parallel transport
frames for smooth geometry without twisting artifacts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widget.R
\name{create_ggseg3d_widget}
\alias{create_ggseg3d_widget}
\title{Create ggseg3d htmlwidget}
\usage{
create_ggseg3d_widget(meshes, legend_data)
}
\arguments{
\item{meshes}{List of mesh data structures}

\item{legend_data}{Legend specification (or NULL)}
}
\value{
An htmlwidget object of class "ggseg3d"
}
\description{
Creates the final htmlwidget object with default options and sizing policy.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{build_subcortical_meshes}
\alias{build_subcortical_meshes}
\title{Build mesh list for subcortical atlases}
\usage{
build_subcortical_meshes(
  atlas_data,
  na_colour,
  text_by = NULL,
  label_by = "region"
)
}
\arguments{
\item{atlas_data}{Prepared atlas data frame with label, colour, and mesh
columns}

\item{na_colour}{Colour for NA values}
}
\value{
List of mesh data structures
}
\description{
Creates mesh data structures for subcortical atlases with per-region mesh
data using face-based colouring (each structure is a separate mesh).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additions.R
\name{snapshot_brain}
\alias{snapshot_brain}
\title{Save ggseg3d widget as image}
\usage{
snapshot_brain(p, file, width = 600, height = 500, delay = 1, zoom = 2, ...)
}
\arguments{
\item{p}{ggseg3d widget object}

\item{file}{string. Output file path (should end in .png)}

\item{width}{numeric. Image width in pixels (default: 600)}

\item{height}{numeric. Image height in pixels (default: 500)}

\item{delay}{numeric. Seconds to wait for widget to render before capture
(default: 1)}

\item{zoom}{numeric. Zoom factor for higher resolution (default: 2)}

\item{...}{Additional arguments passed to webshot2::webshot}
}
\value{
The file path (invisibly)
}
\description{
Takes a screenshot of a ggseg3d widget and saves it as a PNG image.
Requires a Chrome-based browser to be installed.
}
\examples{
\dontrun{
ggseg3d() |>
  pan_camera("left lateral") |>
  snapshot_brain("brain.png")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widget-output.R
\name{updateGgseg3dCamera}
\alias{updateGgseg3dCamera}
\title{Update camera in Shiny}
\usage{
updateGgseg3dCamera(session, outputId, camera)
}
\arguments{
\item{session}{The Shiny session object}

\item{outputId}{The output ID of the ggseg3d widget}

\item{camera}{Camera position preset or custom position}
}
\value{
None, called for side effects (sends message to client)
}
\description{
Sends a message to update the camera position of a ggseg3d widget
in a Shiny app.
}
\examples{
\dontshow{if (interactive() && rlang::is_installed("shiny")) withAutoprint(\{ # examplesIf}
\dontrun{
updateGgseg3dCamera(session, "brain", "left lateral")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brain-mesh.R
\name{tangents_to_colors}
\alias{tangents_to_colors}
\title{Convert tangent vectors to orientation RGB colours}
\usage{
tangents_to_colors(mesh_data)
}
\arguments{
\item{mesh_data}{Mesh data with vertices data.frame and metadata list}
}
\value{
Character vector of hex colours (one per mesh vertex)
}
\description{
Computes direction-based RGB colours from centerline tangent vectors.
Standard tractography colouring: R = left-right (x),
G = anterior-posterior (y), B = superior-inferior (z).
}
\keyword{internal}
